/* $Id$
 *
 * dhcpcd - DHCP client daemon -
 * Copyright (C) 1996 - 1997 Yoichi Hariguchi <yoichi@fore.com>
 *
 * Dhcpcd is an RFC2131 and RFC1541 compliant DHCP client daemon.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <sys/types.h>
#include <string.h>
#include <netinet/in.h>

#include <sys/stat.h> /* for write() DEBUG */
#include <fcntl.h> /* for write() DEBUG */
#include <unistd.h> /* for write() DEBUG */

#include "dhcp.h"
#include "dhcp-options.h"
#include "error-handler.h"
#include "memory.h"

extern int acceptBootpPackets;
extern int sizeOptBuf;
int next_128plus_option = 128;


/* length of each option. -1 means 'n'.
 */
static char OptLen[256] = {
     0,			/* 0: pad */
     4,
     4,
    -1,
    -1,
    -1,			/* 5: Name Server */
    -1,
    -1,
    -1,
    -1,
    -1,			/* 10: Impress Server */
    -1,
    -1,
     2,
    -1,
    -1,			/* 15: Domain Name */
    -1,
    -1,
    -1,
     1,
     1,			/* 20: Non-Local Source Routing Enable/Disable */
    -1,
     2,
     1,
     4,
    -1,			/* 25: Path MTU Plateau Table */
     2,
     1,
     4,
     1,
     1,			/* 30: Mask Supplier */
     1,
     4,
    -1,
     1,
     4,			/* 35: ARP Cache Timeout */
     1,
     1,
     4,
     1,
    -1,			/* 40: NIS Domain */
    -1,
    -1,
    -1,
    -1,
    -1,			/* 45: NetBIOS over TCP/IP... */
     1,
    -1,
    -1,
    -1,
     4,			/* 50: Requested IP address (DHCP) */
     4,
     1,
     1,
     4,
    -1,			/* 55: Parameter Request List (DHCP) */
    -1,
     2,
     4,
     4,
    -1,			/* 60: Class Identifier (DHCP) */
    -1,

    /* New options */
    -1,			/* 62: */
    -1,			/* 63: */
    -1,			/* 64: nis plus domain */
    -1,			/* 65: nis plus server */
    -1,			/* 66: TFTP server */
    -1,			/* 67: Bootfile */
    -1,			/* 68: Mobile IP Home agent */
    -1,			/* 69: SMTP server */
    -1,			/* 70: Pop3server */
    -1,			/* 71: NNTP server */
    -1,			/* 72: WWW server */
    -1,			/* 73: Finger server */
    -1,			/* 74: IRC server */
    -1,			/* 75: StreetTalk server */
     0,
};

static u_char *vendor_option_ptr = NULL;


/* opt must have at least 312 elements.
 */
int
sub_getOptions(u_char *optp[], u_char *p, u_char *end, int *option_overload,
	       int i)
{
    int len;
    while ( p < end )
    {
	if ( *p == endOption ) { /* end */
	    return i;
	}
	if ( *p == dhcpOptionOverload )
	{
	    *option_overload = *p + 1;
	    p+=2;
	    continue;
	}
	if ( *p == padOption ) { /* pad */
	    ++p;
	    continue;
	}
	if ( *p == vendorSpecificInfo )
	{
	    vendor_option_ptr = p;
	    p += p[1] + 2;
	    continue;
	}


	/* get the length of this tag code
	 */
	if (*p >= 128 && *p <= 254) /* User option */
	    len = p[1];
	else
	    len = (OptLen[*p] == -1) ? p[1] : OptLen[*p];
	optp[i++] = p;			/* store the pointer to the tag */
	p += len + 2;			/* set the pointer to the next tag */
    }
    return i;
}

/* Process option 43 and map to normal space (127+WyseOptionNumber) */
int
sub_getVendorOptions(u_char *optp[], u_char *p, u_char *end,
		     int *option_overload, int i)
{
    int len;
    char buf[512], *q;

    while ( p < end )
    {
	if ( *p == endOption ) { /* end */
	    return i;
	}
	if ( *p == padOption ) { /* pad */
	    ++p;
	    continue;
	}

	len = p[1];
	buf[0] = '\0';
	switch(*p)
	{
	case WyseHeader: /* Just skip the option */
	    break;
	case WyseNetSvc:
	    buf[0] = 128;
	    strcpy(buf+2, "Q080_NETSVC=");
	    q = strchr(buf+2, '=') + 1;
	    memcpy(q, p+2, p[1]);
	    q[p[1]] = '\0';
	    break;
	case WyseNoSwap:
	    buf[0] = 128;
	    strcpy(buf+2, "Q080_NOSWAP=YES");
	    break;
	case WyseReflash:
	    buf[0] = 128;
	    strcpy(buf+2, "Q080_REFLASH=YES");
	    break;
	case WyseMfgDir:
	    buf[0] = 128;
	    strcpy(buf+2, "Q080_MF_DIR=");
	    q = strchr(buf+2, '=') + 1;
	    memcpy(q, p+2, p[1]);
	    q[p[1]] = '\0';
	    break;
	case WyseMfgCfg:
	    buf[0] = 128;
	    strcpy(buf+2, "Q080_MF_CFG=");
	    q = strchr(buf+2, '=') + 1;
	    memcpy(q, p+2, p[1]);
	    q[p[1]] = '\0';
	    break;
	case WyseXFS:
	    buf[0] = 128;
	    strcpy(buf+2, "Q080_XFS=");
	    q = strchr(buf+2, '=') + 1;
	    memcpy(q, p+2, p[1]);
	    q[p[1]] = '\0';
	    break;
	case WyseXDMCP:
	    buf[0] = 128;
	    strcpy(buf+2, "Q080_XDMCP=");
	    q = strchr(buf+2, '=') + 1;
	    memcpy(q, p+2, p[1]);
	    q[p[1]] = '\0';
	    break;
	case WyseBuddyBoot:
	    buf[0] = 128;
	    strcpy(buf+2, "Q080_BUDDY_BOOT=YES");
	    break;
	default: /* Invalid, just skip the option */
	    break;
	}
	if (buf[0])
	{
	    buf[1] = strlen(buf+2) + 1;
	    optp[i++] = strdup(buf);	/* store the pointer to the tag */
	}
	p += len + 2;			/* set the pointer to the next tag */
    }
    return i;
}

void
getOptions(u_char *optp[], dhcpInMessage *msg)
{
    u_char *p;
    u_char *end;
    int	i, j;
    int overload_value = 0;

    vendor_option_ptr = NULL; /* initialize */
    p	= msg->options + 4;		/* skip magic cookie */
    end = msg-> options + sizeof(msg->options);	/* last element + 1 */

    i = sub_getOptions(optp, p, end, &overload_value, 0);

    /* If there were option overload values (option 51), append that set here.
     */
    if (overload_value == '1' || overload_value == '3')
    {
	/* file chunk */
	p = msg->file;
	end = msg->file + sizeof(msg->file);
	i = sub_getOptions(optp, p, end, &j, i); /*skip option 51 replacement */
    }
    if (overload_value == '2' || overload_value == '3')
    {
	/* sname chunk */
	p = msg->sname;
	end = msg->sname + sizeof(msg->sname);
	i = sub_getOptions(optp, p, end, &j, i); /*skip option 51 replacement */
    }
    if (vendor_option_ptr != NULL)
    {
	unsigned long *q;
	/* vendor specific option (Option 43) chunk */
	p = vendor_option_ptr;
	p+=2; /* skip header and length */
	end = p + vendor_option_ptr[1];

	/* Don't bother if the option doesn't check correctly */
	q = (unsigned long *) (&p[2]);
	if (p[0] == 1 && p[1] == 4 && ntohl(*q) == 0x76583A1C)
	    i = sub_getVendorOptions(optp, p, end, &j, i);
    }

    /* Extensions path stuff will be processed later on */
    return;
}

/* optp must contain pointers to msg->option
 * return 1 if msg is good
 * return 0 if msg is not good
 * optp[Oxxx] has the pointer to the option field if msg is good
 * see dhcp.h for more details on Oxxx
 * CAUTION: This function uses DhcpMsgSend directly!!
 */
int
parseDhcpMsg(u_char *optp[], dhcpInMessage *msg)
{
    char    buf[512];
    u_char *opt[N_SUPPORT_OPTIONS];
    u_char *inopt[N_SUPPORT_OPTIONS]; /* Scratch space */
    u_char **p = inopt;

    if ( msg->op != BOOTREPLY ) {
	return 0;				/* NG */
    }
    if ( msg->xid != DhcpMsgSend.xid ) {
	return 0;				/* NG */
    }
    if ( bcmp(msg->chaddr, DhcpMsgSend.chaddr, DhcpMsgSend.hlen) ) {
	return 0;				/* NG */
    }
    bzero((char *)opt, sizeof(opt));
    bzero((char *)inopt, sizeof(inopt));
    next_128plus_option = 128;

    getOptions(p, msg);			/* get pointers to each option */

    /* At this point, inopt(p) contains an array of the options with their
     * data in the order pulled. Unsupported options have been stripped out.
     *
     * Note that these elements are option number + length + data.
     *
     * The following loop will take these elements and put them in a specific
     * spot within the opt array.  When done, opt will be copied over optp
     * for passing back up to the parent.
     */
    for (; *p != NULL; ++p )
    {
	switch ( **p ) {
	  case dhcpMessageType:
	    if (opt[OmsgType] == NULL)
		opt[OmsgType] = *p + 1;
	    break;
	  case dhcpServerIdentifier:
	    if (opt[OserverInaddr] == NULL)
		opt[OserverInaddr] = *p + 1;
	    break;
	  case timerOffset:
	    if (opt[OtimerOffset] == NULL)
		opt[OtimerOffset] = *p + 1;
	    break;
	  case dhcpIPaddrLeaseTime:
	    if (opt[OleaseTime] == NULL)
		opt[OleaseTime] = *p + 1;
	    break;
	  case dhcpT1value:
	    if (opt[OrenewalTime] == NULL)
		opt[OrenewalTime] = *p + 1;
	    break;
	  case dhcpT2value:
	    if (opt[OrebindTime] == NULL)
		opt[OrebindTime] = *p + 1;
	    break;
	  case subnetMask:
	    if (opt[Onetmask] == NULL)
		opt[Onetmask] = *p + 1;
	    break;
	  case broadcastAddr:
	    if (opt[ObcastInaddr] == NULL)
		opt[ObcastInaddr] = *p + 1;
	    break;
	  case timeServer:
	    if (opt[OtimeServer] == NULL)
		opt[OtimeServer] = ( (*p)[1]%4 ) ? NULL : *p + 1;
	    break;
	  case dns:
	    if (opt[Odns] == NULL)
		opt[Odns] = ( (*p)[1]%4 ) ? NULL : *p + 1;
	    break;
	  case lprServer:
	    if (opt[OlprServer] == NULL)
		opt[OlprServer] = ( (*p)[1]%4 ) ? NULL : *p + 1;
	    break;
	  case xFontServer:
	    if (opt[OxFontServer] == NULL)
		opt[OxFontServer] = ( (*p)[1]%4 ) ? NULL : *p + 1;
	    break;
	  case xDisplayManager:
	    if (opt[OxDisplayManager] == NULL)
		opt[OxDisplayManager] = ( (*p)[1]%4 ) ? NULL : *p + 1;
	    break;
	  case hostName:
	    if (opt[OhostName] == NULL)
		opt[OhostName] = *p + 1;
	    break;
	  case domainName:
	    if (opt[OdomainName] == NULL)
		opt[OdomainName] = *p + 1;
	    break;
	  case nisDomainName:
	    if (opt[OnisDomName] == NULL)
		opt[OnisDomName] = *p + 1;
	    break;
	  case ntpServers:
	    if (opt[OntpServer] == NULL)
		opt[OntpServer] = ( (*p)[1]%4 ) ? NULL : *p + 1;
	    break;
	  case dhcpMsg:
	    strcpy(buf, "msg from the DHCP server: ");
	    strncat(buf, *p + 2, (*p)[1]);
	    logRet(buf);
	    if (opt[OdhcpMessage] == NULL)
		opt[OdhcpMessage] = *p + 1;
	    break;
	  case dhcpClassIdentifier:
	    if (opt[OdhcpClassId] == NULL)
		opt[OdhcpClassId] = ( (*p)[1]%4 ) ? NULL : *p + 1;
	    break;
	  case routersOnSubnet:
	    if (opt[Orouter] == NULL)
		opt[Orouter] = ( (*p)[1]%4 ) ? NULL : *p + 1;
	    break;
	  case nisServers:
	    if (opt[OnisServer] == NULL)
		opt[OnisServer] = ( (*p)[1]%4 ) ? NULL : *p + 1;
	    break;
	  case nisplusDomain:
	    if (opt[OnisplusDomain] == NULL)
		opt[OnisplusDomain] = *p + 1;
	    break;
	  case nisplusServer:
	    if (opt[OnisplusServer] == NULL)
		opt[OnisplusServer] = ( (*p)[1]%4 ) ? NULL : *p + 1;
	    break;
	  case SMTPServer:
	    if (opt[OSMTPServer] == NULL)
		opt[OSMTPServer] = ( (*p)[1]%4 ) ? NULL : *p + 1;
	    break;
	  case POP3Server:
	    if (opt[OPOP3Server] == NULL)
		opt[OPOP3Server] = ( (*p)[1]%4 ) ? NULL : *p + 1;
	    break;
	  case extentionsPath:
	    if (opt[OextensionsPath] == NULL)
		opt[OextensionsPath] = *p + 1;
	    break;
	  default:
	    /* 128-254 */
	    if (**p >= 128 && **p <= 254)
	    {
		/* If it is a Wyse option, save it, otherwise ignore it */
		if ((strncmp("0080_", *p + 2, 5) == 0) ||
		    (strncmp("Q080_", *p + 2, 5) == 0))
		{
		    opt[(next_128plus_option - 128) + OBaseUserOptions] =
			*p + 1;
		    next_128plus_option++;
		}
	    }
	    break;
	}
    }

    /* check option field
     */
    if ( opt[OmsgType] == NULL)	/* illegal DHCP msg, or BOOTP msg */
    {
	if ( acceptBootpPackets) {
	    static unsigned char null_msgtype[2] = {1,0};
	    opt[OmsgType] = smalloc(sizeof(null_msgtype));
	    memcpy(opt[OmsgType], null_msgtype, sizeof(null_msgtype));
	    bcopy(opt, optp, sizeof(opt));
	    return 1;
	} else {
	    return 0;
	}
    }
    /* DHCPOFFER & DHCPACK msgs must have server IP addres and lease time
     */
    switch ( opt[OmsgType][1] )
    {
      case DHCP_OFFER:
	if ( opt[OleaseTime] == NULL )
	{
	    logRet("No leasetime (msgType: %d) (parseDhcpMsg)", opt[OmsgType][1]);
	    return 0;
	}
      case DHCP_ACK:
	if ( opt[OserverInaddr] == NULL )
	{
	    logRet("No server IP address (msgType: %d) (parseDhcpMsg)", opt[OmsgType][1]);
	    return 0;
	}
	break;
      case DHCP_INFORM:
      	/* If we ever get an INFORM packet from anywhere, we must ignore it.
	 * Microsoft Windows 2000 being used as a server will apparently
	 * have the partially-running client send an inform out with option
	 * 43 data but not a request for data.  This packet MUST be absorbed
	 * without action and should be absorbed quietly.
	 */
      	return 0;
      default:
	if ( opt[OleaseTime] == NULL && !acceptBootpPackets)
	{
	    logRet("No leasetime (msgType: %d) (parseDhcpMsg)", opt[OmsgType][1]);
	    return 0;
	}
	break;
    }
    bcopy(opt, optp, sizeof(opt));
    return 1;
}
