/* 
	arcd.c	(c) 1997  Grant R. Guenther <grant@torque.net>
		          Under the terms of the GNU public license.

	arcd.c is a driver for ATAPI CDroms based on an old version of
        the DataStor parallel port IDE adapter.  The driver was created
        to support the Arista ParaCD 525 product, on a specific request.
        Currently, the driver supports the 4-bit and 8-bit modes *only*.  
	I do not have one of these devices, the driver was created 
        using data traces supplied by Corey Stotts (cstotts@avalon.net).

        Ensure that /usr/include/linux and /usr/include/asm are 
	links to the correct include files for the target system.  Compile 
        the driver with

                cc -D__KERNEL__ -DMODULE -O2 -c arcd.c

        You must then load it with insmod.  If you are using
        MODVERSIONS, add the following to the cc command:

		-DMODVERSIONS -include /usr/include/linux/modversions.h

	Before attempting to access the new driver, you will need to
        create a new device special file.  The following commands will
	do that for you:

		mknod /dev/arcd b 63 0
		chown root:disk /dev/arcd
		chmod 660 /dev/arcd

	Load the driver with a command like:

		insmod arcd

	Afterward, you can mount a disk in the usual way:

		mount -t iso9660 /dev/arcd /cdrom

	(assuming you have made a directory /cdrom to use as a
        mount point).

	If you load the module with

       		insmod arcd ar_debug=1

	the driver will log verbose messages as it probes
        for the appropriate port and mode.  If you have problems 
        loading the driver, please include this output with your
        report.

*/

#define	AR_VERSION	"0.03" 

#define	AR_BASE		0	/* set to 0 for autoprobe */
#define AR_HOLD         4       /* micro-second port settling time */
#define AR_DEBUG	1	/* enable debug messages */

#include <linux/version.h>
#include <linux/module.h>

#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/fs.h>
#include <linux/major.h>
#include <linux/kernel.h>
#include <linux/tqueue.h>
#include <linux/delay.h>
#include <linux/cdrom.h>
#include <linux/ioport.h>

#include <asm/system.h>
#include <asm/io.h>
#include <asm/segment.h>   /* change to uaccess.h for 2.1 kernels ... */

#ifndef ARCD_MAJOR
#define ARCD_MAJOR   63	
#endif

#define MAJOR_NR ARCD_MAJOR

/* set up defines for blk.h,  why don't all drivers do it this way ? */

#define DEVICE_NAME "ParaCD"
#define DEVICE_REQUEST do_ar_request
#define DEVICE_NR(device) (MINOR(device))
#define DEVICE_ON(device)
#define DEVICE_OFF(device)

#include <linux/blk.h>

#define AR_TMO		   300		/* timeout in jiffies */
#define AR_DELAY            50          /* spin delay in uS */

#define AR_SPIN		(10000/AR_DELAY)*AR_TMO

int arcd_init(void);
void arcd_setup(char * str, int * ints);
void cleanup_module( void );

static int 	ar_open(struct inode *inode, struct file *file);
static void 	do_ar_request(void);
static void 	do_ar_read(void);
static int 	ar_ioctl(struct inode *inode,struct file *file,
                         unsigned int cmd, unsigned long arg);
static void 	ar_release (struct inode *inode, struct file *file);

static int 	ar_detect(void);
static int      ar_lock(void);
static void     ar_unlock(void);
static void     ar_eject(void);
static void     ar_interrupt(void *data);

static int	ar_base = AR_BASE;
static int      ar_mode = 0;
static int	ar_hold = AR_HOLD;
static int      ar_debug = AR_DEBUG;
static int      ar_found_r0;
static int      ar_found_r2;

static int ar_access = 0;		/* count of active opens ... */
static int ar_busy = 0;			/* request being processed ? */
static int ar_timeout;                  /* "interrupt" loop limiter */
static int ar_sector;			/* address of next requested sector */
static int ar_count;			/* number of blocks still to do */
static char * ar_buf;			/* buffer for request in progress */
static char ar_buffer[2048];		/* raw block buffer */
static int ar_bufblk = -1;		/* block in buffer, in CD units,
					   -1 for nothing there */

static struct tq_struct ar_tq = {0,0,ar_interrupt,NULL}; 

/* kernel glue structures */

static struct file_operations ar_fops = {
	NULL,			/* lseek - default */
	block_read,		/* read - general block-dev read */
	block_write,		/* write - general block-dev write */
	NULL,			/* readdir - bad */
	NULL,			/* select */
	ar_ioctl,		/* ioctl */
	NULL,			/* mmap */
	ar_open,		/* open */
	ar_release,		/* release */
	block_fsync,		/* fsync */
	NULL,			/* fasync */
	NULL,                   /* media change ? */
	NULL			/* revalidate new media */
};

int arcd_init (void)	/* preliminary initialisation */

{       if (ar_detect()) return -1;

	if (register_blkdev(MAJOR_NR,"arcd",&ar_fops)) {
		printk("arcd: unable to get major number %d\n",MAJOR_NR);
		return -1;
	}
	blk_dev[MAJOR_NR].request_fn = DEVICE_REQUEST;
	read_ahead[MAJOR_NR] = 8;	/* 8 sector (4kB) read ahead */

	return 0;
}

static int ar_open (struct inode *inode, struct file *file)

{	
	if (file->f_mode & 2) return -EROFS;  /* wants to write ? */

	MOD_INC_USE_COUNT;

	if (!ar_access)
	   if (ar_lock()) {
		MOD_DEC_USE_COUNT;
		return -ENXIO;
	   }
	ar_access++;
	return 0;
}

static void do_ar_request (void)

{       if (ar_busy) return;
        while (1) {
	    if ((!CURRENT) || (CURRENT->rq_status == RQ_INACTIVE)) return;
	    INIT_REQUEST;
	    if (CURRENT->cmd == READ) {
	        ar_sector = CURRENT->sector;
	        ar_count = CURRENT->nr_sectors;
	        ar_buf = CURRENT->buffer;
	        do_ar_read();
		if (ar_busy) return;
	    } 
	    else end_request(0);
	}
}

static int ar_ioctl(struct inode *inode,struct file *file,
		    unsigned int cmd, unsigned long arg)

/* we currently support only the EJECT ioctl. */

{	switch (cmd) {
            case CDROMEJECT: if (ar_access == 1) {
				ar_eject();
				return 0;
			     }
	    default:
	        return -EINVAL;
	}
}

static void ar_release (struct inode *inode, struct file *file)

{	kdev_t	devp;

	ar_access--;
	if (!ar_access) { 
	    devp = inode->i_rdev;
	    fsync_dev(devp);
	    invalidate_inodes(devp);
	    invalidate_buffers(devp);
	    ar_unlock();
	}
	MOD_DEC_USE_COUNT;
}

#ifdef MODULE

/* Glue for modules ... */

int	init_module(void)

{	int	err;
	long    flags;

	save_flags(flags);
	cli();

	err = arcd_init();

	restore_flags(flags);
	return err;
}

void	cleanup_module(void)

{	long flags;

	save_flags(flags);
	cli();
	unregister_blkdev(MAJOR_NR,"arcd");
	release_region(ar_base,3);
	restore_flags(flags);
}

#else 

/* arcd_setup:  process lilo command parameters ...

   syntax:	arcd=base[,hold]
*/

void    arcd_setup(char *str, int *ints)

{       if (ints[0] > 0) ar_base = ints[1];
        if (ints[0] > 1) ar_mode = ints[2];
        if (ints[0] > 2) ar_hold = ints[3];
} 

#endif

#define out_p(port,byte)  outb(byte,ar_base+port);udelay(ar_hold);
#define in_p(port)	  (udelay(ar_hold),inb(ar_base+port))

#define w0(byte)  		out_p(0,byte)
#define r0()      		(in_p(0) & 0xff)
#define w1(byte)  		out_p(1,byte)
#define r1()      		(in_p(1) & 0xff)
#define r2()			(in_p(2) & 0xff)
#define w2(byte)  		out_p(2,byte)

#define j44(a,b)  (((a>>3)&0x0f)|((b<<1)&0xf0))

static int ar_read_regr( char regr )

{       int l, h;

	switch (ar_mode) {

	case 0: w0(regr+8); w2(5); w2(0xd); w2(0xd); w2(5); w2(4);
		w0(0); w2(6); l = r1(); w0(0x80); h = r1(); w2(4);
		return j44(l,h);

	case 1: w0(regr+0x28); w2(5); w2(0xd); w2(0xd); w2(5); w2(4);
		w0(0); w2(0x26); h = r0(); w2(4);
		return h;

	}

	return -1;
}	

static void ar_write_regr( char regr, char val )

{	w0(regr+8); w2(5); w2(0xd); w2(0xd); w2(5); w2(4);
	w0(val); w2(5); w2(7); w2(7); w2(5); w2(4);
}

static void ar_write_cmd( char * cmd, int len )

{	int i;

	w0(0x68); w2(5); w2(0xd); w2(0xd); w2(5); w2(4);
	for(i=0;i<len;i++) { w2(5); w0(cmd[i^1]); w2(7); }
	w2(5); w2(4);
}

static void ar_read_data( char * buf, int len )

{	int i, l, h;

	switch (ar_mode) {

	case 0:	w0(0x48); w2(5); w2(0xd); w2(0xd); w2(5); w2(4);
		for(i=0;i<len;i++) {   
			w0(0); w2(6); l = r1(); 
			w0(0x80); h = r1(); w2(4);
	        	buf[i] = j44(l,h);
		}
		break;

	case 1: w0(0x68); w2(5); w2(0xd); w2(0xd); w2(5); w2(4); w0(0);
		for(i=0;i<len;i++) {
			w2(0x26); buf[i] = r0(); w2(0x24);
			}
		w2(4);
		break;
	}
}

static void ar_connect ( void  )

{       ar_found_r0 = r0();
	ar_found_r2 = r2();
	w2(4); w0(0xff); w2(6);
	w2(4); w0(0xaa); w2(6);
	w2(4); w0(0x00); w2(6);
	w2(4); w0(0x87); w2(6);
	w2(4); w0(0xe8); w2(0xc); w2(0xc); w2(4);
}

static void ar_disconnect ( void )

{	w2(0); w2(0); w2(0); w2(4);
	w2(ar_found_r2);
	w0(ar_found_r0);
} 

static inline int ar_ready( void )

{	return (r1() & 0x40);
}

static int ar_test_proto( void )

{	int i, e;

	e = 0;
	ar_connect();
	for(i=0;i<256;i++) {
		ar_write_regr(2,i^0xaa);
		ar_write_regr(3,i^0x55);
		if (ar_read_regr(2) != (i^0xaa)) e++;
		}
	ar_disconnect();	

	if (ar_debug) 
	   printk("arcd TEST port: 0x%x mode: %d errors: %d\n",
			ar_base,ar_mode,e);

	return e;
}

static int ar_wait_drq( char * fun, char * msg1, char * msg2 )

{	int j, r, e;

	j = 0;
	while (!(ar_read_regr(7)&8)&&(j++<AR_SPIN)) udelay(AR_DELAY);
        r = ar_read_regr(7);
	e = ar_read_regr(1);
	if (j >= AR_SPIN) {
	    printk("arcd: %s%s: DRQ timeout on %s, status: %x error: %x\n",
			     fun,msg1,msg2,r,e);
	    return -1;
	}
	return 0;
}

static int ar_wait( char * fun, char * msg1, char * msg2 )

{	int j, r, e;

	j = 0;
	while ((!(ar_ready()))&&(j++<AR_SPIN)) udelay(AR_DELAY);
        r = ar_read_regr(7);
	e = ar_read_regr(1);
	if ((j >= AR_SPIN) || (r & 1)) {
	    if (j >= AR_SPIN) e |= 0x100;
	    printk("arcd: %s%s: Error %s, status: %x error: %x\n",
			  fun,msg1,msg2,r,e);
	    return -1;
	}
	return 0;
}

static int ar_req_sense( char * fun, char * msg )

{	char	rs_cmd[12] = { 0x03,0,0,0,18,0,0,0,0,0,0,0 };
	int n, r;

        ar_write_regr(4,18);
        ar_write_regr(5,0);
        ar_write_regr(6,0xa0);          /* drive 0 */
        ar_write_regr(7,0xa0);          /* ATAPI packet command */
	if (ar_wait_drq(fun,msg,"request sense command")) return -1;
	ar_write_cmd(rs_cmd,12);
	if (!(r=ar_wait(fun,msg,"getting request sense data"))) {
            if (ar_read_regr(2) == 2) {
                n = (ar_read_regr(4) + 256*ar_read_regr(5));
                ar_read_data(ar_buffer,n);
                r=ar_wait(fun,msg,"on request sense done");
                if (!r)
	          printk("arcd: %s%s: sense key: %x, ASC: %x, ASQ: %x\n",
	            fun, msg, ar_buffer[2]&0xf, ar_buffer[12], ar_buffer[13]);
	    }
	}
	return r;
}

static int ar_command( char * cmd, int dlen, char * fun )

{	int	r;

	ar_connect();

	ar_write_regr(4,dlen % 256);
	ar_write_regr(5,dlen / 256);
	ar_write_regr(6,0xa0);		/* drive 0 */
	ar_write_regr(7,0xa0);		/* ATAPI packet command */

        if (ar_read_regr(7)&1) {
            if (ar_req_sense(fun," before command")) { 
                ar_disconnect();
                return -1;
                }
            ar_write_regr(4,dlen % 256);
            ar_write_regr(5,dlen / 256);
            ar_write_regr(6,0xa0);          /* drive 0 */
            ar_write_regr(7,0xa0);          /* ATAPI packet command */
	}

	if ((r=ar_wait_drq(fun,"","sending  command"))) {
	   ar_disconnect();
	   return r;
	}

	ar_write_cmd(cmd,12);
	return 0;
}

static int ar_completion( char * fun )

{	int r, n;

	if (!(r=ar_wait(fun,"","reading data"))) {
	    if (ar_read_regr(2) == 2) {
	        n = (ar_read_regr(4) + 256*ar_read_regr(5));
	        ar_read_data(ar_buffer,n);
	        r=ar_wait(fun,"","waiting after data");
	    }
	}
	if (r) r = ar_req_sense(fun," after command");

	ar_disconnect(); 
	return r;
}

static int ar_atapi( char * cmd, int dlen, char * fun )

{	int r;

	if (!(r=ar_command(cmd,dlen,fun)))
	  r = ar_completion(fun);
	return r;
}

static int ar_lock(void)

{	char	lo_cmd[12] = { 0x1e,0,0,0,1,0,0,0,0,0,0,0 };
	char	cl_cmd[12] = { 0x1b,0,0,0,3,0,0,0,0,0,0,0 };

	ar_atapi(cl_cmd,0,"close door");
	return ar_atapi(lo_cmd,0,"lock door");
}

static void ar_unlock( void)

{	char	un_cmd[12] = { 0x1e,0,0,0,0,0,0,0,0,0,0,0 };

	ar_atapi(un_cmd,0,"unlock door");
}

static void ar_eject( void)

{	char	ej_cmd[12] = { 0x1b,0,0,0,2,0,0,0,0,0,0,0 };

	ar_unlock();
	ar_atapi(ej_cmd,0,"eject");
}

static int ar_reset( void )

/* the ATAPI standard actually specifies the contents of all 7 registers
   after a reset, but the specification is ambiguous concerning the last
   two bytes, and different drives interpret the standard differently.
*/

{	int	i, flg;
	int	expect[5] = {1,1,1,0x14,0xeb};
	long	flags;

	ar_connect();
	ar_write_regr(6,0xa0);
	ar_write_regr(7,8);

	save_flags(flags);
	sti();
	udelay(500000);  	/* delay 0.5 seconds */
	restore_flags(flags);

	flg = 1;
	for(i=0;i<5;i++) flg &= (ar_read_regr(i+1) == expect[i]);

	if (ar_debug) {
		printk("arcd RESET: ");
		for (i=0;i<5;i++) printk("%3x",ar_read_regr(i+1));
		printk("\n");
	}
	
	ar_disconnect();
	return flg-1;	
}

static int ar_identify( char * id )

{	int k;
	char   id_cmd[12] = {0x12,0,0,0,36,0,0,0,0,0,0,0};

	ar_bufblk = -1;
        if (ar_atapi(id_cmd,36,"identify")) return -1;
	for (k=0;k<16;k++) id[k] = ar_buffer[16+k];
	id[16] = 0;
	return 0;
}

static void ar_probe ( int port )

{ 	if (check_region(port,3)) return;

	ar_base = port;
	ar_mode = 0;

	if (ar_test_proto()) {
		ar_base = 0;
		return;
	}

	ar_mode = 1;
	if (ar_test_proto()) ar_mode = 0;
	
}
 
static int ar_detect( void )

{	char    id[18];
	char    *mode_str[2] = { "4-bit", "8-bit" };

	if (!ar_base) ar_probe(0x3bc);
	if (!ar_base) ar_probe(0x378);
	if (!ar_base) ar_probe(0x278);

	if (!ar_base) {
		printk("arcd: autoprobe failed\n");
		return -1;
	}

	if (check_region(ar_base,3)) {
		printk("arcd: Ports at 0x%x are not available\n",ar_base);
		return -1;
	}

	if (ar_test_proto()) {
		printk("arcd: Can't access ParaCD at 0x%x in %s mode\n",
				ar_base,mode_str[ar_mode]);
		return -1;
	}

	if (ar_debug) printk("arcd CHECK: port = 0x%x mode = %d\n", 
				ar_base, ar_mode);

	if (ar_reset()) {
		printk("arcd: Failed to reset CD drive\n");
		return -1;
	}

	if (ar_identify(id)) {
		printk("arcd: ATAPI inquiry failed\n");
		return -1;
	}

	request_region(ar_base,3,"arcd");

	printk("arcd %s: Found %s, using port 0x%x in %s mode\n",
		AR_VERSION,id,ar_base,mode_str[ar_mode]);

	return 0;
}

static void ar_transfer( void )

{	int	k, o;

	while (ar_count && (ar_sector/4 == ar_bufblk)) {
		o = (ar_sector % 4) * 512;
		for(k=0;k<512;k++) ar_buf[k] = ar_buffer[o+k];
		ar_count--;
		ar_buf += 512;
		ar_sector++;
	}
}

static void do_ar_read( void )

{	int	b, i;
	char	rd_cmd[12] = {0xa8,0,0,0,0,0,0,0,0,1,0,0};

	ar_busy = 1;
	ar_transfer();
	if (!ar_count) {
		end_request(1);
		ar_busy = 0;
		return;
	}
	sti();

	ar_bufblk = ar_sector / 4;
        b = ar_bufblk;
	for(i=0;i<4;i++) { 
	   rd_cmd[5-i] = b & 0xff;
	   b = b >> 8;
	}

	if (ar_command(rd_cmd,2048,"read block")) {
		ar_bufblk = -1; 
		ar_busy = 0;
		cli();
		end_request(0);
		return;
	}
	ar_timeout = jiffies + AR_TMO;
	queue_task(&ar_tq,&tq_scheduler); 
}

static void ar_interrupt( void *data)

{	if (!(ar_ready())) {
		if (jiffies > ar_timeout) {
			ar_bufblk = -1;
			ar_busy = 0;
			end_request(0);
			do_ar_request();
		}
		queue_task(&ar_tq,&tq_scheduler); 
		return;
	}
	sti();
	if (ar_completion("read block")) {
		cli();
		ar_busy = 0;
		ar_bufblk = -1;
		end_request(0);
		do_ar_request();
	}
	do_ar_read();
	do_ar_request();
}
	
/* end of arcd.c */

