/* 
	dscd.c	(c) 1997  Grant R. Guenther <grant@torque.net>
		          Under the terms of the GNU public license.

	dscd.c is a driver for ATAPI CDroms based on the DataStor EP2000
        parallel port IDE adapter. 

        Ensure that /usr/include/linux and /usr/include/asm are 
	links to the correct include files for the target system.  Compile 
        the driver with

                cc -D__KERNEL__ -DMODULE -O2 -c dscd.c

        You must then load it with insmod.  If you are using
        MODVERSIONS, add the following to the cc command:

		-DMODVERSIONS -include /usr/include/linux/modversions.h

	Before attempting to access the new driver, you will need to
        create a new device special file.  The following commands will
	do that for you:

		mknod /dev/dscd b 63 0
		chown root:disk /dev/dscd
		chmod 660 /dev/dscd

	Load the driver with a command like:

		insmod dscd

	Afterward, you can mount a disk in the usual way:

		mount -t iso9660 /dev/dscd /cdrom

	(assuming you have made a directory /cdrom to use as a
        mount point).

	If you load the module with

       		insmod dscd ds_debug=1

	the driver will log verbose messages as it probes
        for the appropriate port and mode.  If you have problems 
        loading the driver, please include this output with your
        report.

*/

#define	DS_VERSION	"0.01" 

#define	DS_BASE		0	/* set to 0 for autoprobe */
#define DS_HOLD         4       /* micro-second port settling time */
#define DS_DEBUG	1	/* enable debug messages */

#include <linux/version.h>
#include <linux/module.h>

#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/fs.h>
#include <linux/major.h>
#include <linux/kernel.h>
#include <linux/tqueue.h>
#include <linux/delay.h>
#include <linux/cdrom.h>
#include <linux/ioport.h>

#include <asm/system.h>
#include <asm/io.h>
#include <asm/segment.h>   /* change to uaccess.h for 2.1 kernels ... */

#ifndef DSCD_MAJOR
#define DSCD_MAJOR   63	
#endif

#define MAJOR_NR DSCD_MAJOR

/* set up defines for blk.h,  why don't all drivers do it this way ? */

#define DEVICE_NAME "EP2000/CD"
#define DEVICE_REQUEST do_ds_request
#define DEVICE_NR(device) (MINOR(device))
#define DEVICE_ON(device)
#define DEVICE_OFF(device)

#include <linux/blk.h>

#define DS_TMO		   300		/* timeout in jiffies */
#define DS_DELAY            50          /* spin delay in uS */

#define DS_SPIN		(10000/DS_DELAY)*DS_TMO

int dscd_init(void);
void dscd_setup(char * str, int * ints);
void cleanup_module( void );

static int 	ds_open(struct inode *inode, struct file *file);
static void 	do_ds_request(void);
static void 	do_ds_read(void);
static int 	ds_ioctl(struct inode *inode,struct file *file,
                         unsigned int cmd, unsigned long arg);
static void 	ds_release (struct inode *inode, struct file *file);

static int 	ds_detect(void);
static int      ds_lock(void);
static void     ds_unlock(void);
static void     ds_eject(void);
static void     ds_interrupt(void *data);

static int	ds_base = DS_BASE;
static int      ds_mode = 0;
static int	ds_hold = DS_HOLD;
static int      ds_debug = DS_DEBUG;
static int      ds_found_0;
static int      ds_found_2;

static char ds_nmap[256];               /* nybble decoder */

static int ds_access = 0;		/* count of active opens ... */
static int ds_busy = 0;			/* request being processed ? */
static int ds_timeout;                  /* "interrupt" loop limiter */
static int ds_sector;			/* address of next requested sector */
static int ds_count;			/* number of blocks still to do */
static char * ds_buf;			/* buffer for request in progress */
static char ds_buffer[2048];		/* raw block buffer */
static int ds_bufblk = -1;		/* block in buffer, in CD units,
					   -1 for nothing there */

static struct tq_struct ds_tq = {0,0,ds_interrupt,NULL}; 

/* kernel glue structures */

static struct file_operations ds_fops = {
	NULL,			/* lseek - default */
	block_read,		/* read - general block-dev read */
	block_write,		/* write - general block-dev write */
	NULL,			/* readdir - bad */
	NULL,			/* select */
	ds_ioctl,		/* ioctl */
	NULL,			/* mmap */
	ds_open,		/* open */
	ds_release,		/* release */
	block_fsync,		/* fsync */
	NULL,			/* fasync */
	NULL,                   /* media change ? */
	NULL			/* revalidate new media */
};

int dscd_init (void)	/* preliminary initialisation */

{       if (ds_detect()) return -1;

	if (register_blkdev(MAJOR_NR,"dscd",&ds_fops)) {
		printk("dscd: unable to get major number %d\n",MAJOR_NR);
		return -1;
	}
	blk_dev[MAJOR_NR].request_fn = DEVICE_REQUEST;
	read_ahead[MAJOR_NR] = 8;	/* 8 sector (4kB) read ahead */

	return 0;
}

static int ds_open (struct inode *inode, struct file *file)

{	
	if (file->f_mode & 2) return -EROFS;  /* wants to write ? */

	MOD_INC_USE_COUNT;

	if (!ds_access)
	   if (ds_lock()) {
		MOD_DEC_USE_COUNT;
		return -ENXIO;
	   }
	ds_access++;
	return 0;
}

static void do_ds_request (void)

{       if (ds_busy) return;
        while (1) {
	    if ((!CURRENT) || (CURRENT->rq_status == RQ_INACTIVE)) return;
	    INIT_REQUEST;
	    if (CURRENT->cmd == READ) {
	        ds_sector = CURRENT->sector;
	        ds_count = CURRENT->nr_sectors;
	        ds_buf = CURRENT->buffer;
	        do_ds_read();
		if (ds_busy) return;
	    } 
	    else end_request(0);
	}
}

static int ds_ioctl(struct inode *inode,struct file *file,
		    unsigned int cmd, unsigned long arg)

/* we currently support only the EJECT ioctl. */

{	switch (cmd) {
            case CDROMEJECT: if (ds_access == 1) {
				ds_eject();
				return 0;
			     }
	    default:
	        return -EINVAL;
	}
}

static void ds_release (struct inode *inode, struct file *file)

{	kdev_t	devp;

	ds_access--;
	if (!ds_access) { 
	    devp = inode->i_rdev;
	    fsync_dev(devp);
	    invalidate_inodes(devp);
	    invalidate_buffers(devp);
	    ds_unlock();
	}
	MOD_DEC_USE_COUNT;
}

#ifdef MODULE

/* Glue for modules ... */

int	init_module(void)

{	int	err;
	long    flags;

	save_flags(flags);
	cli();

	err = dscd_init();

	restore_flags(flags);
	return err;
}

void	cleanup_module(void)

{	long flags;
	int range;

	save_flags(flags);
	cli();
	unregister_blkdev(MAJOR_NR,"dscd");
	if (ds_mode == 2) range = 8; else range = 3;
	release_region(ds_base,range);
	restore_flags(flags);
}

#else 

/* dscd_setup:  process lilo command parameters ...

   syntax:	dscd=base[,hold]
*/

void    dscd_setup(char *str, int *ints)

{       if (ints[0] > 0) ds_base = ints[1];
        if (ints[0] > 1) ds_mode = ints[2];
        if (ints[0] > 2) ds_hold = ints[3];
} 

#endif


/*------------------------------------------------------------------*/

/* mode codes:  0  nybble reads, 8-bit writes
                1  8-bit reads and writes
                2  8-bit EPP mode
*/


#define out_p(port,byte)  outb(byte,ds_base+port);udelay(ds_hold);
#define in_p(port)	  (udelay(ds_hold),(inb(ds_base+port)&0xff))

#define w0(byte)  out_p(0,byte)
#define w2(byte)  out_p(2,byte)
#define r0()      in_p(0)
#define r1()      in_p(1)
#define r2()      in_p(2) 
#define w3(byte)  out_p(3,byte)
#define w4(byte)  out_p(4,byte)
#define r4()      in_p(4)

#define j44(a,b)  (ds_nmap[a] + (ds_nmap[b] << 4))

#define P1      w2(5);w2(0xd);w2(5);w2(4);
#define P2      w2(5);w2(7);w2(5);w2(4);
#define P3      w2(6);w2(4);w2(6);w2(4);


/* To access the control registers add 0x80 to the regr number */

static int ds_read_regr( char regr )

{       int     a, b, r;

	if (regr & 0x80) r = (regr&7) + 0x40; else r = (regr&7) + 0x20;

	w0(0x81); P1;
	if (ds_mode) { w0(0x11); } else {w0(1); }  P2;
	w0(r); P1;

        switch (ds_mode)  {

        case 0: w2(6); a = r1(); w2(4); w2(6); b = r1(); w2(4);
                return j44(a,b);

        case 1: w0(0); w2(0x26); a = r0(); w2(4);
                return a;

        case 2: w2(0x24); a = r4(); w2(4);
                return a;

        }
        return -1;
}       

static inline int ds_ready( void )

{	w0(0x81); P1; 
        if (ds_mode) { w0(0x11); } else { w0(1); } P2;
        w0(0); P1;
	return (r1() & 0x40);
}


static void ds_write_regr( char regr, char val )

{       int  r;

        if (regr & 0x80) r = (regr&7) + 0x40; else r = (regr&7) + 0x20;

	w0(0x81); P1; 
	if (ds_mode == 2) { w0(0x11); } else { w0(1); } P2;
	w0(r); P1;
	
        switch (ds_mode)  {

        case 0:
        case 1: w0(val); w2(5); w2(7); w2(5); w2(4);
		break;

        case 2: w4(val); 
                break;
        }
}

#define  CCP(x)  w0(0xff);w2(0xc);w2(4);\
		 w0(0xaa);w0(0x55);w0(0);w0(0xff);w0(0x87);w0(0x78);\
		 w0(x);w2(5);w2(4);

static void ds_connect ( void  )

{       ds_found_0 = r0();
        ds_found_2 = r2();
        w2(4); CCP(0xe0); w0(0xff);
}

static void ds_disconnect ( void )

{       CCP(0x30);
        w0(ds_found_0);
        w2(ds_found_2);
} 

static void ds_read_data( char * buf, int len )

{       int     k, a, b;

        w0(0x81); P1;
        if (ds_mode) { w0(0x19); } else { w0(9); } P2;
        w0(0x82); P1; P3;
	w0(0x20); P1;

        switch (ds_mode) {

        case 0: for (k=0;k<len;k++) {
                        w2(6); a = r1(); w2(4);
                        w2(6); b = r1(); w2(4);
                        buf[k] = j44(a,b);
                } 
                break;

        case 1: w0(0);
                for (k=0;k<len;k++) {
                        w2(0x26); buf[k] = r0(); w2(0x24);
                }
                w2(4);
                break;

        case 2: w2(0x24); 
                for (k=0;k<len;k++) buf[k] = r4();
                w2(4);
                break;

        }
}

static void ds_write_cmd( char * cmd, int len )

{       int	k;

        w0(0x81); P1;
        if (ds_mode == 2) { w0(0x19); } else { w0(9); } P2;
        w0(0x82); P1; P3;
        w0(0x20); P1;


        switch (ds_mode) {

        case 0:
        case 1: for (k=0;k<len;k++) {
                        w2(5); w0(cmd[k]); w2(7);
                }
                w2(5); w2(4);
                break;

        case 2: w2(0xc5);
                for (k=0;k<len;k++) w4(cmd[k]);
		w2(0xc4);
                break;

        }

}

static void ds_init_nmap( void )

{	int i;

	for (i=0;i<128;i++) {
		ds_nmap[i] = ((i >> 3) & 0x07) | 8;
		ds_nmap[i+128] = ((i >> 3) & 0x07);
	}
}


static int ds_test_proto( void )

{       int	k, e;

	ds_init_nmap();   /* as good a place as any to do this ... */

        e = 0;
        ds_connect();
        for (k=0;k<256;k++) {
                ds_write_regr(2,k^0xaa);
                ds_write_regr(3,k^0x55);
                if (ds_read_regr(2) != (k^0xaa)) e++;
                }
        ds_disconnect();

	if (ds_debug)  {
            printk("ds: port 0x%x, mode %d, register test %d errors\n",
                    ds_base,ds_mode,e);

            ds_connect();
            ds_write_regr(6,0xa0);  udelay(30);
            if (ds_read_regr(0x86) != ds_read_regr(7)) 
               printk("ds: oops, ALT_STATUS != STATUS\n");
            ds_disconnect();

        }

        return e;
}

/*-------------------------------------------------------------*/



static int ds_wait_drq( char * fun, char * msg1, char * msg2 )

{	int j, r, e;

	j = 0;
	while (!(ds_read_regr(7)&8)&&(j++<DS_SPIN)) udelay(DS_DELAY);
        r = ds_read_regr(7);
	e = ds_read_regr(1);
	if (j >= DS_SPIN) {
	    printk("dscd: %s%s: DRQ timeout on %s, status: %x error: %x\n",
			     fun,msg1,msg2,r,e);
	    return -1;
	}
	return 0;
}

static int ds_wait( char * fun, char * msg1, char * msg2 )

{	int j, r, e;

	j = 0;
	while ((!(ds_ready()))&&(j++<DS_SPIN)) udelay(DS_DELAY);
        r = ds_read_regr(7);
	e = ds_read_regr(1);
	if ((j >= DS_SPIN) || (r & 1)) {
	    if (j >= DS_SPIN) e |= 0x100;
	    printk("dscd: %s%s: Error %s, status: %x error: %x\n",
			  fun,msg1,msg2,r,e);
	    return -1;
	}
	return 0;
}

static int ds_req_sense( char * fun, char * msg )

{	char	rs_cmd[12] = { 0x03,0,0,0,18,0,0,0,0,0,0,0 };
	int n, r;

        ds_write_regr(4,18);
        ds_write_regr(5,0);
        ds_write_regr(6,0xa0);          /* drive 0 */
        ds_write_regr(7,0xa0);          /* ATAPI packet command */
	if (ds_wait_drq(fun,msg,"request sense command")) return -1;
	ds_write_cmd(rs_cmd,12);
	if (!(r=ds_wait(fun,msg,"getting request sense data"))) {
            if (ds_read_regr(2) == 2) {
                n = (ds_read_regr(4) + 256*ds_read_regr(5));
                ds_read_data(ds_buffer,n);
                r=ds_wait(fun,msg,"on request sense done");
                if (!r)
	          printk("dscd: %s%s: sense key: %x, ASC: %x, ASQ: %x\n",
	            fun, msg, ds_buffer[2]&0xf, ds_buffer[12], ds_buffer[13]);
	    }
	}
	return r;
}

static int ds_command( char * cmd, int dlen, char * fun )

{	int	r;

	ds_connect();

	ds_write_regr(4,dlen % 256);
	ds_write_regr(5,dlen / 256);
	ds_write_regr(6,0xa0);		/* drive 0 */
	ds_write_regr(7,0xa0);		/* ATAPI packet command */

        if (ds_read_regr(7)&1) {
            if (ds_req_sense(fun," before command")) { 
                ds_disconnect();
                return -1;
                }
            ds_write_regr(4,dlen % 256);
            ds_write_regr(5,dlen / 256);
            ds_write_regr(6,0xa0);          /* drive 0 */
            ds_write_regr(7,0xa0);          /* ATAPI packet command */
	}

	if ((r=ds_wait_drq(fun,"","sending  command"))) {
	   ds_disconnect();
	   return r;
	}

	ds_write_cmd(cmd,12);
	return 0;
}

static int ds_completion( char * fun )

{	int r, n;

	if (!(r=ds_wait(fun,"","reading data"))) {
	    if (ds_read_regr(2) == 2) {
	        n = (ds_read_regr(4) + 256*ds_read_regr(5));
	        ds_read_data(ds_buffer,n);
	        r=ds_wait(fun,"","waiting after data");
	    }
	}
	if (r) r = ds_req_sense(fun," after command");

	ds_disconnect(); 
	return r;
}

static int ds_atapi( char * cmd, int dlen, char * fun )

{	int r;

	if (!(r=ds_command(cmd,dlen,fun)))
	  r = ds_completion(fun);
	return r;
}

static int ds_lock(void)

{	char	lo_cmd[12] = { 0x1e,0,0,0,1,0,0,0,0,0,0,0 };
	char	cl_cmd[12] = { 0x1b,0,0,0,3,0,0,0,0,0,0,0 };

	ds_atapi(cl_cmd,0,"close door");
	return ds_atapi(lo_cmd,0,"lock door");
}

static void ds_unlock( void)

{	char	un_cmd[12] = { 0x1e,0,0,0,0,0,0,0,0,0,0,0 };

	ds_atapi(un_cmd,0,"unlock door");
}

static void ds_eject( void)

{	char	ej_cmd[12] = { 0x1b,0,0,0,2,0,0,0,0,0,0,0 };

	ds_unlock();
	ds_atapi(ej_cmd,0,"eject");
}

static int ds_reset( void )

/* the ATAPI standard actually specifies the contents of all 7 registers
   after a reset, but the specification is ambiguous concerning the last
   two bytes, and different drives interpret the standard differently.
*/

{	int	i, flg;
	int	expect[5] = {1,1,1,0x14,0xeb};
	long	flags;

	ds_connect();
	ds_write_regr(6,0xa0);
	ds_write_regr(7,8);

	save_flags(flags);
	sti();
	udelay(500000);  	/* delay 0.5 seconds */
	restore_flags(flags);

	flg = 1;
	for(i=0;i<5;i++) flg &= (ds_read_regr(i+1) == expect[i]);

	if (ds_debug) {
		printk("dscd RESET: ");
		for (i=0;i<5;i++) printk("%3x",ds_read_regr(i+1));
		printk("\n");
	}
	
	ds_disconnect();
	return flg-1;	
}

static int ds_identify( char * id )

{	int k;
	char   id_cmd[12] = {0x12,0,0,0,36,0,0,0,0,0,0,0};

	ds_bufblk = -1;
        if (ds_atapi(id_cmd,36,"identify")) return -1;
	for (k=0;k<16;k++) id[k] = ds_buffer[16+k];
	id[16] = 0;
	return 0;
}


static void ds_probe( int port, int max_mode)

{       int  best, range;

        range = 3;
        if (max_mode == 2) range = 8;
        if (check_region(port,range)) {
                max_mode = 2;
                range = 3;
                if (check_region(port,range)) return;
        }

        ds_base = port;  best = -1;

        for (ds_mode=0;ds_mode<=max_mode;ds_mode++)
                if (!ds_test_proto())  best = ds_mode;

        ds_mode = best;
        if (ds_mode < 0) ds_base = 0; 
}

 
static int ds_detect( void )

{	char    id[18];
	char    *mode_str[3] = { "4-bit", "8-bit", "EPP" };
	int	range;

	if (!ds_base) ds_probe(0x3bc,1);
	if (!ds_base) ds_probe(0x378,2);
	if (!ds_base) ds_probe(0x278,2);

	if (!ds_base) {
		printk("dscd: autoprobe failed\n");
		return -1;
	}

	range = 3; if (ds_mode == 2) range = 8;
	if (check_region(ds_base,range)) {
		printk("dscd: Ports at 0x%x are not available\n",ds_base);
		return -1;
	}

	if (ds_test_proto()) {
		printk("dscd: No EP2000 found at 0x%x in %s mode\n",
				ds_base,mode_str[ds_mode]);
		return -1;
	}

	if (ds_debug) printk("dscd CHECK: port = 0x%x mode = %d\n", 
				ds_base, ds_mode);

	if (ds_reset()) {
		printk("dscd: Failed to reset CD drive\n");
		return -1;
	}

	if (ds_identify(id)) {
		printk("dscd: ATAPI inquiry failed\n");
		return -1;
	}

	request_region(ds_base,range,"dscd");

	printk("dscd %s: Found %s, using port 0x%x in %s mode\n",
		DS_VERSION,id,ds_base,mode_str[ds_mode]);

	return 0;
}

static void ds_transfer( void )

{	int	k, o;

	while (ds_count && (ds_sector/4 == ds_bufblk)) {
		o = (ds_sector % 4) * 512;
		for(k=0;k<512;k++) ds_buf[k] = ds_buffer[o+k];
		ds_count--;
		ds_buf += 512;
		ds_sector++;
	}
}

static void do_ds_read( void )

{	int	b, i;
	char	rd_cmd[12] = {0xa8,0,0,0,0,0,0,0,0,1,0,0};

	ds_busy = 1;
	ds_transfer();
	if (!ds_count) {
		end_request(1);
		ds_busy = 0;
		return;
	}
	sti();

	ds_bufblk = ds_sector / 4;
        b = ds_bufblk;
	for(i=0;i<4;i++) { 
	   rd_cmd[5-i] = b & 0xff;
	   b = b >> 8;
	}

	if (ds_command(rd_cmd,2048,"read block")) {
		ds_bufblk = -1; 
		ds_busy = 0;
		cli();
		end_request(0);
		return;
	}
	ds_timeout = jiffies + DS_TMO;
	queue_task(&ds_tq,&tq_scheduler); 
}

static void ds_interrupt( void *data)

{	if (!(ds_ready())) {
		if (jiffies > ds_timeout) {
			ds_bufblk = -1;
			ds_busy = 0;
			end_request(0);
			do_ds_request();
		}
		queue_task(&ds_tq,&tq_scheduler); 
		return;
	}
	sti();
	if (ds_completion("read block")) {
		cli();
		ds_busy = 0;
		ds_bufblk = -1;
		end_request(0);
		do_ds_request();
	}
	do_ds_read();
	do_ds_request();
}
	
/* end of dscd.c */

