/* 
	eicd.c	(c) 1996,1997  Grant R. Guenther <grant@torque.net>
		               Under the terms of the GNU public license.

	eicd.c is a driver for parallel port CDrom drives based on the
        EPIA parallel to IDE adapter chip from Shuttle Technologies.
	There is also an ei driver for IDE disks using this interface.

        To compile the driver, ensure that /usr/include/linux and
        /usr/include/asm are links to the correct include files for 
        the target system. Then compile the driver with 

                cc -D__KERNEL__ -DMODULE -Wall -O2 -c eicd.c

        Alternatively, if you are using MODVERSIONS, do

                cc -D__KERNEL__ -DMODULE -DMODVERSIONS \
                   -include /usr/src/linux/include/linux/modversions.h \
                   -Wall -O2 -c eicd.c

        You must then load it with insmod.

	Before attempting to access the new driver, you will need to
        create a new device special file.  The following commands will
	do that for you:

		mknod /dev/eicd b 63 0
		chown root:disk /dev/eicd
		chmod 660 /dev/eicd

	Afterward, you can mount a disk in the usual way:

		mount -t iso9660 /dev/eicd /cdrom

	(assuming you have made a directory /cdrom to use as a
        mount point).

        (The number '63' above corresponds to the #define EI_MAJOR 63
        in the source below.  If it is necessary to change it, you
        must be sure that all uses of the device number are consistent.)

        The driver will search for an EPIA device on the available
        standard parallel ports.  The driver supports unidirectional
        and bidirectional parallel ports as well as some EPP ports.

        It is not currently possible to connect a printer to a chained
        printer port and expect Linux to use both devices at once.

        Keep an eye on http://www.torque.net/epia.html for news and
        other information about the driver.  If you have any problems
        with this driver, please send me, grant@torque.net, some mail 
        directly before posting into the newsgroups or mailing lists.

*/

#define	EI_VERSION	"0.01" 

#define	EI_BASE		0	/* set to 0 for autoprobe */
#define EI_HOLD         1       /* micro-second port settling time */
#define EI_DEBUG	0

#include <linux/version.h>
#include <linux/module.h>

#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/fs.h>
#include <linux/major.h>
#include <linux/kernel.h>
#include <linux/tqueue.h>
#include <linux/delay.h>
#include <linux/cdrom.h>
#include <linux/ioport.h>

#include <asm/system.h>
#include <asm/io.h>
#include <asm/segment.h>   /* change to uaccess.h for 2.1 kernels ... */

#ifndef EICD_MAJOR
#define EICD_MAJOR   63
#endif

#define MAJOR_NR EICD_MAJOR

/* set up defines for blk.h,  why don't all drivers do it this way ? */

#define DEVICE_NAME "EPIA CDrom"
#define DEVICE_REQUEST do_ei_request
#define DEVICE_NR(device) (MINOR(device))
#define DEVICE_ON(device)
#define DEVICE_OFF(device)

#include <linux/blk.h>

#define EI_RETRIES	     5
#define EI_TMO		   600		/* timeout in jiffies */
#define EI_DELAY            50          /* spin delay in uS */

#define EI_SPIN		(10000/EI_DELAY)*EI_TMO

int eicd_init(void);
void eicd_setup(char * str, int * ints);
void cleanup_module( void );

static int 	ei_open(struct inode *inode, struct file *file);
static void 	do_ei_request(void);
static void 	do_ei_read(void);
static int 	ei_ioctl(struct inode *inode,struct file *file,
                         unsigned int cmd, unsigned long arg);
static void 	ei_release (struct inode *inode, struct file *file);

static int 	ei_detect(void);
static void     ei_lock(void);
static void     ei_unlock(void);
static void     ei_eject(void);
static void     ei_interrupt(void *data);

static int	ei_base = EI_BASE;
static int      ei_mode = 0;
static int	ei_hold = EI_HOLD;
static int      ei_debug = EI_DEBUG;;

static int      ei_retries;

static int ei_access = 0;		/* count of active opens ... */
static int ei_busy = 0;			/* request being processed ? */
static int ei_timeout;                  /* "interrupt" loop limiter */
static int ei_sector;			/* address of next requested sector */
static int ei_count;			/* number of blocks still to do */
static char * ei_buf;			/* buffer for request in progress */
static char ei_buffer[2048];		/* raw block buffer */
static int ei_bufblk = -1;		/* block in buffer, in CD units,
					   -1 for nothing there */
static int ei_found_0;
static int ei_found_2;

static struct tq_struct ei_tq = {0,0,ei_interrupt,NULL}; 

/* kernel glue structures */

static struct file_operations ei_fops = {
	NULL,			/* lseek - default */
	block_read,		/* read - general block-dev read */
	block_write,		/* write - general block-dev write */
	NULL,			/* readdir - bad */
	NULL,			/* select */
	ei_ioctl,		/* ioctl */
	NULL,			/* mmap */
	ei_open,		/* open */
	ei_release,		/* release */
	block_fsync,		/* fsync */
	NULL,			/* fasync */
	NULL,                   /* media change ? */
	NULL			/* revalidate new media */
};

int eicd_init (void)	/* preliminary initialisation */

{       if (ei_detect()) return -1;

	if (register_blkdev(MAJOR_NR,"eicd",&ei_fops)) {
		printk("eicd: unable to get major number %d\n",MAJOR_NR);
		return -1;
	}
	blk_dev[MAJOR_NR].request_fn = DEVICE_REQUEST;
	read_ahead[MAJOR_NR] = 8;	/* 8 sector (4kB) read ahead */

	return 0;
}

static int ei_open (struct inode *inode, struct file *file)

{	
	if (file->f_mode & 2) return -EROFS;  /* wants to write ? */

	MOD_INC_USE_COUNT;

	ei_lock();

	ei_access++;
	return 0;
}

static void do_ei_request (void)

{       if (ei_busy) return;
        while (1) {
	    if ((!CURRENT) || (CURRENT->rq_status == RQ_INACTIVE)) return;
	    INIT_REQUEST;
	    if (CURRENT->cmd == READ) {
	        ei_sector = CURRENT->sector;
	        ei_count = CURRENT->nr_sectors;
	        ei_buf = CURRENT->buffer;
	        do_ei_read();
		if (ei_busy) return;
	    } 
	    else end_request(0);
	}
}

static int ei_ioctl(struct inode *inode,struct file *file,
		    unsigned int cmd, unsigned long arg)

/* we currently support only the EJECT ioctl. */

{	switch (cmd) {
            case CDROMEJECT: if (ei_access == 1) {
				ei_eject();
				return 0;
			     }
	    default:
	        return -EINVAL;
	}
}

static void ei_release (struct inode *inode, struct file *file)

{	kdev_t	devp;

	ei_access--;
	if (!ei_access) { 
	    devp = inode->i_rdev;
	    fsync_dev(devp);
	    invalidate_inodes(devp);
	    invalidate_buffers(devp);
	    ei_unlock();
	}
	MOD_DEC_USE_COUNT;
}

#ifdef MODULE

/* Glue for modules ... */

int	init_module(void)

{	int	err;
	long    flags;

	save_flags(flags);
	cli();

	err = eicd_init();

	restore_flags(flags);
	return err;
}

void	cleanup_module(void)

{	long flags;
	int  range;

	if (ei_mode == 3) range = 8; else range = 3;

	save_flags(flags);
	cli();
	unregister_blkdev(MAJOR_NR,"eicd");
	release_region(ei_base,range);
	restore_flags(flags);
}

#else 

/* eicd_setup:  process lilo command parameters ...

   syntax:	eicd=base[,mode[,hold]]
*/

void    eicd_setup(char *str, int *ints)

{       if (ints[0] > 0) ei_base = ints[1];
        if (ints[0] > 1) ei_mode = ints[2];
        if (ints[0] > 2) ei_hold = ints[3];
} 

#endif

/*------------------------------------------------------------------*/

/* mode codes:  0  nybble reads on port 1, 8-bit writes
                1  5/3 reads on ports 1 & 2, 8-bit writes
                2  8-bit reads and writes
                3  8-bit EPP mode
*/

/* Now the actual hardware interface */

#define out_p(port,byte)   outb(byte,ei_base+port); udelay(ei_hold);
#define in_p(port)         (udelay(ei_hold),inb(ei_base+port))

#define w0(byte)                out_p(0,byte)
#define r0()                    (in_p(0) & 0xff)
#define w1(byte)                out_p(1,byte)
#define r1()                    (in_p(1) & 0xff)
#define w2(byte)                out_p(2,byte)
#define r2()                    (in_p(2) & 0xff)
#define w3(byte)                out_p(3,byte)
#define w4(byte)                out_p(4,byte)
#define r4()                    (in_p(4) & 0xff)

#define j44(a,b)                (((a>>4)&0x0f)+(b&0xf0))
#define j53(a,b)                (((a>>3)&0x1f)+((b<<4)&0xe0))


/* To access the control registers add 0x80 to the regr number */

static int ei_read_regr( char regr )

{       int     a, b, r;

        switch (ei_mode)  {

        case 0: r = regr^0x39;
                w0(r); w2(1); w2(3); w0(r);
                a = r1(); w2(1); b = r1(); w2(4);
                return j44(a,b);

        case 1: r = regr^0x31;
                w0(r); w2(1); w0(r&0x37); 
                w2(3); w2(5); w0(r|0xf0);
                a = r1(); b = r2(); w2(4);
                return j53(a,b);

        case 2: r = regr^0x29;
                w0(r); w2(1); w2(0X21); w2(0x23); 
                a = r0(); w2(4);
                return a;

        case 3: w3(regr); w2(0x24); a = r4(); w2(4);
                return a;

        }
        return -1;
}    

static void ei_write_regr( char regr, char val )

{       int  r;

        switch (ei_mode)  {

        case 0:
        case 1:
        case 2: r = regr^0x19;
                w0(r); w2(1); w0(val); w2(3); w2(4);
                break;

        case 3: r = regr^0x40;
                w3(r); w4(val); w2(4);
                break;
        }
}

static void ei_write_cmd( char * buf, int len )

{       int     ph, k, last, d;

        switch (ei_mode) {

        case 0:
        case 1:
        case 2: w0(0xa1); w2(1); w2(3); w2(1); w2(5);
                ph = 0;  last = 0x8000;
                for (k=0;k<len;k++) {
                        d = buf[k];
                        if (d != last) { last = d; w0(d); }
                        w2(4+ph);
                        ph = 1 - ph;
                }
                w2(7); w2(4);
                break;

        case 3: w3(0xc4); w4(1); w3(0x40);
                for (k=0;k<len;k++) w4(buf[k]);
		w3(0xc4); w4(0);
                break;

        }

}

static void ei_read_data( char * buf, int len )

{       int     k, ph, a, b;

        switch (ei_mode) {

        case 0: w0(0x81); w2(1); w2(3); w0(0xc1);
                ph = 1;
                for (k=0;k<len;k++) {
                        w2(2+ph); a = r1();
                        w2(4+ph); b = r1();
                        buf[k] = j44(a,b);
                        ph = 1 - ph;
                } 
                w0(0); w2(4);
                break;

        case 1: w0(0x91); w2(1); w0(0x10); w2(3); 
                w0(0x51); w2(5); w0(0xd1); 
                ph = 1;
                for (k=0;k<len;k++) {
                        w2(4+ph);
                        a = r1(); b = r2();
                        buf[k] = j53(a,b);
                        ph = 1 - ph;
                }
                w0(0); w2(4);
                break;

        case 2: w0(0x89); w2(1); w2(0x23); w2(0x21); 
                ph = 1;
                for (k=0;k<len;k++) {
                        w2(0x24+ph);
                        buf[k] = r0();
                        ph = 1 - ph;
                }
                w2(6); w2(4);
                break;

        case 3: if (len > 512) { w3(0xc4); w4((len-1)/512); }
		w3(0); w2(0x24);
                for (k=0;k<len;k++) buf[k] = r4();
                w2(4);
		w3(0xc4); w4(0);
                break;

        }
}


static void ei_connect ( void  )

{       ei_found_0 = r0();
        ei_found_2 = r2();
        w2(4); w0(0xa0); w0(0x50); w0(0xc0); w0(0x30); w0(0xa0); w0(0);
        w2(1); w2(4);
        if (ei_mode == 3) { 
                w0(0xa); w2(1); w2(4); w0(0x82); w2(4); w2(0xc); w2(4);
                w2(0x24); w2(0x26); w2(4);
        }
        ei_write_regr(4,0x55);  
        ei_write_regr(5,0xaa); 
        ei_read_regr(4);
        ei_write_regr(0x86,8);  
}

static void ei_disconnect ( void )

{       ei_write_regr(0x84,0x10);
        w0(ei_found_0);
        w2(1); w2(4);
        w0(ei_found_0);
        w2(ei_found_2);
}

static int ei_ready( void )

{	int s;

	w2(4);
	if (ei_mode == 3) return (!(r1() & 8));
	w0(0x90); w2(1); w2(1); w2(4); w2(4);
	s = r1();
	w0(0); w2(1); w2(1); w2(4); w2(4);
	return (!(s & 8));
}

static int ei_test_proto( void )

{       int     tp[8] = {0xff,0x55,0xaa,0x0,0xf0,0x0f,0x5a,0xa5};
        int     k, e, f;

        e = 0;
        ei_connect();

        for (k=0;k<8;k++) {
                ei_write_regr(2,tp[k]);
                if (ei_read_regr(2) != tp[k]) e++;
                }
        ei_disconnect();

        f = 0;
        ei_connect();
        ei_write_regr(0x84,8);
        ei_read_data(ei_buffer,512);
        for (k=0;k<256;k++) {
            if ((ei_buffer[2*k] & 0xff) != ((k+1) & 0xff)) f++;
            if ((ei_buffer[2*k+1] & 0xff) != ((-2-k) & 0xff)) f++;
        }
        ei_write_regr(0x84,0);
        ei_disconnect();

        if (ei_debug)
            printk("ei: port 0x%x, mode %d, %d + %d errors\n",
                   ei_base,ei_mode,e,f);

        return e+f;
}

/*-------------------------------------------------------------------*/

static int ei_wait_drq( char * fun, char * msg1, char * msg2 )

{       int j, r, e;

        j = 0;
        while (!(ei_read_regr(7)&8)&&(j++<EI_SPIN)) udelay(EI_DELAY);
        r = ei_read_regr(7);
        e = ei_read_regr(1);
        if (j >= EI_SPIN) {
            if (fun) 
              printk("eicd: %s%s: DRQ timeout on %s, status: %x error: %x\n",
                             fun,msg1,msg2,r,e);
            return -1;
        }
        return 0;
}

static int ei_wait_cmd( )

{       int j, r, e;

        j = 0;
        while ((ei_read_regr(7)&0x88)&&(j++<EI_SPIN)) udelay(EI_DELAY);
        r = ei_read_regr(7);
        e = ei_read_regr(1);
        if (j >= EI_SPIN) {
            printk("eicd: Pre-command timeout:  status: %x error: %x\n",
                             r,e);
            return -1;
        }
        return 0;
}

static int ei_wait( char * fun, char * msg1, char * msg2 )

{       int j, r, e;

        j = 0;
        while ((!(ei_ready()))&&(j++<EI_SPIN)) udelay(EI_DELAY);
        r = ei_read_regr(7);
        e = ei_read_regr(1);
        if ((j >= EI_SPIN) || (r & 1)) {
            if (j >= EI_SPIN) e |= 0x100;
            if (fun) printk("eicd: %s%s: Error %s, status: %x error: %x\n",
                          fun,msg1,msg2,r,e);
            return -1;
        }
        return 0;
}

static int ei_req_sense( char * fun, char * msg )

{	char	rs_cmd[12] = { 0x03,0,0,0,18,0,0,0,0,0,0,0 };
	int n, r, p;

        if (ei_wait_cmd()) return -1;

        ei_write_regr(4,18);
        ei_write_regr(5,0);
        ei_write_regr(6,0xa0);          /* drive 0 */
        ei_write_regr(7,0xa0);          /* ATAPI packet command */
	if (ei_wait_drq(fun,msg,"request sense command")) return -1;
	ei_write_cmd(rs_cmd,12);
	if (!(r=ei_wait(fun,msg,"getting request sense data"))) {
	    p = ei_read_regr(2);
            if (p == 2) {
                n = (ei_read_regr(4) + 256*ei_read_regr(5));
		ei_wait_drq(fun,msg,"getting sense data");
                ei_read_data(ei_buffer,n);
                r=ei_wait(fun,msg,"on request sense done");
                if (!r) if (fun)
	          printk("eicd: %s%s: sense key: %x, ASC: %x, ASQ: %x\n",
	            fun, msg, ei_buffer[2]&0xf, ei_buffer[12], ei_buffer[13]);
	    } else if (p == 3) {
                r = (ei_read_regr(7) & 1);
                if (r) printk("eicd: phase error, status %x, error %x\n",
                                r,ei_read_regr(1));
	    }
	}
	return r;
}

static int ei_command( char * cmd, int dlen, char * fun )

{	int	r;

	ei_connect();

        if (ei_wait_cmd()) { ei_disconnect(); return -1; }

	ei_write_regr(4,dlen % 256);
	ei_write_regr(5,dlen / 256);
	ei_write_regr(6,0xa0);		/* drive 0 */
	ei_write_regr(7,0xa0);		/* ATAPI packet command */

        if (ei_read_regr(7)&1) {
            if (ei_req_sense(fun," before command")) { 
                ei_disconnect();
                return -1;
                }

            if (ei_wait_cmd()) { ei_disconnect(); return -1; }

            ei_write_regr(4,dlen % 256);
            ei_write_regr(5,dlen / 256);
            ei_write_regr(6,0xa0);          /* drive 0 */
            ei_write_regr(7,0xa0);          /* ATAPI packet command */
	}

	if ((r=ei_wait_drq(fun,"","sending command"))) {
	   ei_disconnect();
	   return r;
	}
	ei_write_cmd(cmd,12);
	return 0;
}

static int ei_completion( char * fun )

{	int r, n, p;

	if (!(r=ei_wait(fun,"","command completion"))) {
	    p = ei_read_regr(2);
	    if (p == 2) {
	        n = (ei_read_regr(4) + 256*ei_read_regr(5));
		ei_wait_drq(fun,"","getting data");
	        ei_read_data(ei_buffer,n);
	        r=ei_wait(fun,"","waiting after data");
	    } else if (p == 3) {
		r = (ei_read_regr(7) & 1);
		if (r) printk("eicd: phase error, status %x, error %x\n",
				r,ei_read_regr(1));
	    }
	}
	if (r) r = ei_req_sense(fun," after command");
	ei_disconnect(); 
	return r;
}

static int ei_atapi( char * cmd, int dlen, char * fun )

{	int r;

	if (!(r=ei_command(cmd,dlen,fun)))
	  r = ei_completion(fun);
	return r;
}

static void ei_lock(void)

{	char	lo_cmd[12] = { 0x1e,0,0,0,1,0,0,0,0,0,0,0 };
	char	cl_cmd[12] = { 0x1b,0,0,0,3,0,0,0,0,0,0,0 };

	ei_atapi(cl_cmd,0,NULL);
	ei_atapi(cl_cmd,0,NULL);
	ei_atapi(cl_cmd,0,NULL);
	ei_atapi(cl_cmd,0,NULL);
	ei_atapi(cl_cmd,0,"close door");
        ei_atapi(lo_cmd,0,NULL);
        ei_atapi(lo_cmd,0,"lock door");
}

static void ei_unlock( void)

{	char	un_cmd[12] = { 0x1e,0,0,0,0,0,0,0,0,0,0,0 };

	ei_atapi(un_cmd,0,"unlock door");
}

static void ei_eject( void)

{	char	ej_cmd[12] = { 0x1b,0,0,0,2,0,0,0,0,0,0,0 };

	ei_unlock();
	ei_atapi(ej_cmd,0,"eject");
}

static int ei_reset( void )

/* the ATAPI standard actually specifies the contents of all 7 registers
   after a reset, but the specification is ambiguous concerning the last
   two bytes, and different drives interpret the standard differently.
*/

{	int	i, flg;
	int	expect[5] = {1,1,1,0x14,0xeb};
	long	flags;

/* first do an ATA reset */

	ei_connect();
        ei_write_regr(0x86,4);
        udelay(50);
        ei_write_regr(0x86,0);
        ei_disconnect();
	udelay(1000);

/* now do an ATAPI reset */

	ei_connect();
	ei_write_regr(6,0xa0);
	ei_write_regr(7,8);

	save_flags(flags);
	sti();
	udelay(500000);  	/* delay 0.5 seconds */
	restore_flags(flags);

	flg = 1;
	for(i=0;i<5;i++) flg &= (ei_read_regr(i+1) == expect[i]);

	if (ei_debug) {
		printk("eicd RESET: ");
		for (i=0;i<7;i++) printk("%3x",ei_read_regr(i+1));
		printk("\n");
	}
	
	ei_disconnect();
	return flg-1;	
}

static int ei_identify( char * id )

{	int k;
	char   id_cmd[12] = {0x12,0,0,0,36,0,0,0,0,0,0,0};

	ei_bufblk = -1;
        if (ei_atapi(id_cmd,36,"identify")) return -1;
	for (k=0;k<16;k++) id[k] = ei_buffer[16+k];
	id[16] = 0;
	return 0;
}

static void ei_probe ( int port, int max_mode )

{	int	best, range;

	range = 3;
	if (max_mode == 3) range = 8;
	if (check_region(port,range)) {
		max_mode = 1;
		range = 3;
		if (check_region(port,range)) return;
	}

	ei_base = port;  best = -1;

	if (ei_debug) printk("eicd PROBE: port 0x%x, max_mode %d\n",
			port, max_mode);

	for (ei_mode=0;ei_mode<=max_mode;ei_mode++)
	   if (!ei_test_proto()) best = ei_mode;

	ei_mode = best;
	if (ei_mode < 0) ei_base = 0;
}
 
static int ei_detect( void )

{	char    id[18];
	int	range;
        char    *mode_str[4] = {"4-bit","Shuttle 5/3","8-bit","EPP"};

	if (!ei_base) ei_probe(0x278,3);
	if (!ei_base) ei_probe(0x378,3);
	if (!ei_base) ei_probe(0x3bc,2);

	if (!ei_base) {
		printk("eicd: Autoprobe failed\n");
		return -1;
	}

	if (ei_mode == 3) range = 8; else range = 3;
	if (check_region(ei_base,range)) {
		printk("eicd: Ports at 0x%x are not available\n",ei_base);
		return -1;
	}

	if (ei_test_proto()) {
		printk("eicd: Can't access EPIA adapter in %s mode\n",
			mode_str[ei_mode]);
		return -1;
	}

        printk("eicd %s: Shuttle EPIA adapter "
	       "using port 0x%x in %s mode\n",
                EI_VERSION,ei_base,mode_str[ei_mode]);

	if (ei_reset()) {
		printk("eicd: Failed to reset CD drive\n");
		return -1;
	}

	ei_identify(id);
	if (ei_identify(id)) {
		printk("eicd: ATAPI inquiry failed\n");
		return -1;
	}

	if (ei_mode == 3) range = 8; else range = 3;	
	request_region(ei_base,range,"eicd");

	printk("eicd: Found %s\n",id);

	return 0;
}

static void ei_transfer( void )

{	int	k, o;

	while (ei_count && (ei_sector/4 == ei_bufblk)) {
		o = (ei_sector % 4) * 512;
		for(k=0;k<512;k++) ei_buf[k] = ei_buffer[o+k];
		ei_count--;
		ei_buf += 512;
		ei_sector++;
	}
}

static void ei_start( void )

{	int	b, i;
	char	rd_cmd[12] = {0xa8,0,0,0,0,0,0,0,0,1,0,0};

	ei_bufblk = ei_sector / 4;
        b = ei_bufblk;
	for(i=0;i<4;i++) { 
	   rd_cmd[5-i] = b & 0xff;
	   b = b >> 8;
	}
	while (1) {
	        if (!ei_command(rd_cmd,2048,"read block")) break;
		if (ei_retries >= EI_RETRIES) {
			ei_bufblk = -1; 
			ei_busy = 0;
			cli();
			end_request(0);
			do_ei_request();
			return;
		}
                ei_retries++;
		/* ei_reset(); */
	}
	ei_timeout = jiffies + EI_TMO;
	queue_task(&ei_tq,&tq_scheduler); 
}

static void do_ei_read( void )

{	ei_busy = 1;
	ei_retries = 0;

	ei_transfer();
	if (!ei_count) {
		end_request(1);
		ei_busy = 0;
		return;
	}
	sti();

	ei_start();
}

static void ei_interrupt( void *data )

{	if (!ei_ready()) {
		if (jiffies > ei_timeout) {
			ei_bufblk = -1;
			printk("eicd: pseudo-interrupt timeout, "
			       "status %x, error %x\n", 
				ei_read_regr(7),ei_read_regr(1));
			ei_disconnect();
			if (ei_retries < EI_RETRIES) {
				ei_retries++;
				/* ei_reset(); */
				ei_start();
				return;
			}
			cli();
			ei_busy = 0;
			end_request(0);
			do_ei_request();
			return;
		}
		queue_task(&ei_tq,&tq_scheduler); 
		return;
	}
	sti();
	if (ei_completion("read block")) {
                if (ei_retries < EI_RETRIES) {
                        ei_retries++;
			/* ei_reset(); */
			ei_start();
                        return;
                        }
		cli();
		ei_busy = 0;
		ei_bufblk = -1;
		end_request(0);
		do_ei_request();
		return;
	}
	do_ei_read();
	do_ei_request();
}
	
/* end of eicd.c */
