/* 
	epcd.c	(c) 1996,1997  Grant R. Guenther <grant@torque.net>
		               Under the terms of the GNU public license.

	epcd.c is a driver for parallel port CDrom drives based on the
        EPAT or EPEZ parallel to IDE adapter chip from Shuttle 
        Technologies.

        To compile the driver, ensure that /usr/include/linux and
        /usr/include/asm are links to the correct include files for 
        the target system. Then compile the driver with 

                cc -D__KERNEL__ -DMODULE -Wall -O2 -c epcd.c

        Alternatively, if you are using MODVERSIONS, do

                cc -D__KERNEL__ -DMODULE -DMODVERSIONS \
                   -include /usr/src/linux/include/linux/modversions.h \
                   -Wall -O2 -c epcd.c

        You must then load it with insmod.

	Before attempting to access the new driver, you will need to
        create a new device special file.  The following commands will
	do that for you:

		mknod /dev/epcd b 63 0
		chown root:disk /dev/epcd
		chmod 660 /dev/epcd

	Afterward, you can mount a disk in the usual way:

		mount -t iso9660 /dev/epcd /cdrom

	(assuming you have made a directory /cdrom to use as a
        mount point).

        (The number '63' above corresponds to the #define EP_MAJOR 63
        in the source below.  If it is necessary to change it, you
        must be sure that all uses of the device number are consistent.)

        The driver will search for an EPAT device on the available
        standard parallel ports.  The driver supports unidirectional
        and bidirectional parallel ports as well as some EPP ports.

        It is not currently possible to connect a printer to a chained
        printer port and expect Linux to use both devices at once.

        Keep an eye on http://www.torque.net/epez.html for news and
        other information about the driver.  If you have any problems
        with this driver, please send me, grant@torque.net, some mail 
        directly before posting into the newsgroups or mailing lists.

*/

#define	EP_VERSION	"0.03" 

#define	EP_BASE		0	/* set to 0 for autoprobe */
#define EP_HOLD         1       /* micro-second port settling time */
#define EP_DEBUG	0

#include <linux/version.h>
#include <linux/module.h>

#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/fs.h>
#include <linux/major.h>
#include <linux/kernel.h>
#include <linux/tqueue.h>
#include <linux/delay.h>
#include <linux/cdrom.h>
#include <linux/ioport.h>

#include <asm/system.h>
#include <asm/io.h>
#include <asm/segment.h>   /* change to uaccess.h for 2.1 kernels ... */

#ifndef EPCD_MAJOR
#define EPCD_MAJOR   63
#endif

#define MAJOR_NR EPCD_MAJOR

/* set up defines for blk.h,  why don't all drivers do it this way ? */

#define DEVICE_NAME "EPAT CDrom"
#define DEVICE_REQUEST do_ep_request
#define DEVICE_NR(device) (MINOR(device))
#define DEVICE_ON(device)
#define DEVICE_OFF(device)

#include <linux/blk.h>

#define EP_RETRIES	     5
#define EP_TMO		   600		/* timeout in jiffies */
#define EP_DELAY            50          /* spin delay in uS */

#define EP_SPIN		(10000/EP_DELAY)*EP_TMO

int epcd_init(void);
void epcd_setup(char * str, int * ints);
void cleanup_module( void );

static int 	ep_open(struct inode *inode, struct file *file);
static void 	do_ep_request(void);
static void 	do_ep_read(void);
static int 	ep_ioctl(struct inode *inode,struct file *file,
                         unsigned int cmd, unsigned long arg);
static void 	ep_release (struct inode *inode, struct file *file);

static int 	ep_detect(void);
static void     ep_lock(void);
static void     ep_unlock(void);
static void     ep_eject(void);
static void     ep_interrupt(void *data);

static int	ep_base = EP_BASE;
static int      ep_mode = 0;
static int	ep_hold = EP_HOLD;
static int      ep_debug = EP_DEBUG;;
static int      ep_fast_deglitch = 1;

static int      ep_retries;

static int ep_access = 0;		/* count of active opens ... */
static int ep_busy = 0;			/* request being processed ? */
static int ep_timeout;                  /* "interrupt" loop limiter */
static int ep_sector;			/* address of next requested sector */
static int ep_count;			/* number of blocks still to do */
static char * ep_buf;			/* buffer for request in progress */
static char ep_buffer[2048];		/* raw block buffer */
static int ep_bufblk = -1;		/* block in buffer, in CD units,
					   -1 for nothing there */
static int ep_epat_ver;
static int ep_found_0;
static int ep_found_2;

static struct tq_struct ep_tq = {0,0,ep_interrupt,NULL}; 

/* kernel glue structures */

static struct file_operations ep_fops = {
	NULL,			/* lseek - default */
	block_read,		/* read - general block-dev read */
	block_write,		/* write - general block-dev write */
	NULL,			/* readdir - bad */
	NULL,			/* select */
	ep_ioctl,		/* ioctl */
	NULL,			/* mmap */
	ep_open,		/* open */
	ep_release,		/* release */
	block_fsync,		/* fsync */
	NULL,			/* fasync */
	NULL,                   /* media change ? */
	NULL			/* revalidate new media */
};

int epcd_init (void)	/* preliminary initialisation */

{       if (ep_detect()) return -1;

	if (register_blkdev(MAJOR_NR,"epcd",&ep_fops)) {
		printk("epcd: unable to get major number %d\n",MAJOR_NR);
		return -1;
	}
	blk_dev[MAJOR_NR].request_fn = DEVICE_REQUEST;
	read_ahead[MAJOR_NR] = 8;	/* 8 sector (4kB) read ahead */

	return 0;
}

static int ep_open (struct inode *inode, struct file *file)

{	
	if (file->f_mode & 2) return -EROFS;  /* wants to write ? */

	MOD_INC_USE_COUNT;

	ep_lock();

	ep_access++;
	return 0;
}

static void do_ep_request (void)

{       if (ep_busy) return;
        while (1) {
	    if ((!CURRENT) || (CURRENT->rq_status == RQ_INACTIVE)) return;
	    INIT_REQUEST;
	    if (CURRENT->cmd == READ) {
	        ep_sector = CURRENT->sector;
	        ep_count = CURRENT->nr_sectors;
	        ep_buf = CURRENT->buffer;
	        do_ep_read();
		if (ep_busy) return;
	    } 
	    else end_request(0);
	}
}

static int ep_ioctl(struct inode *inode,struct file *file,
		    unsigned int cmd, unsigned long arg)

/* we currently support only the EJECT ioctl. */

{	switch (cmd) {
            case CDROMEJECT: if (ep_access == 1) {
				ep_eject();
				return 0;
			     }
	    default:
	        return -EINVAL;
	}
}

static void ep_release (struct inode *inode, struct file *file)

{	kdev_t	devp;

	ep_access--;
	if (!ep_access) { 
	    devp = inode->i_rdev;
	    fsync_dev(devp);
	    invalidate_inodes(devp);
	    invalidate_buffers(devp);
	    ep_unlock();
	}
	MOD_DEC_USE_COUNT;
}

#ifdef MODULE

/* Glue for modules ... */

int	init_module(void)

{	int	err;
	long    flags;

	save_flags(flags);
	cli();

	err = epcd_init();

	restore_flags(flags);
	return err;
}

void	cleanup_module(void)

{	long flags;
	int  range;

	if (ep_mode == 3) range = 8; else range = 3;

	save_flags(flags);
	cli();
	unregister_blkdev(MAJOR_NR,"epcd");
	release_region(ep_base,range);
	restore_flags(flags);
}

#else 

/* epcd_setup:  process lilo command parameters ...

   syntax:	epcd=base[,mode[,hold]]
*/

void    epcd_setup(char *str, int *ints)

{       if (ints[0] > 0) ep_base = ints[1];
        if (ints[0] > 1) ep_mode = ints[2];
        if (ints[0] > 2) ep_hold = ints[3];
} 

#endif

/*------------------------------------------------------------------*/

/* Now the actual hardware interface */

#define out_p(port,byte)   outb(byte,ep_base+port); udelay(ep_hold);
#define in_p(port)         (udelay(ep_hold),inb(ep_base+port))

#define w0(byte)                out_p(0,byte)
#define r0()                    (in_p(0) & 0xff)
#define w1(byte)                out_p(1,byte)
#define r1()                    (in_p(1) & 0xff)
#define w2(byte)                out_p(2,byte)
#define r2()                    (in_p(2) & 0xff)
#define w3(byte)                out_p(3,byte)
#define w4(byte)                out_p(4,byte)
#define r4()                    (in_p(4) & 0xff)

#define j44(a,b)                (((a>>4)&0x0f)+(b&0xf0))
#define j53(a,b)                (((a>>3)&0x1f)+((b<<4)&0xe0))


/* To access the control registers add 0x80 to the regr number */

static int ep_read_regr( char regr )

{       int  a, b, r;

        r = (regr & 7) + 0x18;
        if (regr & 0x80) r = (regr & 7) + 0x10; 
        if (regr & 0x40) r = regr & 0x1f;

        switch (ep_mode) {

        case 0: w0(r); w2(1); w2(3); 
                a = r1(); w2(4); b = r1();
                return j44(a,b);

        case 1: w0(0x40+r); w2(1); w2(4);
                a = r1(); b = r2(); w0(0xff);
                return j53(a,b);

        case 2: w0(0x20+r); w2(1); w2(0x25);
                a = r0(); w2(4);
                return a;

        case 3: w3(r); w2(0x24); a = r4(); w2(4);
                return a;

        }
        return -1;      /* never gets here */
}

static void ep_write_regr( char regr, char val )

{       int r;

        r = (regr & 7) + 0x18;
        if (regr & 0x80) r = (regr & 7) + 0x10; 
        if (regr & 0x40) r = regr & 0x1f;

        switch (ep_mode) {

        case 0:
        case 1:
        case 2: w0(0x60+r); w2(1); w0(val); w2(4);
                break;

        case 3: w3(0x40+r); w4(val);
                break;

        }
}

static void ep_write_cmd( char * buf, int len )

{       int ph, k;

        switch (ep_mode) {

        case 0:
        case 1:
        case 2: w0(0x67); w2(1); w2(5);
                ph = 0;
                for(k=0;k<len;k++) {
                        w0(buf[k]);
                        w2(4+ph);
                        ph = 1 - ph;
                }
                w2(7); w2(4);
                break;

        case 3: w3(0xc0); 
                for(k=0;k<len;k++) { w4(buf[k]) };
                w2(4);
                break;

        }
}

static void ep_read_data( char * buf, int len )

{       int  k, ph, a, b;

        switch (ep_mode) {

        case 0: w0(7); w2(1); w2(3); w0(0xff);
                ph = 0;
                for(k=0;k<len;k++) {
                        if (k == len-1) { w0(0xfd); }
                        w2(6+ph); a = r1();
                        if (a & 8) b = a; 
                          else { w2(4+ph); b = r1(); }
                        buf[k] = j44(a,b);
                        ph =  1 - ph;
                }
                w0(0); w2(4);
                break;

        case 1: w0(0x47); w2(1); w2(5); w0(0xff);
                ph = 0;
                for(k=0;k<len;k++) {
                        if (k == len-1) { w0(0xfd); }
                        w2(4+ph);
                        a = r1(); b = r2();
                        buf[k] = j53(a,b);
                        ph = 1 - ph;
                }
                w0(0); w2(4);
                break;

        case 2: w0(0x27); w2(1); w2(0x25); w0(0);
                ph = 0;
                for(k=0;k<len-1;k++) {
                        w2(0x24+ph);
                        buf[k] = r0();
                        ph = 1 - ph;
                }
                w2(0x26); w2(0x27); buf[len-1] = r0(); 
                w2(0x25); w2(4);
                break;

        case 3: w3(0x80); w2(0x24);
                for(k=0;k<len-1;k++) buf[k] = r4();
                w2(4); w3(0xa0); w2(0x24); buf[len-1] = r4();
                w2(4);
                break;

        }
}

#define WR(r,v)         ep_write_regr(r,v)
#define RR(r)           (ep_read_regr(r))

#define CCP(x)  w2(4);w0(0x22);w0(0xaa);w0(0x55);w0(0);w0(0xff);\
                w0(0x87);w0(0x78);w0(x);w2(4);w2(5);w2(4);w0(0xff);

static void ep_connect ( void  )

{       ep_found_0 = r0();
        ep_found_2 = r2();
        CCP(0); CCP(0xe0);
        w0(0); w2(1); w2(4);
        if (ep_mode == 3) {
                w0(0); w2(1); w2(4); w2(0xc);
                w0(0x40); w2(6); w2(7); w2(4); w2(0xc); w2(4);
        }
        if (ep_fast_deglitch) WR(0x48,0x10); 
	WR(0x4c,0x14); WR(0x4a,0x38); WR(0x52,0x10);

}

static void ep_disconnect ( void )

{       CCP(0x30);
        w0(ep_found_0);
        w2(ep_found_2);
} 

static int ep_ready( void )

{	return RR(0x52) & 0x20;
}

static int ep_test_proto( void )

{       int     k, e, f, cc;

        e = 0;
        ep_connect();
        cc = RR(0x4d);
        ep_disconnect();
        ep_connect();
        WR(6,0xa0);
        for (k=0;k<256;k++) {
                WR(2,k^0xaa);
                WR(3,k^0x55);
                if (RR(2) != (k^0xaa)) e++;
                }
        ep_disconnect();
        f = 0;
        ep_connect();
        WR(0x53,1); WR(0x53,0); WR(0x4a,0x11);
        ep_read_data(ep_buffer,512);
        
        for (k=0;k<256;k++) {
            if ((ep_buffer[2*k] & 0xff) != k) f++;
            if ((ep_buffer[2*k+1] & 0xff) != (0xff-k)) f++;
        }
        WR(0x4a,0x38);
        ep_epat_ver = RR(0x4b);
        ep_disconnect();

        if (ep_debug)  {
            printk("ep: port 0x%x, mode %d, ccr %x, test %d+%d errors\n"
                   ,ep_base,ep_mode,cc,e,f);
        }
        
        if (ep_debug && (e+f == 0)) {
                ep_connect();
                WR(6,0xa0);  udelay(30);
                if (RR(0x86) != RR(7)) 
                   printk("ep: oops, ALT_STATUS != STATUS\n");
                ep_disconnect();
        }

        return e+f;
}

/*-------------------------------------------------------------------*/

static int ep_wait_drq( char * fun, char * msg1, char * msg2 )

{       int j, r, e;

        j = 0;
        while (!(ep_read_regr(7)&8)&&(j++<EP_SPIN)) udelay(EP_DELAY);
        r = ep_read_regr(7);
        e = ep_read_regr(1);
        if (j >= EP_SPIN) {
            if (fun) 
              printk("epcd: %s%s: DRQ timeout on %s, status: %x error: %x\n",
                             fun,msg1,msg2,r,e);
            return -1;
        }
        return 0;
}

static int ep_wait_cmd( )

{       int j, r, e;

        j = 0;
        while ((ep_read_regr(7)&0x88)&&(j++<EP_SPIN)) udelay(EP_DELAY);
        r = ep_read_regr(7);
        e = ep_read_regr(1);
        if (j >= EP_SPIN) {
            printk("epcd: Pre-command timeout:  status: %x error: %x\n",
                             r,e);
            return -1;
        }
        return 0;
}

static int ep_wait( char * fun, char * msg1, char * msg2 )

{       int j, r, e;

        j = 0;
        while ((!(ep_ready()))&&(j++<EP_SPIN)) udelay(EP_DELAY);
        r = ep_read_regr(7);
        e = ep_read_regr(1);
        if ((j >= EP_SPIN) || (r & 1)) {
            if (j >= EP_SPIN) e |= 0x100;
            if (fun) printk("epcd: %s%s: Error %s, status: %x error: %x\n",
                          fun,msg1,msg2,r,e);
            return -1;
        }
        return 0;
}

static int ep_req_sense( char * fun, char * msg )

{	char	rs_cmd[12] = { 0x03,0,0,0,18,0,0,0,0,0,0,0 };
	int n, r, p;

        if (ep_wait_cmd()) { ep_disconnect(); return -1; }

        ep_write_regr(4,18);
        ep_write_regr(5,0);
        ep_write_regr(6,0xa0);          /* drive 0 */
        ep_write_regr(7,0xa0);          /* ATAPI packet command */
	if (ep_wait_drq(fun,msg,"request sense command")) return -1;
	ep_write_cmd(rs_cmd,12);
	if (!(r=ep_wait(fun,msg,"getting request sense data"))) {
	    p = ep_read_regr(2);
            if (p == 2) {
                n = (ep_read_regr(4) + 256*ep_read_regr(5));
		ep_wait_drq(fun,msg,"getting sense data");
                ep_read_data(ep_buffer,n);
                r=ep_wait(fun,msg,"on request sense done");
                if (!r) if (fun)
	          printk("epcd: %s%s: sense key: %x, ASC: %x, ASQ: %x\n",
	            fun, msg, ep_buffer[2]&0xf, ep_buffer[12], ep_buffer[13]);
	    } else if (p == 3) {
                r = (ep_read_regr(7) & 1);
                if (r) printk("epcd: phase error, status %x, error %x\n",
                                r,ep_read_regr(1));
	    }
	}
	return r;
}

static int ep_command( char * cmd, int dlen, char * fun )

{	int	r;

	ep_connect();

        if (ep_wait_cmd()) { ep_disconnect(); return -1; }

	ep_write_regr(4,dlen % 256);
	ep_write_regr(5,dlen / 256);
	ep_write_regr(6,0xa0);		/* drive 0 */
	ep_write_regr(7,0xa0);		/* ATAPI packet command */

        if (ep_read_regr(7)&1) {
            if (ep_req_sense(fun," before command")) { 
                ep_disconnect();
                return -1;
                }

            if (ep_wait_cmd()) { ep_disconnect(); return -1; }

            ep_write_regr(4,dlen % 256);
            ep_write_regr(5,dlen / 256);
            ep_write_regr(6,0xa0);          /* drive 0 */
            ep_write_regr(7,0xa0);          /* ATAPI packet command */
	}

	if ((r=ep_wait_drq(fun,"","sending command"))) {
	   ep_disconnect();
	   return r;
	}

	ep_write_cmd(cmd,12);
	return 0;
}

static int ep_completion( char * fun )

{	int r, n, p;

	if (!(r=ep_wait(fun,"","command completion"))) {
	    p = ep_read_regr(2);
	    if (p == 2) {
	        n = (ep_read_regr(4) + 256*ep_read_regr(5));
		ep_wait_drq(fun,"","getting data");
	        ep_read_data(ep_buffer,n);
	        r=ep_wait(fun,"","waiting after data");
	    } else if (p == 3) {
		r = (ep_read_regr(7) & 1);
		if (r) printk("epcd: phase error, status %x, error %x\n",
				r,ep_read_regr(1));
	    }
	}
	if (r) r = ep_req_sense(fun," after command");

	ep_disconnect(); 
	return r;
}

static int ep_atapi( char * cmd, int dlen, char * fun )

{	int r;

	if (!(r=ep_command(cmd,dlen,fun)))
	  r = ep_completion(fun);
	return r;
}

static void ep_lock(void)

{	char	lo_cmd[12] = { 0x1e,0,0,0,1,0,0,0,0,0,0,0 };
	char	cl_cmd[12] = { 0x1b,0,0,0,3,0,0,0,0,0,0,0 };

	ep_atapi(cl_cmd,0,NULL);
	ep_atapi(cl_cmd,0,NULL);
	ep_atapi(cl_cmd,0,NULL);
	ep_atapi(cl_cmd,0,NULL);
	ep_atapi(cl_cmd,0,"close door");
        ep_atapi(lo_cmd,0,NULL);
        ep_atapi(lo_cmd,0,"lock door");
}

static void ep_unlock( void)

{	char	un_cmd[12] = { 0x1e,0,0,0,0,0,0,0,0,0,0,0 };

	ep_atapi(un_cmd,0,"unlock door");
}

static void ep_eject( void)

{	char	ej_cmd[12] = { 0x1b,0,0,0,2,0,0,0,0,0,0,0 };

	ep_unlock();
	ep_atapi(ej_cmd,0,"eject");
}

static int ep_reset( void )

/* the ATAPI standard actually specifies the contents of all 7 registers
   after a reset, but the specification is ambiguous concerning the last
   two bytes, and different drives interpret the standard differently.
*/

{	int	i, flg;
	int	expect[5] = {1,1,1,0x14,0xeb};
	long	flags;

	ep_connect();
	ep_write_regr(6,0xa0);
	ep_write_regr(7,8);

	save_flags(flags);
	sti();
	udelay(500000);  	/* delay 0.5 seconds */
	restore_flags(flags);

	flg = 1;
	for(i=0;i<5;i++) flg &= (ep_read_regr(i+1) == expect[i]);

	if (ep_debug) {
		printk("epcd RESET: ");
		for (i=0;i<7;i++) printk("%3x",ep_read_regr(i+1));
		printk("\n");
	}
	
	ep_disconnect();
	return flg-1;	
}

static int ep_identify( char * id )

{	int k;
	char   id_cmd[12] = {0x12,0,0,0,36,0,0,0,0,0,0,0};

	ep_bufblk = -1;
        if (ep_atapi(id_cmd,36,"identify")) return -1;
	for (k=0;k<16;k++) id[k] = ep_buffer[16+k];
	id[16] = 0;
	return 0;
}

static void ep_probe ( int port, int max_mode )

{	int	best, range;

	range = 3;
	if (max_mode == 3) range = 8;
	if (check_region(port,range)) {
		max_mode = 1;
		range = 3;
		if (check_region(port,range)) return;
	}

	ep_base = port;  best = -1;

	if (ep_debug) printk("epcd PROBE: port 0x%x, max_mode %d\n",
			port, max_mode);

	for (ep_mode=0;ep_mode<=max_mode;ep_mode++)
	   if (!ep_test_proto()) best = ep_mode;

	ep_mode = best;
	if (ep_mode < 0) ep_base = 0;
}
 
static int ep_detect( void )

{	char    id[18];
	int	range;
        char    *mode_str[4] = {"4-bit","Shuttle 5/3","8-bit","EPP"};

	if (!ep_base) ep_probe(0x278,3);
	if (!ep_base) ep_probe(0x378,3);
	if (!ep_base) ep_probe(0x3bc,2);

	if (!ep_base) {
		printk("epcd: Autoprobe failed\n");
		return -1;
	}

	if (ep_mode == 3) range = 8; else range = 3;
	if (check_region(ep_base,range)) {
		printk("epcd: Ports at 0x%x are not available\n",ep_base);
		return -1;
	}

	if (ep_test_proto()) {
		printk("epcd: Can't access EPAT adapter in %s mode\n",
			mode_str[ep_mode]);
		return -1;
	}

        printk("epcd %s: Shuttle EPAT adapter version %x "
	       "using port 0x%x in %s mode\n",
                EP_VERSION,ep_epat_ver,ep_base,mode_str[ep_mode]);

	if (ep_reset()) {
		printk("epcd: Failed to reset CD drive\n");
		return -1;
	}

	ep_identify(id);
	if (ep_identify(id)) {
		printk("epcd: ATAPI inquiry failed\n");
		return -1;
	}

	if (ep_mode == 3) range = 8; else range = 3;	
	request_region(ep_base,range,"epcd");

	printk("epcd: Found %s\n",id);

	return 0;
}

static void ep_transfer( void )

{	int	k, o;

	while (ep_count && (ep_sector/4 == ep_bufblk)) {
		o = (ep_sector % 4) * 512;
		for(k=0;k<512;k++) ep_buf[k] = ep_buffer[o+k];
		ep_count--;
		ep_buf += 512;
		ep_sector++;
	}
}

static void ep_start( void )

{	int	b, i;
	char	rd_cmd[12] = {0xa8,0,0,0,0,0,0,0,0,1,0,0};

	ep_bufblk = ep_sector / 4;
        b = ep_bufblk;
	for(i=0;i<4;i++) { 
	   rd_cmd[5-i] = b & 0xff;
	   b = b >> 8;
	}
	while (1) {
	        if (!ep_command(rd_cmd,2048,"read block")) break;
		if (ep_retries >= EP_RETRIES) {
			ep_bufblk = -1; 
			ep_busy = 0;
			cli();
			end_request(0);
			do_ep_request();
			return;
		}
		udelay(1000);
                ep_retries++;
		ep_disconnect();
	}
	ep_timeout = jiffies + EP_TMO;
	queue_task(&ep_tq,&tq_scheduler); 
}

static void do_ep_read( void )

{	ep_busy = 1;
	ep_retries = 0;

	ep_transfer();
	if (!ep_count) {
		end_request(1);
		ep_busy = 0;
		return;
	}
	sti();

	ep_start();
}

static void ep_interrupt( void *data )

{	if (!ep_ready()) {
		if (jiffies > ep_timeout) {
			ep_bufblk = -1;
			ep_disconnect();
			printk("epcd: pseudo-interrupt timeout, "
			       "status %x, error %x\n", 
				ep_read_regr(7),ep_read_regr(1));
			if (ep_retries < EP_RETRIES) {
				udelay(1000);
				ep_retries++;
				ep_start();
				return;
			}
			cli();
			ep_busy = 0;
			end_request(0);
			do_ep_request();
			return;
		}
		queue_task(&ep_tq,&tq_scheduler); 
		return;
	}
	sti();
	if (ep_completion("read block")) {
                if (ep_retries < EP_RETRIES) {
                        udelay(1000);
                        ep_retries++;
			ep_start();
                        return;
                        }
		cli();
		ep_busy = 0;
		ep_bufblk = -1;
		end_request(0);
		do_ep_request();
		return;
	}
	do_ep_read();
	do_ep_request();
}
	
/* end of epcd.c */
