/* 
	kicd.c	(c) 1996,1997  Grant R. Guenther <grant@torque.net>
		               Under the terms of the GNU public license.

	kicd.c is a driver for parallel port CDrom drives based on the
        KBIC-951A parallel to IDE adapter chip from KingByte
        Information Systems.

        To compile the driver, ensure that /usr/include/linux and
        /usr/include/asm are links to the correct include files for 
        the target system. Then compile the driver with 

                cc -D__KERNEL__ -DMODULE -Wall -O2 -c kicd.c

        Alternatively, if you are using MODVERSIONS, do

                cc -D__KERNEL__ -DMODULE -DMODVERSIONS \
                   -include /usr/src/linux/include/linux/modversions.h \
                   -Wall -O2 -c kicd.c

        You must then load it with insmod.

	Before attempting to access the new driver, you will need to
        create a new device special file.  The following commands will
	do that for you:

		mknod /dev/kicd b 63 0
		chown root:disk /dev/kicd
		chmod 660 /dev/kicd

	Afterward, you can mount a disk in the usual way:

		mount -t iso9660 /dev/kicd /cdrom

	(assuming you have made a directory /cdrom to use as a
        mount point).

        (The number '63' above corresponds to the #define KI_MAJOR 63
        in the source below.  If it is necessary to change it, you
        must be sure that all uses of the device number are consistent.)

        The driver will search for an KBIC device on the available
        standard parallel ports.  The driver supports unidirectional
        and bidirectional parallel ports as well as some EPP ports.

        It is not currently possible to connect a printer to a chained
        printer port and expect Linux to use both devices at once.

        Keep an eye on http://www.torque.net/kbic.html for news and
        other information about the driver.  If you have any problems
        with this driver, please send me, grant@torque.net, some mail 
        directly before posting into the newsgroups or mailing lists.

*/

#define	KI_VERSION	"0.02" 

#define	KI_BASE		0	/* set to 0 for autoprobe */
#define KI_HOLD         1       /* micro-second port settling time */
#define KI_DEBUG	0

#include <linux/version.h>
#include <linux/module.h>

#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/fs.h>
#include <linux/major.h>
#include <linux/kernel.h>
#include <linux/tqueue.h>
#include <linux/delay.h>
#include <linux/cdrom.h>
#include <linux/ioport.h>

#include <asm/system.h>
#include <asm/io.h>
#include <asm/segment.h>   /* change to uaccess.h for 2.1 kernels ... */

#ifndef KICD_MAJOR
#define KICD_MAJOR   63
#endif

#define MAJOR_NR KICD_MAJOR

/* set up defines for blk.h,  why don't all drivers do it this way ? */

#define DEVICE_NAME "KingByte CDrom"
#define DEVICE_REQUEST do_ki_request
#define DEVICE_NR(device) (MINOR(device))
#define DEVICE_ON(device)
#define DEVICE_OFF(device)

#include <linux/blk.h>

#define KI_RETRIES	     5
#define KI_TMO		   600		/* timeout in jiffies */
#define KI_DELAY            50          /* spin delay in uS */

#define KI_SPIN		(10000/KI_DELAY)*KI_TMO

int kicd_init(void);
void kicd_setup(char * str, int * ints);
void cleanup_module( void );

static int 	ki_open(struct inode *inode, struct file *file);
static void 	do_ki_request(void);
static void 	do_ki_read(void);
static int 	ki_ioctl(struct inode *inode,struct file *file,
                         unsigned int cmd, unsigned long arg);
static void 	ki_release (struct inode *inode, struct file *file);

static int 	ki_detect(void);
static void     ki_lock(void);
static void     ki_unlock(void);
static void     ki_eject(void);
static void     ki_interrupt(void *data);

static int	ki_base = KI_BASE;
static int      ki_mode = 0;
static int	ki_hold = KI_HOLD;
static int      ki_debug = KI_DEBUG;;

static int      ki_found_0;
static int      ki_found_2;
static int      ki_retries;

static int ki_access = 0;		/* count of active opens ... */
static int ki_busy = 0;			/* request being processed ? */
static int ki_timeout;                  /* "interrupt" loop limiter */
static int ki_sector;			/* address of next requested sector */
static int ki_count;			/* number of blocks still to do */
static char * ki_buf;			/* buffer for request in progress */
static char ki_buffer[2048];		/* raw block buffer */
static int ki_bufblk = -1;		/* block in buffer, in CD units,
					   -1 for nothing there */

static struct tq_struct ki_tq = {0,0,ki_interrupt,NULL}; 

/* kernel glue structures */

static struct file_operations ki_fops = {
	NULL,			/* lseek - default */
	block_read,		/* read - general block-dev read */
	block_write,		/* write - general block-dev write */
	NULL,			/* readdir - bad */
	NULL,			/* select */
	ki_ioctl,		/* ioctl */
	NULL,			/* mmap */
	ki_open,		/* open */
	ki_release,		/* release */
	block_fsync,		/* fsync */
	NULL,			/* fasync */
	NULL,                   /* media change ? */
	NULL			/* revalidate new media */
};

int kicd_init (void)	/* preliminary initialisation */

{       if (ki_detect()) return -1;

	if (register_blkdev(MAJOR_NR,"kicd",&ki_fops)) {
		printk("kicd: unable to get major number %d\n",MAJOR_NR);
		return -1;
	}
	blk_dev[MAJOR_NR].request_fn = DEVICE_REQUEST;
	read_ahead[MAJOR_NR] = 8;	/* 8 sector (4kB) read ahead */

	return 0;
}

static int ki_open (struct inode *inode, struct file *file)

{	
	if (file->f_mode & 2) return -EROFS;  /* wants to write ? */

	MOD_INC_USE_COUNT;

	ki_lock();

	ki_access++;
	return 0;
}

static void do_ki_request (void)

{       if (ki_busy) return;
        while (1) {
	    if ((!CURRENT) || (CURRENT->rq_status == RQ_INACTIVE)) return;
	    INIT_REQUEST;
	    if (CURRENT->cmd == READ) {
	        ki_sector = CURRENT->sector;
	        ki_count = CURRENT->nr_sectors;
	        ki_buf = CURRENT->buffer;
	        do_ki_read();
		if (ki_busy) return;
	    } 
	    else end_request(0);
	}
}

static int ki_ioctl(struct inode *inode,struct file *file,
		    unsigned int cmd, unsigned long arg)

/* we currently support only the EJECT ioctl. */

{	switch (cmd) {
            case CDROMEJECT: if (ki_access == 1) {
				ki_eject();
				return 0;
			     }
	    default:
	        return -EINVAL;
	}
}

static void ki_release (struct inode *inode, struct file *file)

{	kdev_t	devp;

	ki_access--;
	if (!ki_access) { 
	    devp = inode->i_rdev;
	    fsync_dev(devp);
	    invalidate_inodes(devp);
	    invalidate_buffers(devp);
	    ki_unlock();
	}
	MOD_DEC_USE_COUNT;
}

#ifdef MODULE

/* Glue for modules ... */

int	init_module(void)

{	int	err;
	long    flags;

	save_flags(flags);
	cli();

	err = kicd_init();

	restore_flags(flags);
	return err;
}

void	cleanup_module(void)

{	long flags;
	int  range;

	if (ki_mode == 3) range = 8; else range = 3;

	save_flags(flags);
	cli();
	unregister_blkdev(MAJOR_NR,"kicd");
	release_region(ki_base,range);
	restore_flags(flags);
}

#else 

/* kicd_setup:  process lilo command parameters ...

   syntax:	kicd=base[,mode[,hold]]
*/

void    kicd_setup(char *str, int *ints)

{       if (ints[0] > 0) ki_base = ints[1];
        if (ints[0] > 1) ki_mode = ints[2];
        if (ints[0] > 2) ki_hold = ints[3];
} 

#endif

#define out_p(port,byte)  outb(byte,ki_base+port);udelay(ki_hold);
#define in_p(port)	  (udelay(ki_hold),inb(ki_base+port))
#define in_p_w(port)	  (udelay(ki_hold),inw(ki_base+port))

#define w0(byte)  		out_p(0,byte)
#define w2(byte)  		out_p(2,byte)
#define w3(byte)  		out_p(3,byte)
#define w4(byte)  		out_p(4,byte)

#define r0()      		(in_p(0) & 0xff)
#define r1()      		(in_p(1) & 0xff)
#define r2()			(in_p(2) & 0xff)
#define r12()			(in_p_w(1) & 0xffff)
#define r4()			(in_p(4) & 0xff)

#define j44(a,b)                ((((a>>4)&0x0f)|(b&0xf0))^0x88)
#define j53(w)                  (((w>>3)&0x1f)|((w>>4)&0xe0))

/* To access the control registers add 0x80 to the regr number */

static int ki_read_regr( char regr )

{       int     a, b, r, s;

        if (regr & 0x80) s = 0x40; else s = 0x80;
        r = (regr & 7);

        switch (ki_mode) {

        case 0: w0(r|0x18|s); w2(4); w2(6); w2(4); w2(1); w0(8);
                a = r1(); w0(0x28); b = r1(); w2(4);
                return j44(a,b);

        case 1: w0(r|0x38|s); w2(4); w2(6); w2(4); w2(5); w0(8);
                a = r12(); w2(4);
                return j53(a);

        case 2: w0(r|0x08|s); w2(4); w2(6); w2(4); w2(0xa5); w2(0xa1);
                a = r0(); w2(4);
                return a;

        case 3: w0(0x20|s); w2(4); w2(6); w2(4); w3(r);
                a = r4(); b = r4(); w2(4); w2(0); w2(4);
                return a;

        }
        return -1;
}

static void ki_write_regr( char regr, char val )

{       int  r, s;

        if (regr & 0x80) s = 0x40; else s = 0x80;
        r = (regr & 7);

        switch (ki_mode) {

        case 0: 
        case 1:
        case 2: w0(r|0x10|s); w2(4); w2(6); w2(4); 
                w0(val); w2(5); w2(4);
                break;

        case 3: w0(0x20|s); w2(4); w2(6); w2(4); w3(r);
                w4(val); w4(val);
                w2(4); w2(0); w2(4);
                break;

        }
}

static void ki_write_cmd( char * buf, int len )

{       int     k;

        switch (ki_mode) {

        case 0:
        case 1:
        case 2: w0(0x90); w2(4); w2(6); w2(4); 
                for(k=0;k<len/2;k++) {
                        w0(buf[2*k+1]); w2(0); w2(4); 
                        w0(buf[2*k]);   w2(5); w2(4); 
                }
                break;

        case 3: w0(0xa0); w2(4); w2(6); w2(4); w3(0);
                for(k=0;k<len/2;k++) {
                        w4(buf[2*k+1]); 
                        w4(buf[2*k]);
                }
                w2(4); w2(0); w2(4);
                break;
        }

}

static void ki_read_data( char * buf, int len )

{       int     k, a, b;

        switch (ki_mode) {

        case 0: w0(0x98); w2(4); w2(6); w2(4);
                for (k=0;k<len/2;k++) {
                        w2(1); w0(8);    a = r1();
                               w0(0x28); b = r1();
                        buf[2*k]   = j44(a,b);
                        w2(5);           b = r1();
                               w0(8);    a = r1();
                        buf[2*k+1] = j44(a,b);
                        w2(4);
                } 
                break;

        case 1: w0(0xb8); w2(4); w2(6); w2(4); 
                for (k=0;k<len/2;k++) {
                        w2(4); w2(5); 
                        w0(8);    buf[2*k]   = j53(r12());
                        w0(0xb8); buf[2*k+1] = j53(r12());
                }
                w2(4);
                break;

        case 2: w0(0x88); w2(4); w2(6); w2(4);
                for (k=0;k<len/2;k++) {
                        w2(0xa0); w2(0xa1); buf[2*k] = r0();
                        w2(0xa5); buf[2*k+1] = r0();
                }
                w2(4);
                break;

        case 3: w0(0xa0); w2(4); w2(6); w2(4); w3(0);
                for (k=0;k<len;k++) buf[k] = r4();
                w2(4); w2(0); w2(4);
                break;

        }
}

static void ki_connect ( void  )

{       ki_found_0 = r0();
        ki_found_2 = r2();
        w2(4); 
}

static void ki_disconnect ( void )

{       w2(ki_found_2);
        w0(ki_found_0);
} 

static int ki_ready( void )

{	w2(4);
	return (r1() & 0x10);
}

static int ki_test_proto( void )

{       int     k, e;

        e = 0;
        ki_connect();
        for (k=0;k<256;k++) {
                ki_write_regr(2,k^0xaa);
                ki_write_regr(3,k^0x55);
                if (ki_read_regr(2) != (k^0xaa)) e++;
                }
        ki_disconnect();

        if (ki_debug) {

            printk("ki: port 0x%x, mode %d, mode test %d errors\n",
                ki_base,ki_mode,e);

            if (ki_debug > 1) {

                int t;

                ki_connect();
                printk("..... ");
                for (k=0;k<8;k++) {
                    t = 1 << k;
                    ki_write_regr(2,t);
                    ki_write_regr(3,0xaa);
                    printk("%x/%x ",t,(ki_read_regr(2)));
                }
                printk("\n");
                ki_disconnect();
            }

            if (e > 0) return e;

            ki_connect();
            ki_write_regr(6,0xa0);  udelay(30);
            if (ki_read_regr(0x86) != ki_read_regr(7)) 
               printk("ki: oops, ALT_STATUS != STATUS\n");
            ki_disconnect();
        }

        return e;
}

/*-------------------------------------------------------------------*/

static int ki_wait_drq( char * fun, char * msg1, char * msg2 )

{	int j, r, e;

	j = 0;
	while (!(ki_read_regr(7)&8)&&(j++<KI_SPIN)) udelay(KI_DELAY);
        r = ki_read_regr(7);
	e = ki_read_regr(1);
	if (j >= KI_SPIN) {
	    if (fun) 
	      printk("kicd: %s%s: DRQ timeout on %s, status: %x error: %x\n",
			     fun,msg1,msg2,r,e);
	    return -1;
	}
	return 0;
}

static int ki_wait( char * fun, char * msg1, char * msg2 )

{	int j, r, e;

	j = 0;
	while ((!(ki_ready()))&&(j++<KI_SPIN)) udelay(KI_DELAY);
        r = ki_read_regr(7);
	e = ki_read_regr(1);
	if ((j >= KI_SPIN) || (r & 1)) {
	    if (j >= KI_SPIN) e |= 0x100;
	    if (fun) printk("kicd: %s%s: Error %s, status: %x error: %x\n",
			  fun,msg1,msg2,r,e);
	    return -1;
	}
	return 0;
}

static int ki_req_sense( char * fun, char * msg )

{	char	rs_cmd[12] = { 0x03,0,0,0,18,0,0,0,0,0,0,0 };
	int n, r;

        ki_write_regr(4,18);
        ki_write_regr(5,0);
        ki_write_regr(6,0xa0);          /* drive 0 */
        ki_write_regr(7,0xa0);          /* ATAPI packet command */
	if (ki_wait_drq(fun,msg,"request sense command")) return -1;
	ki_write_cmd(rs_cmd,12);
	if (!(r=ki_wait(fun,msg,"getting request sense data"))) {
            if (ki_read_regr(2) == 2) {
                n = (ki_read_regr(4) + 256*ki_read_regr(5));
		ki_wait_drq(NULL,NULL,NULL);
                ki_read_data(ki_buffer,n);
                r=ki_wait(fun,msg,"on request sense done");
                if (!r) if (fun)
	          printk("kicd: %s%s: sense key: %x, ASC: %x, ASQ: %x\n",
	            fun, msg, ki_buffer[2]&0xf, ki_buffer[12], ki_buffer[13]);
	    }
	}
	return r;
}

static int ki_command( char * cmd, int dlen, char * fun )

{	int	r;

	ki_connect();

	ki_write_regr(4,dlen % 256);
	ki_write_regr(5,dlen / 256);
	ki_write_regr(6,0xa0);		/* drive 0 */
	ki_write_regr(7,0xa0);		/* ATAPI packet command */

        if (ki_read_regr(7)&1) {
            if (ki_req_sense(fun," before command")) { 
                ki_disconnect();
                return -1;
                }
            ki_write_regr(4,dlen % 256);
            ki_write_regr(5,dlen / 256);
            ki_write_regr(6,0xa0);          /* drive 0 */
            ki_write_regr(7,0xa0);          /* ATAPI packet command */
	}

	if ((r=ki_wait_drq(fun,"","sending  command"))) {
	   ki_disconnect();
	   return r;
	}

	ki_write_cmd(cmd,12);
	return 0;
}

static int ki_completion( char * fun )

{	int r, n;

	if (!(r=ki_wait(fun,"","command completion"))) {
	    if (ki_read_regr(2) == 2) {
	        n = (ki_read_regr(4) + 256*ki_read_regr(5));
		ki_wait_drq(NULL,NULL,NULL);
	        ki_read_data(ki_buffer,n);
	        r=ki_wait(fun,"","waiting after data");
	    } 
	}
	if (r) r = ki_req_sense(fun," after command");

	ki_disconnect(); 
	return r;
}

static int ki_atapi( char * cmd, int dlen, char * fun )

{	int r;

	if (!(r=ki_command(cmd,dlen,fun)))
	  r = ki_completion(fun);
	return r;
}

static void ki_lock(void)

{	char	lo_cmd[12] = { 0x1e,0,0,0,1,0,0,0,0,0,0,0 };
	char	cl_cmd[12] = { 0x1b,0,0,0,3,0,0,0,0,0,0,0 };

	ki_atapi(cl_cmd,0,NULL);
	ki_atapi(cl_cmd,0,NULL);
	ki_atapi(cl_cmd,0,NULL);
	ki_atapi(cl_cmd,0,"close door");

        ki_atapi(lo_cmd,0,NULL);
        ki_atapi(lo_cmd,0,"lock door");
}

static void ki_unlock( void)

{	char	un_cmd[12] = { 0x1e,0,0,0,0,0,0,0,0,0,0,0 };

	ki_atapi(un_cmd,0,"unlock door");
}

static void ki_eject( void)

{	char	ej_cmd[12] = { 0x1b,0,0,0,2,0,0,0,0,0,0,0 };

	ki_unlock();
	ki_atapi(ej_cmd,0,"eject");
}

static int ki_reset( void )

/* the ATAPI standard actually specifies the contents of all 7 registers
   after a reset, but the specification is ambiguous concerning the last
   two bytes, and different drives interpret the standard differently.
*/

{	int	i, flg;
	int	expect[5] = {1,1,1,0x14,0xeb};
	long	flags;

	ki_connect();
	ki_write_regr(6,0xa0);
	ki_write_regr(7,8);

	save_flags(flags);
	sti();
	udelay(500000);  	/* delay 0.5 seconds */
	restore_flags(flags);

	flg = 1;
	for(i=0;i<5;i++) flg &= (ki_read_regr(i+1) == expect[i]);

	if (ki_debug) {
		printk("kicd RESET: ");
		for (i=0;i<5;i++) printk("%3x",ki_read_regr(i+1));
		printk("\n");
	}
	
	ki_disconnect();
	return flg-1;	
}

static int ki_identify( char * id )

{	int k;
	char   id_cmd[12] = {0x12,0,0,0,36,0,0,0,0,0,0,0};

	ki_bufblk = -1;
        if (ki_atapi(id_cmd,36,"identify")) return -1;
	for (k=0;k<16;k++) id[k] = ki_buffer[16+k];
	id[16] = 0;
	return 0;
}

static void ki_probe ( int port, int max_mode )

{	int	best, range;

	range = 3;
	if (max_mode == 3) range = 8;
	if (check_region(port,range)) {
		max_mode = 2;
		range = 3;
		if (check_region(port,range)) return;
	}

	ki_base = port;  best = -1;

	if (ki_debug) printk("kicd PROBE: port 0x%x, max_mode %d\n",
			port, max_mode);

	for (ki_mode=0;ki_mode<=max_mode;ki_mode++)
	   if (!ki_test_proto()) best = ki_mode;

	ki_mode = best;
	if (ki_mode < 0) ki_base = 0;
}
 
static int ki_detect( void )

{	char    id[18];
	int	range;
        char    *mode_str[4] = {"4-bit","KBIC 5/3","8-bit","EPP"};

	if (!ki_base) ki_probe(0x278,3);
	if (!ki_base) ki_probe(0x378,3);
	if (!ki_base) ki_probe(0x3bc,2);

	if (!ki_base) {
		printk("kicd: Autoprobe failed\n");
		return -1;
	}

	if (ki_mode == 3) range = 8; else range = 3;
	if (check_region(ki_base,range)) {
		printk("kicd: Ports at 0x%x are not available\n",ki_base);
		return -1;
	}

	if (ki_test_proto()) {
		printk("kicd: Can't access KBIC adapter in %s mode\n",
			mode_str[ki_mode]);
		return -1;
	}

        printk("kicd %s: KBIC adapter using port 0x%x in %s mode\n",
                KI_VERSION,ki_base,mode_str[ki_mode]);

	if (ki_reset()) {
		printk("kicd: Failed to reset CD drive\n");
		return -1;
	}

	if (ki_identify(id)) {
		printk("kicd: ATAPI inquiry failed\n");
		return -1;
	}

	if (ki_mode == 3) range = 8; else range = 3;	
	request_region(ki_base,range,"kicd");

	printk("kicd: Found %s\n",id);

	return 0;
}

static void ki_transfer( void )

{	int	k, o;

	while (ki_count && (ki_sector/4 == ki_bufblk)) {
		o = (ki_sector % 4) * 512;
		for(k=0;k<512;k++) ki_buf[k] = ki_buffer[o+k];
		ki_count--;
		ki_buf += 512;
		ki_sector++;
	}
}

static void ki_start( void )

{	int	b, i;
	char	rd_cmd[12] = {0xa8,0,0,0,0,0,0,0,0,1,0,0};

	ki_bufblk = ki_sector / 4;
        b = ki_bufblk;
	for(i=0;i<4;i++) { 
	   rd_cmd[5-i] = b & 0xff;
	   b = b >> 8;
	}
	if (ki_command(rd_cmd,2048,"read block")) {
		ki_bufblk = -1; 
		ki_busy = 0;
		cli();
		end_request(0);
		do_ki_request();
		return;
	}
	ki_timeout = jiffies + KI_TMO;
	queue_task(&ki_tq,&tq_scheduler); 
}

static void do_ki_read( void )

{	ki_busy = 1;
	ki_retries = 0;
	ki_transfer();
	if (!ki_count) {
		end_request(1);
		ki_busy = 0;
		return;
	}
	sti();

	ki_start();
}

static void ki_interrupt( void *data)

{	if (!(ki_ready())) {
		if (jiffies > ki_timeout) {
			ki_bufblk = -1;
			ki_busy = 0;
			end_request(0);
			do_ki_request();
			return;
		}
		queue_task(&ki_tq,&tq_scheduler); 
		return;
	}
	sti();
	if (ki_completion("read block")) {
                if (ki_retries < KI_RETRIES) {
                        udelay(100000);
                        ki_retries++;
                        
			ki_start();

                        return;
                        }
		cli();
		ki_busy = 0;
		ki_bufblk = -1;
		end_request(0);
		do_ki_request();
		return;
	}
	do_ki_read();
	do_ki_request();
}
	
/* end of kicd.c */
