/* 
	oicd.c	(c) 1996,1997  Grant R. Guenther <grant@torque.net>
		               Under the terms of the GNU public license.

	oicd.c is a driver for parallel port CDrom drives based on the
        90c26 parallel to IDE adapter chip from OnSpec Electronics.

        To compile the driver, ensure that /usr/include/linux and
        /usr/include/asm are links to the correct include files for 
        the target system. Then compile the driver with 

                cc -D__KERNEL__ -DMODULE -Wall -O2 -c oicd.c

        Alternatively, if you are using MODVERSIONS, do

                cc -D__KERNEL__ -DMODULE -DMODVERSIONS \
                   -include /usr/src/linux/include/linux/modversions.h \
                   -Wall -O2 -c oicd.c

        You must then load it with insmod.

	Before attempting to access the new driver, you will need to
        create a new device special file.  The following commands will
	do that for you:

		mknod /dev/oicd b 63 0
		chown root:disk /dev/oicd
		chmod 660 /dev/oicd

	Afterward, you can mount a disk in the usual way:

		mount -t iso9660 /dev/oicd /cdrom

	(assuming you have made a directory /cdrom to use as a
        mount point).

        (The number '63' above corresponds to the #define OI_MAJOR 63
        in the source below.  If it is necessary to change it, you
        must be sure that all uses of the device number are consistent.)

        The driver will search for a 90c26 device on the available
        standard parallel ports.  The driver supports unidirectional
        and bidirectional parallel ports as well as some EPP ports.

        It is not currently possible to connect a printer to a chained
        printer port and expect Linux to use both devices at once.

        Keep an eye on http://www.torque.net/onspec.html for news and
        other information about the driver.  If you have any problems
        with this driver, please send me, grant@torque.net, some mail 
        directly before posting into the newsgroups or mailing lists.

*/

#define	OI_VERSION	"0.02" 

#define	OI_BASE		0	/* set to 0 for autoprobe */
#define OI_HOLD         1       /* micro-second port settling time */
#define OI_DEBUG	0

#include <linux/version.h>
#include <linux/module.h>

#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/fs.h>
#include <linux/major.h>
#include <linux/kernel.h>
#include <linux/tqueue.h>
#include <linux/delay.h>
#include <linux/cdrom.h>
#include <linux/ioport.h>

#include <asm/system.h>
#include <asm/io.h>
#include <asm/segment.h>   /* change to uaccess.h for 2.1 kernels ... */

#ifndef OICD_MAJOR
#define OICD_MAJOR   63
#endif

#define MAJOR_NR OICD_MAJOR

/* set up defines for blk.h,  why don't all drivers do it this way ? */

#define DEVICE_NAME "90c26 CDrom"
#define DEVICE_REQUEST do_oi_request
#define DEVICE_NR(device) (MINOR(device))
#define DEVICE_ON(device)
#define DEVICE_OFF(device)

#include <linux/blk.h>

#define OI_RETRIES	     5
#define OI_TMO		   600		/* timeout in jiffies */
#define OI_DELAY            50          /* spin delay in uS */

#define OI_SPIN		(10000/OI_DELAY)*OI_TMO

int oicd_init(void);
void oicd_setup(char * str, int * ints);
void cleanup_module( void );

static int 	oi_open(struct inode *inode, struct file *file);
static void 	do_oi_request(void);
static void 	do_oi_read(void);
static int 	oi_ioctl(struct inode *inode,struct file *file,
                         unsigned int cmd, unsigned long arg);
static void 	oi_release (struct inode *inode, struct file *file);

static int 	oi_detect(void);
static void     oi_lock(void);
static void     oi_unlock(void);
static void     oi_eject(void);
static void     oi_interrupt(void *data);

static int	oi_base = OI_BASE;
static int      oi_mode = 0;
static int	oi_hold = OI_HOLD;
static int      oi_debug = OI_DEBUG;;

static int      oi_retries;

static int oi_access = 0;		/* count of active opens ... */
static int oi_busy = 0;			/* request being processed ? */
static int oi_timeout;                  /* "interrupt" loop limiter */
static int oi_sector;			/* address of next requested sector */
static int oi_count;			/* number of blocks still to do */
static char * oi_buf;			/* buffer for request in progress */
static char oi_buffer[2048];		/* raw block buffer */
static int oi_bufblk = -1;		/* block in buffer, in CD units,
					   -1 for nothing there */

static struct tq_struct oi_tq = {0,0,oi_interrupt,NULL}; 

/* kernel glue structures */

static struct file_operations oi_fops = {
	NULL,			/* lseek - default */
	block_read,		/* read - general block-dev read */
	block_write,		/* write - general block-dev write */
	NULL,			/* readdir - bad */
	NULL,			/* select */
	oi_ioctl,		/* ioctl */
	NULL,			/* mmap */
	oi_open,		/* open */
	oi_release,		/* release */
	block_fsync,		/* fsync */
	NULL,			/* fasync */
	NULL,                   /* media change ? */
	NULL			/* revalidate new media */
};

int oicd_init (void)	/* preliminary initialisation */

{       if (oi_detect()) return -1;

	if (register_blkdev(MAJOR_NR,"oicd",&oi_fops)) {
		printk("oicd: unable to get major number %d\n",MAJOR_NR);
		return -1;
	}
	blk_dev[MAJOR_NR].request_fn = DEVICE_REQUEST;
	read_ahead[MAJOR_NR] = 8;	/* 8 sector (4kB) read ahead */

	return 0;
}

static int oi_open (struct inode *inode, struct file *file)

{	
	if (file->f_mode & 2) return -EROFS;  /* wants to write ? */

	MOD_INC_USE_COUNT;

	oi_lock();

	oi_access++;
	return 0;
}

static void do_oi_request (void)

{       if (oi_busy) return;
        while (1) {
	    if ((!CURRENT) || (CURRENT->rq_status == RQ_INACTIVE)) return;
	    INIT_REQUEST;
	    if (CURRENT->cmd == READ) {
	        oi_sector = CURRENT->sector;
	        oi_count = CURRENT->nr_sectors;
	        oi_buf = CURRENT->buffer;
	        do_oi_read();
		if (oi_busy) return;
	    } 
	    else end_request(0);
	}
}

static int oi_ioctl(struct inode *inode,struct file *file,
		    unsigned int cmd, unsigned long arg)

/* we currently support only the EJECT ioctl. */

{	switch (cmd) {
            case CDROMEJECT: if (oi_access == 1) {
				oi_eject();
				return 0;
			     }
	    default:
	        return -EINVAL;
	}
}

static void oi_release (struct inode *inode, struct file *file)

{	kdev_t	devp;

	oi_access--;
	if (!oi_access) { 
	    devp = inode->i_rdev;
	    fsync_dev(devp);
	    invalidate_inodes(devp);
	    invalidate_buffers(devp);
	    oi_unlock();
	}
	MOD_DEC_USE_COUNT;
}

#ifdef MODULE

/* Glue for modules ... */

int	init_module(void)

{	int	err;
	long    flags;

	save_flags(flags);
	cli();

	err = oicd_init();

	restore_flags(flags);
	return err;
}

void	cleanup_module(void)

{	long flags;
	int  range;

	if (oi_mode == 2) range = 8; else range = 3;

	save_flags(flags);
	cli();
	unregister_blkdev(MAJOR_NR,"oicd");
	release_region(oi_base,range);
	restore_flags(flags);
}

#else 

/* oicd_setup:  process lilo command parameters ...

   syntax:	oicd=base[,mode[,hold]]
*/

void    oicd_setup(char *str, int *ints)

{       if (ints[0] > 0) oi_base = ints[1];
        if (ints[0] > 1) oi_mode = ints[2];
        if (ints[0] > 2) oi_hold = ints[3];
} 

#endif

/*------------------------------------------------------------------*/

/* Now the actual hardware interface */

static inline void    out_p( short port, char byte)

{       outb(byte,oi_base+port);
        udelay(oi_hold);
}

static inline int     in_p( short port)

{       udelay(oi_hold);
        return (inb(oi_base+port) & 0xff);
}

/* mode codes:  0  nybble reads, 8-bit writes
                1  8-bit reads and writes
                2  8-bit EPP mode
*/

#define w0(byte)  out_p(0,byte)
#define w2(byte)  out_p(2,byte)
#define r0()      in_p(0)
#define r1()      in_p(1)
#define r2()      in_p(2) 
#define w3(byte)  out_p(3,byte)
#define w4(byte)  out_p(4,byte)
#define r4()      in_p(4)

#define j44(a,b)  (((a>>4)&0x0f)|(b&0xf0))

#define P1      w2(5);w2(0xd);w2(5);w2(0xd);w2(5);w2(4);
#define P2      w2(5);w2(7);w2(5);w2(4);

/* To access the control registers add 0x80 to the regr number */

static int oi_read_regr( char regr )

{       int     a, b, r;

        if (regr & 0x80) r = ((regr&7)<<2) + 2; else r = ((regr&7)<<2) + 1;

        switch (oi_mode)  {

        case 0: w0(1); P1; w0(r); P2; w0(0); P1; 
                w2(6); a = r1(); w2(4);
                w2(6); b = r1(); w2(4);
                w2(6); w2(4); w2(6); w2(4);
                return j44(a,b);

        case 1: w0(1); P1; w0(r); P2; w0(0); P1;
                w2(0x26); a = r0(); w2(4); w2(0x26); w2(4);
                return a;

        case 2: w3(1); w3(1); w2(5); w4(r); w2(4);
                w3(0); w3(0); w2(0x24); a = r4(); w2(4);
                w2(0x24); r4(); w2(4);
                return a;

        }
        return -1;
}       

static void oi_write_regr( char regr, char val )

{       int  r;

        if (regr & 0x80) r = ((regr&7)<<2) + 2; else r = ((regr&7)<<2) + 1;

        switch (oi_mode)  {

        case 0:
        case 1: w0(1); P1; w0(r); P2; w0(0); P1;
                w0(val); P2; w0(val); P2;
                break;

        case 2: w3(1); w3(1); w2(5); w4(r); w2(4);
                w3(0); w3(0); 
                w2(5); w4(val); w2(4);
                w2(5); w4(val); w2(4);
                break;
        }
}

static void oi_write_cmd( char * buf, int len )

{       int     k;

        switch (oi_mode) {

        case 0: 
        case 1: w0(1); P1; w0(1); P2; 
                w0(2); P1; w0(0x18+oi_mode); P2; w0(0); P1;
                udelay(10);
                for (k=0;k<len/2;k++) {
                        w2(5); w0(buf[2*k]); 
                        w2(7); w0(buf[2*k+1]);
                }
                w2(5); w2(4);
                w0(2); P1; w0(8+oi_mode); P2;
                break;

        case 2: w3(1); w3(1); w2(5); w4(1); w2(4);
                w3(0); w3(0); w2(0xc5);
                udelay(10);
                for (k=0;k<len;k++) w4(buf[k]);
                w2(0xc4);
                break;

        }
}

static void oi_read_data( char * buf, int len )

{       int     k, a, b;

        switch (oi_mode) {

        case 0: w0(1); P1; w0(1); P2; w0(2); P1; w0(0x18); P2; w0(0); P1;
                udelay(10);
                for (k=0;k<len;k++) {
                        w2(6); a = r1();
                        w2(4); b = r1();
                        buf[k] = j44(a,b);
                }
                w0(2); P1; w0(8); P2; 
                break;

        case 1: w0(1); P1; w0(1); P2; w0(2); P1; w0(0x19); P2; w0(0); P1;
                udelay(10);
                for (k=0;k<len/2;k++) {
                        w2(0x26); buf[2*k] = r0(); 
                        w2(0x24); buf[2*k+1] = r0();
                }
                w0(2); P1; w0(9); P2;
                break;

        case 2: w3(1); w3(1); w2(5); w4(1); w2(4);
                w3(0); w3(0); w2(0x24);
                udelay(10);
                for (k=0;k<len;k++) buf[k] = r4();
                w2(4);
                break;

        }
}

#define  CCP(x)  w0(0xff);w0(0xaa);w0(0x55);w0(0);w0(0xff);\
                 w0(0x87);w0(0x78);w0(x);w2(4);

static void oi_connect ( void  )

{       int     x;

        CCP(0x20);
        w2(0xcd); w2(0xcc); w0(0xff);
        x = 8; if (oi_mode) x = 9;
        w0(2); P1; w0(8); P2;
        w0(2); P1; w0(x); P2;
}

static void oi_disconnect ( void )

{       if (oi_mode == 2) { w3(4); w3(4); w3(4); w3(4); }
          else { w0(4); P1; w0(4); P1; }
        CCP(0x30);
        w2(0xcd); w2(0xcc); w0(0xff);
} 

static int oi_test_proto( void )

{       int     k, e;

        e = 0;
        oi_connect();
        oi_write_regr(6,0xa0);
        for (k=0;k<256;k++) {
                oi_write_regr(2,k^0xaa);
                oi_write_regr(3,k^0x55);
                if (oi_read_regr(2) != (k^0xaa)) e++;
                }
        oi_disconnect();

        if (oi_debug)  {
            printk("oi: port 0x%x, mode %d, register test %d errors\n",
                    oi_base,oi_mode,e);

            oi_connect();
            oi_write_regr(6,0xa0);  udelay(30);
            if (oi_read_regr(0x86) != oi_read_regr(7)) 
               printk("oi: oops, ALT_STATUS != STATUS\n");
            oi_disconnect();

        }

        return e;
}

/*-------------------------------------------------------------------*/

#define	IDE_DRQ  	0x08
#define IDE_READY	0x40
#define IDE_BUSY 	0x80

static int oi_wait( int st, char * fun, char * msg1, char * msg2 )

{	int j, r, e;

	j = 0;
	while ((!((oi_read_regr(0x86)&(st|IDE_BUSY))==st))&&(j++<OI_SPIN)) 
		udelay(OI_DELAY);
        r = oi_read_regr(7);
	e = oi_read_regr(1);
	if ((j >= OI_SPIN) || (r & 1) || ((r & st) != st)) {
	    if (j >= OI_SPIN) e |= 0x100;
	    if (fun) printk("oicd: %s%s: Error %s, status: %x error: %x\n",
			  fun,msg1,msg2,r,e);
	    return -1;
	}
	return 0;
}

static int oi_req_sense( char * fun, char * msg )

{	char	rs_cmd[12] = { 0x03,0,0,0,18,0,0,0,0,0,0,0 };
	int n, r;

        oi_write_regr(4,18);
        oi_write_regr(5,0);
        oi_write_regr(6,0xa0);          /* drive 0 */
        oi_write_regr(7,0xa0);          /* ATAPI packet command */
	if (oi_wait(IDE_DRQ,fun,msg,"request sense command")) return -1;
	oi_write_cmd(rs_cmd,12);
	if (!(r=oi_wait(IDE_READY|IDE_DRQ,
			fun,msg,"getting request sense data"))) {
            if (oi_read_regr(2) == 2) {
                n = (oi_read_regr(4) + 256*oi_read_regr(5));
		oi_wait(IDE_DRQ,NULL,NULL,NULL);
                oi_read_data(oi_buffer,n);
                r=oi_wait(IDE_READY,fun,msg,"on request sense done");
                if (!r) if (fun)
	          printk("oicd: %s%s: sense key: %x, ASC: %x, ASQ: %x\n",
	            fun, msg, oi_buffer[2]&0xf, oi_buffer[12], oi_buffer[13]);
	    }
	}
	return r;
}

static int oi_command( char * cmd, int dlen, char * fun )

{	int	r;

	oi_connect();

	oi_write_regr(4,dlen % 256);
	oi_write_regr(5,dlen / 256);
	oi_write_regr(6,0xa0);		/* drive 0 */
	oi_write_regr(7,0xa0);		/* ATAPI packet command */

        if (oi_read_regr(7)&1) {
            if (oi_req_sense(fun," before command")) { 
                oi_disconnect();
                return -1;
                }
            oi_write_regr(4,dlen % 256);
            oi_write_regr(5,dlen / 256);
            oi_write_regr(6,0xa0);          /* drive 0 */
            oi_write_regr(7,0xa0);          /* ATAPI packet command */
	}

	if ((r=oi_wait(IDE_DRQ,fun,"","sending  command"))) {
	   oi_disconnect();
	   return r;
	}

	oi_write_cmd(cmd,12);
	
	udelay(200);

	return 0;
}

static int oi_completion( char * fun )

{	int r, n;

	if (!(r=oi_wait(IDE_READY,fun,"","command completion"))) {
	    if (oi_read_regr(2) == 2) {
	        n = (oi_read_regr(4) + 256*oi_read_regr(5));
		if (!(r=oi_wait(IDE_DRQ,fun,"","Waiting for DRQ"))) {
	            oi_read_data(oi_buffer,n);
		    udelay(500);
	            r=oi_wait(0,fun,"","waiting after data");
		}
	    } 
	}
	if (r) r = oi_req_sense(fun," after command");

	oi_disconnect(); 
	return r;
}

static int oi_atapi( char * cmd, int dlen, char * fun )

{	int r;

	if (!(r=oi_command(cmd,dlen,fun)))
	  r = oi_completion(fun);
	return r;
}

static void oi_lock(void)

{	char	lo_cmd[12] = { 0x1e,0,0,0,1,0,0,0,0,0,0,0 };
	char	cl_cmd[12] = { 0x1b,0,0,0,3,0,0,0,0,0,0,0 };

	oi_atapi(cl_cmd,0,NULL);  
	oi_atapi(cl_cmd,0,NULL);  
	oi_atapi(cl_cmd,0,NULL);  
	oi_atapi(cl_cmd,0,NULL);  
	oi_atapi(cl_cmd,0,"close door");

        oi_atapi(lo_cmd,0,NULL);
        oi_atapi(lo_cmd,0,"lock door");
}

static void oi_unlock( void)

{	char	un_cmd[12] = { 0x1e,0,0,0,0,0,0,0,0,0,0,0 };

	oi_atapi(un_cmd,0,"unlock door");
}

static void oi_eject( void)

{	char	ej_cmd[12] = { 0x1b,0,0,0,2,0,0,0,0,0,0,0 };

	oi_unlock();
	oi_atapi(ej_cmd,0,"eject");
}

static int oi_reset( void )

/* the ATAPI standard actually specifies the contents of all 7 registers
   after a reset, but the specification is ambiguous concerning the last
   two bytes, and different drives interpret the standard differently.
*/

{	int	i, flg;
	int	expect[5] = {1,1,1,0x14,0xeb};
	long	flags;

	oi_connect();
	oi_write_regr(6,0xa0);
	oi_write_regr(7,8);

	save_flags(flags);
	sti();
	udelay(500000);  	/* delay 0.5 seconds */
	restore_flags(flags);

	flg = 1;
	for(i=0;i<5;i++) flg &= (oi_read_regr(i+1) == expect[i]);

	if (oi_debug) {
		printk("oicd RESET: ");
		for (i=0;i<5;i++) printk("%3x",oi_read_regr(i+1));
		printk("\n");
	}
	
	oi_disconnect();
	return flg-1;	
}

static int oi_identify( char * id )

{	int k;
	char   id_cmd[12] = {0x12,0,0,0,36,0,0,0,0,0,0,0};

	oi_bufblk = -1;
        if (oi_atapi(id_cmd,36,"identify")) return -1;
	for (k=0;k<16;k++) id[k] = oi_buffer[16+k];
	id[16] = 0;
	return 0;
}

static void oi_probe ( int port, int max_mode )

{	int	best, range;

	range = 3;
	if (max_mode == 2) range = 8;
	if (check_region(port,range)) {
		max_mode = 1;
		range = 3;
		if (check_region(port,range)) return;
	}

	oi_base = port;  best = -1;

	if (oi_debug) printk("oicd PROBE: port 0x%x, max_mode %d\n",
			port, max_mode);

	for (oi_mode=0;oi_mode<=max_mode;oi_mode++)
	   if (!oi_test_proto()) best = oi_mode;

	oi_mode = best;
	if (oi_mode < 0) oi_base = 0;
}
 
static int oi_detect( void )

{	char    id[18];
	int	range;
        char    *mode_str[3] = {"4-bit","8-bit","EPP"};

	if (!oi_base) oi_probe(0x278,2);
	if (!oi_base) oi_probe(0x378,2);
	if (!oi_base) oi_probe(0x3bc,1);

	if (!oi_base) {
		printk("oicd: Autoprobe failed\n");
		return -1;
	}

	if (oi_mode == 2) range = 8; else range = 3;
	if (check_region(oi_base,range)) {
		printk("oicd: Ports at 0x%x are not available\n",oi_base);
		return -1;
	}

	if (oi_test_proto()) {
		printk("oicd: Can't access 90c26 adapter in %s mode\n",
			mode_str[oi_mode]);
		return -1;
	}

        printk("oicd %s: OnSpec 90c26 adapter using port 0x%x in %s mode\n",
                OI_VERSION,oi_base,mode_str[oi_mode]);

	if (oi_reset()) {
		printk("oicd: Failed to reset CD drive\n");
		return -1;
	}

	if (oi_identify(id)) {
		printk("oicd: ATAPI inquiry failed\n");
		return -1;
	}

	if (oi_mode == 2) range = 8; else range = 3;	
	request_region(oi_base,range,"oicd");

	printk("oicd: Found %s\n",id);

	return 0;
}

static void oi_transfer( void )

{	int	k, o;

	while (oi_count && (oi_sector/4 == oi_bufblk)) {
		o = (oi_sector % 4) * 512;
		for(k=0;k<512;k++) oi_buf[k] = oi_buffer[o+k];
		oi_count--;
		oi_buf += 512;
		oi_sector++;
	}
}

static void oi_start( void )

{	int	b, i;
	char	rd_cmd[12] = {0xa8,0,0,0,0,0,0,0,0,1,0,0};

	oi_bufblk = oi_sector / 4;
        b = oi_bufblk;
	for(i=0;i<4;i++) { 
	   rd_cmd[5-i] = b & 0xff;
	   b = b >> 8;
	}
	if (oi_command(rd_cmd,2048,"read block")) {
		oi_bufblk = -1; 
		oi_busy = 0;
		cli();
		end_request(0);
		do_oi_request();
		return;
	}
	oi_timeout = jiffies + OI_TMO;
	queue_task(&oi_tq,&tq_scheduler); 
}

static void do_oi_read( void )

{	oi_busy = 1;
	oi_retries = 0;
	oi_transfer();
	if (!oi_count) {
		end_request(1);
		oi_busy = 0;
		return;
	}
	sti();

	oi_start();
}

static void oi_interrupt( void *data)

{	if ( (oi_read_regr(0x86) & 0xc8) != 0x48 ) {
		if (jiffies > oi_timeout) {
			oi_bufblk = -1;
			oi_busy = 0;
			end_request(0);
			do_oi_request();
			return;
		}
		queue_task(&oi_tq,&tq_scheduler); 
		return;
	}
	sti();
	if (oi_completion("read block")) {
                if (oi_retries < OI_RETRIES) {
                        udelay(100000);
                        oi_retries++;
                        
			oi_start();

                        return;
                        }
		cli();
		oi_busy = 0;
		oi_bufblk = -1;
		end_request(0);
		do_oi_request();
		return;
	}
	do_oi_read();
	do_oi_request();
}
	
/* end of oicd.c */
