/* 
	pwcd.c	(c) 1996  Grant R. Guenther <grant@torque.net>
		          Under the terms of the GNU public license

	pwcd is a driver for the Freecom "Power CD", a parallel port
        CDrom device available at popular German discount stores. 


	The Power CD unit consists of a Mitsumi ATAPI drive, together
        with a custom chip that implements Freecom's IDE over parallel
        protocol.  It seems that this custom chip is based on a truth
        table that must be loaded by the driver during the initialisation
	of the device.  The DOS driver POWER-CD.SYS that comes with the
        drive does this. Since it would violate copyrights to include
        the initialisation sequence as part of this driver, the user
        must arrange for the drive to be initialised from DOS prior to
        booting Linux.

	Alternatively, a small tool, pw-loader, is available that will 
	load the truth table from a copy of POWER-CD.SYS.  In this 
	arrangement, you would run pw-loader from Linux, and then insmod
        the pwcd driver.  Check http://www.torque.net/pwcd.html for 
        this program and other information about the driver.


	To compile the driver make sure that /usr/include/linux and 
	/usr/include/asm are links to the correct include files for 
	the target system.  Compile the driver with

                cc -D__KERNEL__ -DMODULE -O2 -c pwcd.c

        If you built your kernel with MODVERSIONS, you should use this
        command instead:

                cc -D__KERNEL__ -DMODULE -DMODVERSIONS \
			-include /usr/include/linux/modversions.h \
			-O2 -c pwcd.c

	Before attempting to access the new driver, you will need to
        create a new device special file.  The following commands will
	do that for you:

		mknod /dev/pwcd b 62 0
		chown root:disk /dev/pwcd
		chmod 660 /dev/pwcd

	(Please note that '62' is a temporary device number.  If 
	an official number is allocated for this device, or you 
	want to use a different number for some reason, change the 
	#define PWCD_MAJOR  below and change the same number in the 
	command above.)

	Afterward, you can mount a disk in the usual way:

		mount -t iso9660 /dev/pwcd /cdrom

	(assuming you have made a directory /cdrom to use as a
        mount point).

	The driver will attempt to detect which parallel port your
        backpack is connected to, and the best mode to use.

	EPP and ECP modes are not supported, nor are interrupts or
        chained printing.

	Keep an eye on http://www.torque.net/pwcd.html for news and
        other information about the driver.  If you have any problems
        with this driver, please send me, grant@torque.net, some mail.

*/

#define	PW_VERSION	"0.01" 

#define	PW_BASE		0	/* set to 0 for autoprobe */
#define PW_REP		2

#include <linux/version.h>
#include <linux/module.h>

#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/fs.h>
#include <linux/major.h>
#include <linux/kernel.h>
#include <linux/tqueue.h>
#include <linux/delay.h>
#include <linux/cdrom.h>
#include <linux/ioport.h>

#include <asm/system.h>
#include <asm/io.h>
#include <asm/segment.h>

#ifndef PWCD_MAJOR
#define PWCD_MAJOR   62	
#endif

#define MAJOR_NR PWCD_MAJOR

/* set up defines for blk.h,  why don't all drivers do it this way ? */

#define DEVICE_NAME "Power CD"
#define DEVICE_REQUEST do_pw_request
#define DEVICE_NR(device) (MINOR(device))
#define DEVICE_ON(device)
#define DEVICE_OFF(device)

#include <linux/blk.h>

#define PW_TMO		   400		/* timeout in jiffies */
#define PW_DELAY            50          /* spin delay in uS */

#define PW_SPIN		(10000/PW_DELAY)*PW_TMO

int pwcd_init(void);
void pwcd_setup(char * str, int * ints);
void cleanup_module( void );

static int 	pw_open(struct inode *inode, struct file *file);
static void 	do_pw_request(void);
static void 	do_pw_read(void);
static int 	pw_ioctl(struct inode *inode,struct file *file,
                         unsigned int cmd, unsigned long arg);
static void 	pw_release (struct inode *inode, struct file *file);

static int 	pw_detect(void);
static int      pw_lock(void);
static void     pw_unlock(void);
static void     pw_eject(void);
static void     pw_interrupt(void *data);

static int	pw_base = PW_BASE;
static int	pw_rep = PW_REP;
static int	pw_nybble = 0;	/* force nybble mode ? */

/* pw_mode:  0 for nybble mode, 1 for byte mode.  Used for block reads only */

static int pw_access = 0;		/* count of active opens ... */
static int pw_mode = 1;			/* 4- or 8-bit mode */
static int pw_busy = 0;			/* request being processed ? */
static int pw_timeout;                  /* "interrupt" loop limiter */
static int pw_sector;			/* address of next requested sector */
static int pw_count;			/* number of blocks still to do */
static char * pw_buf;			/* buffer for request in progress */
static char pw_buffer[2048];		/* raw block buffer */
static int pw_bufblk = -1;		/* block in buffer, in CD units,
					   -1 for nothing there */

static struct tq_struct pw_tq = {0,0,pw_interrupt,NULL}; 

/* kernel glue structures */

static struct file_operations pw_fops = {
	NULL,			/* lseek - default */
	block_read,		/* read - general block-dev read */
	block_write,		/* write - general block-dev write */
	NULL,			/* readdir - bad */
	NULL,			/* select */
	pw_ioctl,		/* ioctl */
	NULL,			/* mmap */
	pw_open,		/* open */
	pw_release,		/* release */
	block_fsync,		/* fsync */
	NULL,			/* fasync */
	NULL,                   /* media change ? */
	NULL			/* revalidate new media */
};


int pwcd_init (void)	/* preliminary initialisation */

{       if (pw_detect()) return -1;

	if (register_blkdev(MAJOR_NR,"pwcd",&pw_fops)) {
		printk("pwcd: unable to get major number %d\n",MAJOR_NR);
		return -1;
	}
	blk_dev[MAJOR_NR].request_fn = DEVICE_REQUEST;
	read_ahead[MAJOR_NR] = 8;	/* 8 sector (4kB) read ahead */

	return 0;
}

static int pw_open (struct inode *inode, struct file *file)

{	
	if (file->f_mode & 2) return -EROFS;  /* wants to write ? */

	MOD_INC_USE_COUNT;

	if (!pw_access)
	   if (pw_lock()) {
		MOD_DEC_USE_COUNT;
		return -ENXIO;
	   }
	pw_access++;
	return 0;
}

static void do_pw_request (void)

{       if (pw_busy) return;
        while (1) {
	    if ((!CURRENT) || (CURRENT->rq_status == RQ_INACTIVE)) return;
	    INIT_REQUEST;
	    if (CURRENT->cmd == READ) {
	        pw_sector = CURRENT->sector;
	        pw_count = CURRENT->nr_sectors;
	        pw_buf = CURRENT->buffer;
	        do_pw_read();
		if (pw_busy) return;
	    } 
	    else end_request(0);
	}
}

static int pw_ioctl(struct inode *inode,struct file *file,
		    unsigned int cmd, unsigned long arg)

/* we currently support only the EJECT ioctl. */

{	switch (cmd) {
            case CDROMEJECT: if (pw_access == 1) {
				pw_eject();
				return 0;
			     }
	    default:
	        return -EINVAL;
	}
}

static void pw_release (struct inode *inode, struct file *file)

{	kdev_t	devp;

	pw_access--;
	if (!pw_access) { 
	    devp = inode->i_rdev;
	    fsync_dev(devp);
	    invalidate_inodes(devp);
	    invalidate_buffers(devp);
	    pw_unlock();
	}
	MOD_DEC_USE_COUNT;
}

#ifdef MODULE

/* Glue for modules ... */

int	init_module(void)

{	int	err;
	long    flags;

	save_flags(flags);
	cli();

	err = pwcd_init();

	restore_flags(flags);
	return err;
}

void	cleanup_module(void)

{	long flags;

	save_flags(flags);
	cli();
	unregister_blkdev(MAJOR_NR,"pwcd");
	release_region(pw_base,3);
	restore_flags(flags);
}

#else 

/* pwcd_setup:  process lilo command parameters ...

   syntax:	pwcd=base[,nybble[,rep]]
*/

void    pwcd_setup(char *str, int *ints)

{       if (ints[0] > 0) pw_base = ints[1];
        if (ints[0] > 1) pw_nybble = ints[2];
        if (ints[0] > 2) pw_rep = ints[3];
} 

#endif

static 	void    out_p( short port, char byte)

{       int i;

	for(i=0;i<pw_rep;i++) outb(byte,pw_base+port);
}

static	int     in_p( short port)

{       int i;
	char c;

	for(i=0;i<pw_rep;i++) c=inb(pw_base+port);
	c=inb(pw_base+port);
	return c & 0xff;
}

#define w0(byte)  		out_p(0,byte)
#define r0()      		(in_p(0) & 0xff)
#define w1(byte)  		out_p(1,byte)
#define r1()      		(in_p(1) & 0xff)
#define w2(byte)  		out_p(2,byte)

#define cec4	w2(0xc);w2(0xe);w2(0xe);w2(0xc);w2(4);w2(4);w2(4);
#define s20	w0(0x20);cec4;


static int pw_read_regr( int regr )

{	int	h,l;

	w2(4);
	w0(regr); cec4;
	w2(6); l = r1();
	w2(4); h = r1();
	w2(4); s20;

	return (((l & 0xf0)>>4) + (h & 0xf0));
}

static void pw_read_block( int regr, int len, char * buf )

{	int	h, l, k, ph;

	w2(4);
	ph = 2; w0(regr + pw_mode*0xc0); cec4;
	if (pw_mode) w0(0xff);

	for (k=0;k<len;k++) {
	   if (pw_mode) {
		w2(0xa4 + ph); 
		buf[k] = r0();
		ph = 2 - ph;
	   } else {
		w2(6); l = r1();
		w2(4); h = r1();
		buf[k] = (((l & 0xf0)>>4) + (h & 0xf0));
	   }
	}
	if (pw_mode) { w2(0xac); w2(0xa4); }
	w2(4); s20;
}

static void pw_write_regr( int regr, int v )

{	w2(4); w0(regr); cec4; w0(v);
	w2(5);w2(7);w2(5);w2(4); s20;
}

static void pw_write_block( int regr, int len, char * buf )
 
{	int	k;

	w2(4); w0(regr); cec4; w2(5);
        for (k=0;k<len;k++) {
		w0(buf[k]);
		w2(7);w2(5);
	}
	w2(4); s20;
}

static int pw_test_adapter( void )

{	int	k, r;

	pw_read_block(0x10,2048,pw_buffer);
	r = 0;
	for (k=0;k<128;k++) r |= (pw_buffer[k] != k);
	return r;
}

static int pw_wait_drq( char * lab, char * fun )

{	int j, r, e;

	j = 0;

	while (1)  {
        	r = pw_read_regr(0xf);
		e = pw_read_regr(0x9);
		if ((r & 9) || (j++ >= PW_SPIN)) break;
		udelay(PW_DELAY);
	}

	if ((j >= PW_SPIN) || (r & 1)) {
	    if (lab && fun) 
		   printk("pwcd: %s (%s): %s status: %x error: %x\n",
			  lab,fun,(j >= PW_SPIN)?"timeout, ":"",r,e);
	    return -1;
	}
	return 0;
}

static int pw_wait( char * lab, char * fun )

{	int j, r, e;

	j = 0;
	while ((pw_read_regr(0xf) & 0x80) && (j++ < PW_SPIN)) udelay(PW_DELAY);
        r = pw_read_regr(0xf);
	e = pw_read_regr(0x9);
	if ((j >= PW_SPIN) || (r & 1)) {
	    if (lab && fun) 
		   printk("pwcd: %s (%s): %s status: %x error: %x\n",
			  lab,fun,(j >= PW_SPIN)?"timeout, ":"",r,e);
	    return -1;
	}
	return 0;
}

static int pw_command( char * cmd, int dlen, char * fun )

{	int	r;

	pw_write_regr(0xc,dlen % 256);
	pw_write_regr(0xd,dlen / 256);
	pw_write_regr(0xe,0xa0);		/* drive 0 */
	pw_write_regr(0xf,0xa0);		/* ATAPI packet command */
	if ((r=pw_wait_drq("pw_command",fun))) return r;
	pw_write_block(8,12,cmd);
	return 0;
}

static int pw_completion( char * fun )

{	int r, n;

	if (!(r=pw_wait("pw_completion",fun))) {
	    if (pw_read_regr(0xa) == 2) {
	        n = (pw_read_regr(0xc) + 256*pw_read_regr(0xd));
	        pw_read_block(8,n,pw_buffer);
	        r=pw_wait("transfer done",fun);
	    }
	}
	return r;
}

static int pw_atapi( char * cmd, int dlen, char * fun )

{	int r;

	if (!(r=pw_command(cmd,dlen,fun)))
	  r = pw_completion(fun);
	return r;
}

static int pw_req_sense( char * msg )

{	char	rs_cmd[12] = { 0x03,0,0,0,18,0,0,0,0,0,0,0 };
	int r;

	r = pw_atapi(rs_cmd,18,"request sense");
	if (msg) printk("pwcd: %s:  sense key: %x, ASC: %x, ASQ: %x\n",msg,
			 pw_buffer[2]&0xf, pw_buffer[12], pw_buffer[13]);
	return r;
}

static int pw_lock(void)

{	char	lo_cmd[12] = { 0x1e,0,0,0,1,0,0,0,0,0,0,0 };
	char	cl_cmd[12] = { 0x1b,0,0,0,3,0,0,0,0,0,0,0 };

	pw_atapi(cl_cmd,0,"close door");
        if (pw_req_sense(NULL)) return -1;  /* check for disk */
	pw_atapi(lo_cmd,0,NULL);
	pw_req_sense(NULL);  /* in case there was a media change */
	pw_atapi(lo_cmd,0,"lock door");
	return 0;
}

static void pw_unlock( void)

{	char	un_cmd[12] = { 0x1e,0,0,0,0,0,0,0,0,0,0,0 };

	pw_atapi(un_cmd,0,"unlock door");
}

static void pw_eject( void)

{	char	ej_cmd[12] = { 0x1b,0,0,0,2,0,0,0,0,0,0,0 };

	pw_unlock();
	pw_atapi(ej_cmd,0,"eject");
}

static int pw_reset( void )

{	int	i, flg;
	int	expect[5] = {1,1,1,0x14,0xeb};
	long	flags;

	pw_write_regr(0xe,0xa0);
	pw_write_regr(0xf,8);

	save_flags(flags);
	sti();
	udelay(500000);  	/* delay 0.5 seconds */
	restore_flags(flags);

	flg = 1;
	for(i=0;i<5;i++) flg &= (pw_read_regr(i+9) == expect[i]);
	
	return flg-1;	
}

static int pw_identify( char * id )

{	int k;
	char   id_cmd[12] = {0x12,0,0,0,36,0,0,0,0,0,0,0};

	pw_bufblk = -1;
        if (pw_atapi(id_cmd,36,"identify")) return -1;
	for (k=0;k<16;k++) id[k] = pw_buffer[16+k];
	id[16] = 0;
	return 0;
}

static int pw_port_check( void ) 	/* check for 8-bit port */

{	int	r;

        w2(0); 
	w0(0x55); if (r0() != 0x55) return 0;
	w0(0xaa); if (r0() != 0xaa) return 0;
	w2(0x20); w0(0x55); r = r0(); w0(0xaa);
	if (r0() == r) return 1;
	if (r0() == 0xaa) return 0;
	return -1;
}

static int pw_do_detect( int autop )

{	char   id[18];

	if (autop) pw_base = autop;

	if (check_region(pw_base,3)) {
		if (!autop) 
		  printk("pwcd: Ports at 0x%x are not available\n",pw_base);
		return -1;
	}

	pw_mode = pw_port_check();

	if (pw_mode < 0) {
		if (!autop)
		  printk("pwcd: No parallel port at 0x%x\n",pw_base);
		return -1;
	}

	if (pw_nybble) pw_mode = 0;

	if (pw_test_adapter()) {
		if (!autop)
		  printk("pwcd: Couldn't find a Freecom adapter at 0x%x\n",
			  pw_base);
		return -1;
	}

	if (pw_reset()) {
		if (!autop)
		  printk("pwcd: Failed to reset CD drive\n");
		return -1;
	}

	if (pw_identify(id)) {
		if (!autop)
		  printk("pwcd: ATAPI inquiry failed\n");
		return -1;
	}

	request_region(pw_base,3,"pwcd");

	printk("pwcd: Found %s, using port 0x%x in %d-bit mode\n",
	       id,pw_base,4+4*pw_mode);

	return 0;
}

/* If you know about some other weird parallel port base address,
   add it here ....
*/

static int pw_detect( void )

{	if (pw_base) return pw_do_detect(0);
	if (!pw_do_detect(0x378)) return 0;
	if (!pw_do_detect(0x278)) return 0;
	if (!pw_do_detect(0x3bc)) return 0;
	printk("pwcd: Autoprobe failed\n");
	return -1;
}


static void pw_transfer( void )

{	int	k, o;

	while (pw_count && (pw_sector/4 == pw_bufblk)) {
		o = (pw_sector % 4) * 512;
		for(k=0;k<512;k++) pw_buf[k] = pw_buffer[o+k];
		pw_count--;
		pw_buf += 512;
		pw_sector++;
	}
}

static void do_pw_read( void )

{	int	b, i;
	char	rd_cmd[12] = {0xa8,0,0,0,0,0,0,0,0,1,0,0};

	pw_busy = 1;
	pw_transfer();
	if (!pw_count) {
		end_request(1);
		pw_busy = 0;
		return;
	}
	sti();

	pw_bufblk = pw_sector / 4;
        b = pw_bufblk;
	for(i=0;i<4;i++) { 
	   rd_cmd[5-i] = b & 0xff;
	   b = b >> 8;
	}

	if (pw_command(rd_cmd,2048,"read block")) {
		pw_bufblk = -1; 
		pw_busy = 0;
		cli();
		pw_req_sense("send read command");
		end_request(0);
		return;
	}
	pw_timeout = jiffies + PW_TMO;
	queue_task(&pw_tq,&tq_scheduler); 
}

static void pw_interrupt( void *data)

{	if ((pw_read_regr(0xf) & 0x80)) {
		if (jiffies > pw_timeout) {
			pw_bufblk = -1;
			pw_busy = 0;
			pw_req_sense("interrupt timeout");
			end_request(0);
			do_pw_request();
		}
		queue_task(&pw_tq,&tq_scheduler); 
		return;
	}
	sti();
	if (pw_completion("read completion")) {
		cli();
		pw_busy = 0;
		pw_bufblk = -1;
		pw_req_sense("read completion");
		end_request(0);
		do_pw_request();
	}
	do_pw_read();
	do_pw_request();
}
	
/* end of pwcd.c */
