/*
 * Wyse NOR driver for Linux
 *
 * Copyright (C) 2000 Wyse Technology, Inc
 *
 * This driver allows you to read or write the contents of the NOR flash
 * devices on Wyse Winterm family products.  The NOR parts are either 256K or
 * 512K.  This driver and some companion software are used to update the boot
 * block on the Winterms from within Linux.
 *
 * The device file will be /dev/wysenor, and the commands supported are read,
 * write, seek, open, and close.  Because of the nature of the device, open
 * will allocate an internal buffer of 256 or 512 Kb (based upon the device)
 * and will fill it from the device, read will return the appropriate portion
 * of this buffer, write will modify the appropriate portion of this buffer,
 * and close will free up the buffer.  If a write was performed, close will
 * also erase and rewrite the NOR device.
 *
 */

#define WYSENOR_VERSION        "1.0"

#include <linux/module.h>

#define CHECK_DRIVER_INIT() 1

#include <linux/types.h>
#include <linux/errno.h>
#include <linux/miscdevice.h>
#include <linux/malloc.h>
#include <linux/ioport.h>
#include <linux/fcntl.h>
#include <linux/pci.h>

#include <asm/io.h>
#include <asm/system.h>
#include <asm/delay.h>

#define OUTL(port, val)    outl(val, port) /* Mike's code was backwards */

#define FBOOT_SIZE      0x00040000
#define FBOOT_ADDR      0xFFFC0000

#define fb_setdata(off, value)  *((volatile unsigned char *) ((off) + wysenor_vaddr)) = value
#define fb_getdata(off)         *((volatile unsigned char *) ((off) + wysenor_vaddr))

int     fb_pagesize, fb_pagesize_m1;
int     fboot_size = FBOOT_SIZE; /* size of NOR device in unit (256 or 512K) */
int        wysenor_vaddr;

static  u_char  mfg, dev;

#define SST28SF040      0xbf

static int wysenor_open_cnt = 0;   /* #times opened */
static int wysenor_open_mode;      /* special open modes */
static int wysenor_dirty = 0;      /* set to 1 if the buffer is modified */

/* Buffer containing the actual image data which is what we work off of,
 * read on open, written on release.
 */
static unsigned char *wysenor_buffer = NULL;

#define    WYSENOR_WRITE       1   /* opened for writing (exclusive) */
#define    WYSENOR_EXCL        2   /* opened with O_EXCL */

static void
bank_low(void)
{
       unsigned short x;
       pcibios_read_config_word(0, 0x90, 0x90, &x);
       x &= ~0x8000; /* select low bank */
       x  |= 0x80;
       pcibios_write_config_word(0, 0x90, 0x90, x);

       pcibios_read_config_word(0, 0x90, 0x90, &x);
       asm("wbinvd");          // Invalid cache
}

static void
bank_high(void)
{
       unsigned short x;
       pcibios_read_config_word(0, 0x90, 0x90, &x);
       x |= 0x8080; /* select high bank */
       pcibios_write_config_word(0, 0x90, 0x90, x);

       pcibios_read_config_word(0, 0x90, 0x90, &x);
       asm("wbinvd");          // Invalid cache
}

/* The first block of routines are all internal routines used for accessing
 * the hardware.
 */

static int
sst_erase(int sa)
{
   int i;

   fb_setdata(0x5555, 0xAA);
   fb_setdata(0x2AAA, 0x55);
   fb_setdata(0x5555, 0xF0);       // reset
   udelay(20);
   fb_setdata(0x5555, 0xAA);
   fb_setdata(0x2AAA, 0x55);
   fb_setdata(0x5555, 0x20);       // sector erase

   udelay(20);

   fb_setdata(sa, 0xD0);           // execute erase

   for (i = 0; i < 0x20000; ++i) {
       udelay(10);
       if (fb_getdata(sa) == 0xff) {   // last byte
           if (fb_getdata(sa) == 0xff)
               return(0);
       }
   }
   printk("sst_erase: timeout\n");
   return(-1);
}

static void
sst_write(int pa, u_char *buf)
{
   int i, j;

   fb_setdata(0x5555, 0xAA);
   fb_setdata(0x2AAA, 0x55);
   fb_setdata(0x5555, 0xF0);       // reset
   udelay(20);

   for (i = 0; i < fb_pagesize; ++i, ++pa, ++buf) {
       fb_setdata(0x5555, 0xAA);
       fb_setdata(0x2AAA, 0x55);
       fb_setdata(0x5555, 0x10);       // byte program
       fb_setdata(pa, *buf);

       for (j = 0; j < 0x10000; ++j) {
           udelay(10);
           if (fb_getdata(pa) == *buf) {
               if (fb_getdata(pa) == *buf)
                   break;
           }
       }
       if (j >= 0x10000)
           printk( "sst_write: timeout\n");
   }
}

static void
fboot_enable(void)
{
   int i;

   OUTL(0xCF8, 0x80009050);
   i = inl(0xCFC) | 0x20000;
   OUTL(0xCFC, i);
}

static void
fboot_disable(int bank_switched)
{
   int i;

   OUTL(0xCF8, 0x80009050);
   i = inl(0xCFC) & ~0x20000;
   OUTL(0xCFC, i);
}

static int
fboot_read(int offset, u_char *buf, int size)
{
   int flags, count, bank_switched = 0;

   size += fb_pagesize_m1;
   size &= ~fb_pagesize_m1;

   if (fb_pagesize <= 0)
       return(-1);

   save_flags(flags); cli();
   fboot_enable();
   if (fboot_size != FBOOT_SIZE) /* Only bank switch the 512K parts */
       bank_low();

   for (count = 0; size > 0; size -= fb_pagesize) {
       if (offset >= FBOOT_SIZE && bank_switched == 0) {       // Bank switch needed
           udelay(10);
	   if (fboot_size != FBOOT_SIZE)
	   {
	       bank_high();
	       bank_switched = 1;
	   }
           offset = 0;
           udelay(10);
       }

       memcpy(buf, (void *) (wysenor_vaddr + offset), fb_pagesize);
       offset += fb_pagesize;
       buf += fb_pagesize;
       count += fb_pagesize;
   }

   fboot_disable(bank_switched);
   restore_flags(flags);
   return(count);
}

static int
fboot_write(int offset, u_char *buf, int size)
{
   int flags, i, count, bank_switched = 0;

   size += fb_pagesize_m1;
   size &= ~fb_pagesize_m1;

   if (fb_pagesize <= 0)
       return(-1);

   save_flags(flags); cli();
   fboot_enable();
   if (fboot_size != FBOOT_SIZE) /* Only bank switch the 512K parts */
       bank_low();

   for (count = 0; size > 0; size -= fb_pagesize) {
       if (offset >= FBOOT_SIZE && bank_switched == 0) {       // Bank switch needed
           udelay(10);
	   if (fboot_size != FBOOT_SIZE)
	   {
		bank_high();
		bank_switched = 1;
	   }
           offset = 0;
           udelay(10);
       }

       if (memcmp((void *) (wysenor_vaddr + offset), buf, fb_pagesize) == 0) {
           offset += fb_pagesize;
           buf += fb_pagesize;
           goto next;          // Skip unchanged
       }

       if (mfg == SST28SF040) {
           sst_erase(offset);
           sst_write(offset, buf);
           offset += fb_pagesize;
           buf += fb_pagesize;
           goto next;
       }

       fb_setdata(0x5555, 0xAA);
       fb_setdata(0x2AAA, 0x55);
       fb_setdata(0x5555, 0xA0);       // Start writing

       for (i = 0; i < fb_pagesize; i++, offset++, buf++)
           fb_setdata(offset, *buf);   // Copy data

       for (i = 0; ; ++i) {
           udelay(10);
           if (buf[-1] == fb_getdata(offset - 1))  // last byte
               break;
           if (i >= 0x10000) {
               printk( "fboot_write error: wrote 0x%x, "
                           "read 0x%x, offset: 0x%x\n",
                   buf[-1], fb_getdata(offset - 1),
                   offset - 1);
               fboot_disable(bank_switched);
               restore_flags(flags);
               return(count);
           }
       }
next:
       count += fb_pagesize;
   }

   fboot_disable(bank_switched);
   restore_flags(flags);
   return(count);
}

/*
 * The are the file operation function for user access to /dev/wysenor
 */

static int wysenor_llseek(struct inode * inode, struct file * file,
                    off_t offset, int origin)
{
   switch( origin ) {
     case 0:
       /* nothing to do */
       break;
     case 1:
       offset += file->f_pos;
       break;
     case 2:
       offset += fboot_size; /* back up from end */
       break;
   }
   return( (offset >= 0) ? (file->f_pos = offset) : -EINVAL );
}

/* Note that this routine will copy from our internal buffer (read at open
 * time) to the user's buffer.
 */
static int wysenor_read(struct inode * inode, struct file * file, const char * buf, int count)
{
   unsigned long flags;
   unsigned i = file->f_pos;
   char *tmp = buf;

   if (i != file->f_pos)
       return -EINVAL;

   save_flags(flags);
   cli();

   for( ; count-- > 0 && i < fboot_size; ++i, ++tmp )
       put_user( wysenor_buffer[i], tmp );
   file->f_pos = i;

   restore_flags(flags);
   return( tmp - buf );
}

/* Note that this routine will copy from the user's buffer to our internal
 * buffer (read at open time and written at close time).
 */
static int wysenor_write(struct inode * inode, struct file * file, const char * buf, int count)
{
   unsigned long flags;
   unsigned i = file->f_pos;
   const char *tmp = buf;

   if (i != file->f_pos)
       return -EINVAL;

   save_flags(flags);
   cli();

   wysenor_dirty = 1; /* buffer has changed */
   for( ; count-- > 0 && i < fboot_size; ++i, ++tmp ) {
       wysenor_buffer[i] = get_user( tmp );
   }
   file->f_pos = i;

   restore_flags(flags);
   return( tmp - buf );
}

static int wysenor_open( struct inode *inode, struct file *file )
{
   if ((wysenor_open_cnt && (file->f_flags & O_EXCL)) ||
       (wysenor_open_mode & WYSENOR_EXCL) ||
       ((file->f_mode & 2) && (wysenor_open_mode & WYSENOR_WRITE)))
       return( -EBUSY );

   if (file->f_flags & O_EXCL)
       wysenor_open_mode |= WYSENOR_EXCL;
   if (file->f_mode & 2)
       wysenor_open_mode |= WYSENOR_WRITE;
   wysenor_open_cnt++;
   wysenor_dirty = 0;

   /* At this point, allocate a buffer, fill it, and read in the device */
   wysenor_buffer = vmalloc(fboot_size);
   if (wysenor_buffer == NULL)
   {
       wysenor_open_cnt--;
       if (file->f_flags & O_EXCL)
           wysenor_open_mode &= ~WYSENOR_EXCL;
       if (file->f_mode & 2)
           wysenor_open_mode &= ~WYSENOR_WRITE;
       return -ENOMEM;
   }
   if (fboot_read(0, wysenor_buffer, fboot_size) < 0)
   {
       vfree(wysenor_buffer);
       wysenor_open_cnt--;
       if (file->f_flags & O_EXCL)
           wysenor_open_mode &= ~WYSENOR_EXCL;
       if (file->f_mode & 2)
           wysenor_open_mode &= ~WYSENOR_WRITE;
       printk("Wyse NOR driver: Unable to read device\n");
       return -EIO;
   }

   MOD_INC_USE_COUNT;
   return( 0 );
}

static int wysenor_release( struct inode *inode, struct file *file )
{
   if (wysenor_open_cnt == 1)
   {
       if (wysenor_dirty)
       {
           /* At this point, erase the NOR and rewrite the buffer to NOR
            * if it has changed
            */
           if (fboot_write(0, wysenor_buffer, fboot_size) < 0)
           {
               printk("Wyse NOR driver: Unable to rewrite device\n");
               return -EIO;
           }
       }
       /* At this point, free the buffer */
       vfree(wysenor_buffer);
   }
   wysenor_open_cnt--;
   if (file->f_flags & O_EXCL)
       wysenor_open_mode &= ~WYSENOR_EXCL;
   if (file->f_mode & 2)
       wysenor_open_mode &= ~WYSENOR_WRITE;

   MOD_DEC_USE_COUNT;
   return( 0 );
}

static struct file_operations wysenor_fops = {
   wysenor_llseek,
   wysenor_read,
   wysenor_write,
   NULL,           /* No readdir */
   NULL,           /* No poll */
   NULL,           /* No ioctl */
   NULL,           /* No mmap */
   wysenor_open,
   wysenor_release
};

static struct miscdevice wysenor_dev = {
   WYSENOR_MINOR,
   "wysenor",
   &wysenor_fops
};


int wysenor_init(void)
{
   int flags;

   /* First test whether the driver should init at all */
   if (!CHECK_DRIVER_INIT())
       return( -ENXIO );

   printk("Wyse NOR driver v%s\n", WYSENOR_VERSION );
   misc_register( &wysenor_dev );

   wysenor_vaddr = (int) vremap(FBOOT_ADDR, FBOOT_SIZE);

   if (wysenor_vaddr == 0)
   {
       printk("Wyse NOR driver: vremap failed\n");
       return -ENOMEM;
   }

   save_flags(flags); cli();
   fboot_enable();
   fb_setdata(0x5555, 0xaa);
   fb_setdata(0x2AAA, 0x55);
   fb_setdata(0x5555, 0x90);               // Read ID
   udelay(20);

   mfg = fb_getdata(0x0000);
   dev = fb_getdata(0x0001);

   fb_setdata(0x5555, 0xaa);
   fb_setdata(0x2AAA, 0x55);
   fb_setdata(0x5555, 0xF0);               // Reset
   udelay(20);
   fboot_disable(0);
   restore_flags(flags);

   switch(mfg) {
   case 0x01:
       if (dev == 0xa4)                // AMD 29F040
           fb_pagesize = 0x100;
       break;
   case 0x1F:                              // Atmel
       if (dev == 0xda)
           fb_pagesize = 0x100;
       break;
   case SST28SF040:                        // SST 280F040 - 0xbf
       if (dev == 0x04) {
           fb_pagesize = 0x100;
           fboot_size = 512 * 1024;
           fb_getdata(0x1823);     // unlock device
           fb_getdata(0x1820);
           fb_getdata(0x1822);
           fb_getdata(0x0418);
           fb_getdata(0x041B);
           fb_getdata(0x0419);
           fb_getdata(0x041A);
       }
       break;
   case 0xDA:                              // WinBond
       if (dev == 0x45) {
           fb_pagesize = 0x080;
       } else if (dev == 0x46) {
           fb_pagesize = 0x100;
           fboot_size = 512 * 1024;
       }
       break;
   default:
       printk("WyseNOR: mfg = 0x%x, dev = 0x%x\n", mfg, dev);
       break;
   }

   fb_pagesize_m1 = fb_pagesize - 1;

   return( 0 );
}

#ifdef MODULE
int init_module (void)
{
   return( wysenor_init() );
}

void cleanup_module (void)
{
   misc_deregister( &wysenor_dev );
}
#endif


/*
 * Local variables:
 *  c-indent-level: 4
 *  tab-width: 4
 * End:
 */
