#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/errno.h>
#include <linux/fcntl.h>
#include <linux/stat.h>
#include <linux/string.h>
#include <linux/locks.h>
#include <linux/malloc.h>
#include <linux/mm.h>
#include <linux/pagemap.h>

#include <asm/segment.h>
#include <asm/io.h>
#include <asm/bitops.h>

#include <linux/fs.h>
#include <linux/flash_fs.h>

STATIC int ffs_readpage(struct inode * inode, struct page * page);
/*
 * We have mostly NULL's here: the current defaults are ok for
 * the coh filesystem.
 */
static struct file_operations ffs_file_operations = {
	NULL,			/* lseek - default */
	ffs_file_read,		/* read */
	ffs_file_write,		/* write */
	NULL,			/* readdir - bad */
	NULL,			/* select - default */
	NULL,			/* ioctl - default */
	NULL,			/* mmap - was generic_mmap */
	NULL,			/* no special open is needed */
	NULL,			/* release */
	ffs_sync_file		/* fsync */
};

struct inode_operations ffs_file_iops = {
	&ffs_file_operations,	/* default file operations */
	NULL,			/* create */
	NULL,			/* lookup */
	NULL,			/* link */
	NULL,			/* unlink */
	NULL,			/* symlink */
	NULL,			/* mkdir */
	NULL,			/* rmdir */
	NULL,			/* mknod */
	NULL,			/* rename */
	NULL,			/* readlink */
	NULL,			/* follow_link */
	ffs_readpage,		/* readpage */
	NULL,			/* writepage */
	NULL,			/* bmap */
	ffs_truncate,		/* truncate */
	NULL			/* permission */
};

STATIC void
memset_tofs(char *addr, char val, int count)
{
	static	char	zeros[16];

	while(count > sizeof(zeros)) {
		memcpy_tofs(addr, zeros, sizeof(zeros));
		addr += sizeof(zeros);
		count -= sizeof(zeros);
	}
	if (count > 0)
		memcpy_tofs(addr, zeros, count);
}

int 
ffs_file_read(struct inode * inode, struct file * filp, char * buf, int count)
{
	char	*b0 = buf, *sectbuf;
	int	size, c, sectno, off, offset;
	struct 	flash_device *fd = flash_device(inode->i_dev);

	ASSERT(inode);
	ASSERT(S_ISREG(inode->i_mode));

	offset = filp->f_pos;
	size = inode->i_size - offset;
	if (count > size)
		count = size;

	while(count > 0) {
		off = SECTOFF(inode->i_blksize, offset);
		c = inode->i_blksize - off;
		if (c > count)
			c = count;
		sectno = ffs_bmap(inode, offset, 0);
		if (sectno >= 0) { 
			if (flash_read(fd, sectno, &sectbuf, inode->i_blksize))
				return -EIO;
			memcpy_tofs(buf, sectbuf + off, c);
		} else {
			memset_tofs(buf, 0, c);
		}
		buf += c;
		count -= c;
		offset += c;
	}
	filp->f_pos = offset;
	
	return(buf - b0);
}
	
int
ffs_file_write(struct inode * inode, struct file * filp, const char * buf, int count)
{
	const char *b0 = buf;
	char	*sectbuf;
	int	i, j, c, sectno, off, offset;
	struct 	flash_device *fd = flash_device(inode->i_dev);

	ASSERT(inode);
	ASSERT(S_ISREG(inode->i_mode));

	if (filp->f_flags & O_APPEND)
		offset = inode->i_size;
	else
		offset = filp->f_pos;
	ASSERT(offset >= 0);

	while(count > 0) {
		off = SECTOFF(inode->i_blksize, offset);
		c = inode->i_blksize - off;
		if (c > count)
			c = count;
		sectno = ffs_bmap(inode, offset, 0);
		if (sectno < 0) {
			if ((sectno = ffs_bmap(inode, offset, 1)) < 0) 
				return -ENOSPC;
			if (flash_read(fd, sectno, &sectbuf, inode->i_blksize))
				return -EIO;
			memset(sectbuf, 0, off);
			memcpy_fromfs(sectbuf + off, buf, c);
			if (offset + c < inode->i_size) {
				i = inode->i_blksize - (off + c);
				if (inode->i_size - (off + c) < i)
					i = inode->i_size - (off + c);
				memset(sectbuf + off + c, 0, i); 
			}
		} else {
			if (flash_read(fd, sectno, &sectbuf, inode->i_blksize))
				return -EIO;
			if (offset >= inode->i_size) { 
				i = inode->i_size + inode->i_blksize - 1;
				i /= inode->i_blksize; 	
				j = (offset / inode->i_blksize) + 1;
				if (j > i)
					flash_remap(fd, sectno);
				else {
					i = SECTOFF(inode->i_blksize, 
						    inode->i_size);
					ASSERT(off >= i);
					memset(sectbuf + i, 0, off - i);
				}
			} else
				flash_remap(fd, sectno);
			memcpy_fromfs(sectbuf + off, buf, c);
		}

		if (flash_write(fd, sectno, sectbuf, inode->i_blksize))
			return -EIO;
			
		buf += c;
		count -= c;
		offset += c;
		if (offset > inode->i_size) {
			inode->i_dirt = 2;
			inode->i_size = offset;
		}
	}

	i = buf - b0;
	if (i > 0 && inode->i_dirt < 1)
		inode->i_dirt = 1;
	filp->f_pos = offset;
	return(i);
}

STATIC int 
ffs_readpage(struct inode * inode, struct page * page)
{
	int	i;
	struct file  fil;

	PRINTF("ffs_readpage: inode = 0x%x, offset = 0x%x\n", 
		(int) inode, (int) page->offset);
	page->count++;
	clear_bit(PG_error, &page->flags);
	fil.f_pos = page->offset; 
	i = ffs_file_read(inode, &fil, (char *) page_address(page), PAGE_SIZE); 
	if (i > 0) {
		set_bit(PG_uptodate, &page->flags);
		i = 0;
	} else {
		set_bit(PG_error, &page->flags);
		i = -EIO;
	}
	free_page(page_address(page));
	return(i);
}

int 
ffs_sync_file(struct inode * inode, struct file * file)
{
	struct 	flash_device *fd = flash_device(inode->i_dev);

	if (!(S_ISREG(inode->i_mode) || S_ISDIR(inode->i_mode) ||
	     S_ISLNK(inode->i_mode)))
		return -EINVAL;
	flash_sync(fd);
	return(0);
}
