#include <linux/config.h>
#include <linux/sched.h>
#include <linux/minix_fs.h>
#include <linux/ext2_fs.h>
#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/mman.h>
#include <linux/malloc.h>
#include <linux/ioctl.h>
#include <linux/module.h>
#include <linux/stat.h>
#include <linux/flash.h>
#include <linux/flash_fs.h>

#include <asm/system.h>
#include <asm/segment.h>
#include <asm/io.h>

/*
 * 35 has been officially registered as the RAMDISK major number, but
 * so is the original MAJOR number of 1.  We're using 1 in
 * include/linux/major.h for now
 */
#define MAJOR_NR 	FLASH_MAJOR
#include <linux/blk.h>

/*
 * Parameters for the boot-loading of the flash.  These are set by
 * init/main.c (from arguments to the kernel command line) or from the
 * architecture-specific setup routine (from the stored bootsector
 * information). 
 */

int uncompress(char *, int *, char *, int);
int compress(char *, int *, char *, int);

int flash_xread(FLASHDEV *, int, char *, int);
int flash_xwrite(FLASHDEV *, int, char *, int);
/*
 *  Basically, my strategy here is to set up a buffer-head which can't be
 *  deleted, and make that my Ramdisk.  If the request is outside of the
 *  allocated size, we must get rid of it...
 *
 */
static void do_flash_request(void)
{
	FLASHDEV *fd;
	int len;

	for ( ;; ) {
		INIT_REQUEST;
	
		fd = flash_device(CURRENT->rq_dev);
		if (fd == 0) {
			end_request(0);
			continue;
		}
	
		len = CURRENT->current_nr_sectors;
		if (len != 1 && len != FLASH_SECTSIZE / fd->pagesize) {
			printk("Flash request length %d\n", len);
			end_request(0);
			continue;
		}

		if (CURRENT->sector + len > fd->pages) {
			end_request(0);
			continue;
		}

		len <<= 9;
#ifdef	LATER
		if (CURRENT->cmd == READ) 
			flash_xread(fd, CURRENT->sector, 
					CURRENT->bh->b_data, len);
		else	
			flash_xwrite(fd, CURRENT->sector, 
					 CURRENT->bh->b_data, len);
#else
		PRINTF("Flash: cmd 0x%x, sector 0x%x, length 0x%x\n",
			CURRENT->cmd, (int) CURRENT->sector, len);
#endif

		end_request(1);
	}
} 

static int 
flash_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
	FLASHDEV *fd;
	int err;
	
	if (!inode || !inode->i_rdev) 	
		return -EINVAL;

	fd = flash_device(inode->i_rdev);
	if (fd == 0)
		return -EIO;

	switch (cmd) {
		case BLKFLSBUF:
			return -EACCES;
			break;
         	case BLKGETSIZE:   /* Return device size */
			if (!arg)  return -EINVAL;
			err = verify_area(VERIFY_WRITE, (long *) arg,
					  sizeof(long));
			if (err)
				return err;
			put_user((fd->pages * fd->pagesize) / 512, 
				 (long *) arg);
			return 0;
			
		default:
			break;
	};

	return 0;
}

static int flash_open(struct inode * inode, struct file * filp)
{
	if (flash_device(inode->i_rdev) == 0)
		return -ENXIO;

	MOD_INC_USE_COUNT;

	return 0;
}

#ifdef MODULE
static void flash_release(struct inode * inode, struct file * filp)
{
	MOD_DEC_USE_COUNT;
}
#endif

static struct file_operations fd_fops = {
	NULL,		/* lseek - default */
	block_read,	/* read - block dev read */
	block_write,	/* write - block dev write */
	NULL,		/* readdir - not here! */
	NULL,		/* select */
	flash_ioctl, 	/* ioctl */
	NULL,		/* mmap */
	flash_open,	/* open */
#ifndef MODULE
	NULL,		/* no special release code... */
#else
	flash_release,	/* module needs to decrement use count */
#endif
	block_fsync		/* fsync */ 
};

int flash_device_init(void);

/* This is the registration and initialization section of the flash driver */
int flash_init(void)
{
	if (flash_device_init() <= 0)
		return -ENODEV;

	if (register_blkdev(MAJOR_NR, "flash", &fd_fops)) {
		printk("Flash: Could not get major %d", MAJOR_NR);
		return -EIO;
	}

	blk_dev[MAJOR_NR].request_fn = &do_flash_request;

	blksize_size[MAJOR_NR] = &flash_device(0)->pagesize;

	printk("Flash driver initialized : %d flash\n", NFLASHS);
	return 0;
}

/* loadable module support */

#ifdef MODULE

int init_module(void)
{
	int error = flash_init();
	if (!error)
		printk(KERN_INFO "Flash: Loaded as module.\n");
	return error;
}

/* Before freeing the module, invalidate all of the protected buffers! */
void cleanup_module(void)
{
	int i;

	for (i = 0 ; i < NFLASHS; i++)
		invalidate_buffers(MKDEV(MAJOR_NR, i));

	unregister_blkdev( MAJOR_NR, "flash" );
	blk_dev[MAJOR_NR].request_fn = 0;
}

#endif  /* MODULE */

#define	FLASH_ADDR	0x2800000
#define	DATA		(*((vubyte *) (FLASH_ADDR + 0x0000)))
#define	ADDRESS		(*((vubyte *) (FLASH_ADDR + 0x1000)))
#define	DISABLE		(*((vubyte *) (FLASH_ADDR + 0x1000)))
#define	COMMAND		(*((vubyte *) (FLASH_ADDR + 0x3000)))

#ifdef	__KERNEL__
#define	OUTB(port, val)	outb(val, port)
#define	INB(port)	inb(port)

#define	printf		printk
#define	malloc(size)	kmalloc(size, GFP_KERNEL)  
typedef volatile unsigned char vubyte;
#else
#define	OUTB(port, val)	outb(port, val)
#define	INB(port)	inb(port)
#endif

int	flash_force, flash_timestamp;

FLASHDEV flashs[NFLASHS];

typedef	struct areab {
	u_char	sequence;
	u_char	inuse;
	u_short	sectno;
	u_short	nextno;
	u_short	size;
} AREAB;

#define	PAGE_FREE	0xff
#define	PAGE_INUSE	0xfe
#define	PAGE_DELETED	0x00

#define	FREEMAP_SET(fd, i)	fd->freemap[i/8] |=  (1 << (i & 0x7))
#define	FREEMAP_CLR(fd, i)	fd->freemap[i/8] &= ~(1 << (i & 0x7))
#define	FREEMAP_TST(fd, i)	fd->freemap[i/8] &   (1 << (i & 0x7))

STATIC int
fd_wait4ready(FLASHDEV *fd)
{
	int	i = 0;

	OUTB(0x22, 0xa8);
	while((INB(0x23) & 1) == 0) {		/* wait for ready */
		if (++i >= 100000) {
			printf("fd_wait4ready: timeout\n");
			return(0);
		}
	}
	return(1);
}

STATIC void
fd_setup(FLASHDEV *fd, int cmd, int pageno)
{
	ASSERT(pageno < fd->pages);

	fd_wait4ready(fd);

        COMMAND = cmd;
	if (cmd != FLASH_CMD_ERASE)
		ADDRESS = 0;
	ADDRESS = pageno;
	ADDRESS = pageno >> 8;

	fd_wait4ready(fd);
}

STATIC int
fd_input(FLASHDEV *fd, char *buf, int size)
{
	int	sum = 0;

	while(size-- > 0) {
		*buf = DATA;
		sum += *buf++;
	}
	return(sum);
}

STATIC int
fd_output(FLASHDEV *fd, char *buf, int size)
{
	int	sum  = 0;

	while(size-- > 0) {
		sum += *buf;
		DATA = *buf++;
	}
	return(sum);
}

STATIC int 
fd_status(FLASHDEV *fd)
{
	if (!fd_wait4ready(fd)) 
		return(FLASH_ERR_TIMEOUT);

        COMMAND = FLASH_CMD_STATUS_READ;

	if (DATA & FLASH_STS_FAIL) {
		printf("flash: status failed\n"); 
		return(FLASH_ERR_STATUS);
	}
	return(0);
}

STATIC int
fd_standby(FLASHDEV *fd)
{
	return(DISABLE);
}

STATIC void
fd_memset(FLASHDEV *fd, int size)
{
	while(--size >= 0)
		DATA = 0xFF; 
}

STATIC int
fd_readb(FLASHDEV *fd, int pageno, AREAB *b)
{
	int	status;

	ASSERT(pageno < fd->pages);
        fd_setup(fd, FLASH_CMD_READ_MODE2, pageno);
	fd_input(fd, (char *) b, sizeof(*b));
	status = fd_status(fd);
	COMMAND = 0;		/* reset back to main data area */
	fd_standby(fd);
	return(status);
}

STATIC int
fd_read(FLASHDEV *fd, int pageno, char *buf, AREAB *b)
{
	int	status;

	ASSERT(pageno < fd->pages);
        fd_setup(fd, FLASH_CMD_READ_MODE1, pageno);
	fd_input(fd, buf, fd->pagesize);
	if (b) fd_input(fd, (char *) b, sizeof(*b));
	status = fd_status(fd);
	fd_standby(fd);
	return(status);
}

STATIC int
fd_writeb(FLASHDEV *fd, int pageno, AREAB *b)
{
	int	status;

	ASSERT(pageno < fd->pages);
	ASSERT(b->sectno != 0);
        fd_setup(fd, FLASH_CMD_WRITE, pageno);
	fd_memset(fd, fd->pagesize);
	fd_output(fd, (char *) b, sizeof(*b));
	fd_memset(fd, fd->rdntsize);
	COMMAND = FLASH_CMD_AUTO_PROGRAM;

	fd_wait4ready(fd);
	status = fd_status(fd);
	fd_standby(fd);
	return(status);
}

STATIC int
fd_write(FLASHDEV *fd, int pageno, char *buf, AREAB *b)
{
	int	status;

	ASSERT(pageno < fd->pages);
	ASSERT(b == 0 || b->sectno != 0);
        fd_setup(fd, FLASH_CMD_WRITE, pageno);
	fd_output(fd, buf, fd->pagesize);
	if (b) 
		fd_output(fd, (char *) b, sizeof(*b));
	else
		fd_memset(fd, sizeof(*b) + fd->rdntsize);
	COMMAND = FLASH_CMD_AUTO_PROGRAM;

	fd_wait4ready(fd);
	status = fd_status(fd);
	fd_standby(fd);
	return(status);
}

STATIC int
fd_erase(FLASHDEV *fd, int blockno)
{
	int	status;
	int	pageno = blockno * fd->ppb;

	ASSERT(pageno < fd->pages);
	if (pageno >= fd->pages)
		return(-1);
	fd_setup(fd, FLASH_CMD_ERASE, pageno);
	COMMAND = FLASH_CMD_ERASE2;

	fd_wait4ready(fd);
	status = fd_status(fd);
	fd_standby(fd);
	
	return(status);
}

STATIC void
fd_rebuild(FLASHDEV *fd)
{
	int	blockno, pageno;
	AREAB	areab;
	DIRENT	*ep;

	printf("flash: initialize all blocks\n");
	for (blockno = SBBLOCKNO; blockno < fd->blocks; ++blockno) {
		fd_erase(fd, blockno);
	}
	for (pageno = SBPAGENO; pageno < fd->pages; pageno += fd->ppb) {
		if (fd_readb(fd, pageno, &areab) == 0) {
			ASSERT(areab.inuse == PAGE_FREE);
			fd->mapping[SBSECTNO] = pageno;
			areab.inuse = PAGE_INUSE;
			areab.sectno = SBSECTNO;
			areab.sequence = 0;
			fd_writeb(fd, pageno, &areab); 

			flash_read(fd, SBSECTNO, (char **) &ep, fd->pagesize);
			ep->mode = S_IFDIR | 0777;	/* all access */
			ep->entsize = sizeof(*ep) + 2;
			ep->noff = ep->name - (char *) ep;
			ep->nlink = 2;
			ep->size = 512;
			strcpy(ep->name, "/");
			flash_write(fd, SBSECTNO, (char *) ep, fd->pagesize);
			flash_sync(fd);
			return;
		}
	}
}

STATIC int
fd_gc(FLASHDEV *fd)
{
	int	i, j;
	AREAB	areab;

	for (j = i = SBBLOCKNO + 1; i < fd->blocks; ++i) {
		if (fd->blkdelcnt[i] > fd->blkdelcnt[j])
			j = i;
	}
	if (fd->blkdelcnt[j] > 0) {
		for (i = j * fd->ppb; i < (j + 1) * fd->ppb; ++i) {
			fd_readb(fd, i, &areab);
			ASSERT(areab.inuse != PAGE_FREE);
			if (areab.inuse == PAGE_INUSE) {
				/* un-finished yet */
			}
		}
	}
	return(0);
}

/*
**	Returns a pageno, clear the bit in the freemap
*/
STATIC int
fd_newpage(FLASHDEV *fd, int sectno, int nextno)
{
	int	i, s, pageno, startindx;
	AREAB	areab;

	ASSERT(sectno < fd->pages);
	startindx = i = fd->freeindx;
	for ( ;; ) {
		if (fd->freemap[i] != 0) {
			fd->freeindx = i;
			pageno = i * 8;
			for (s = 0; s < 8; ++s, ++pageno) {
				if (fd->freemap[i] & (1 << s)) {
					fd->freemap[i] &= ~(1 << s);
					fd->freepages--;
					fd->mapping[sectno] = pageno;
					memset(&areab, 0xff, sizeof(areab));
					areab.inuse = PAGE_INUSE;
					areab.sequence = 0;
					areab.sectno = sectno;
					areab.nextno = nextno;
					fd_writeb(fd, pageno, &areab);
					return(pageno);
				}
			}
		}
		if (++i >= fd->pages / 8)
			i = SBPAGENO / 8;
		if (i == startindx) {
			if (fd_gc(fd) > 0)
				continue;
			printf("flash alloc: no space\n");
			return(0xFFFF);
		}
	}
}

STATIC int
fd_newsect(FLASHDEV *fd)
{
	int	sectno, startindx;

	startindx = sectno = fd->sectindx;
	for ( ;; ) {
		if (++sectno >= fd->pages)
			sectno = SBSECTNO;
		if (fd->mapping[sectno] == 0xFFFF) {
			fd->sectindx = sectno;
			return(sectno);
		}
		if (sectno == startindx) {
			printf("flash panic: out of sectno\n");
			return(0xFFFF);
		}
	}
}

STATIC void
fd_freepage(FLASHDEV *fd, int pageno)
{
	ASSERT(pageno < fd->pages);
	if (FREEMAP_TST(fd, pageno)) {
		printf("flash panic: double free pageno %d\n", pageno);
		return;
	}

	FREEMAP_SET(fd, pageno);
	fd->freepages++;
} 

STATIC void
fd_reclaim(FLASHDEV *fd, int blockno)
{
	int	i, status, pageno;

	ASSERT(blockno < fd->blocks);
	fd->blkdelcnt[blockno] = 0;
	status = fd_erase(fd, blockno);
	if (status == 0) {
		fd->deletepages -= fd->ppb;
		pageno = blockno * fd->ppb;
		for (i = 0; i < fd->ppb; ++i, ++pageno) 
			fd_freepage(fd, pageno);
	}
}

STATIC int
fd_delpage(FLASHDEV *fd, int pageno)
{
	int	blockno = pageno / fd->ppb;
	AREAB	areab;

	ASSERT(pageno < fd->pages);
	fd_readb(fd, pageno, &areab);
	ASSERT(areab.inuse == PAGE_INUSE);
	areab.inuse = PAGE_DELETED;	/* mark old page deleted */
	fd_writeb(fd, pageno, &areab);

	fd->deletepages++;
	if (++fd->blkdelcnt[blockno] >= fd->ppb)
		fd_reclaim(fd, blockno);
	return(areab.nextno);
}

STATIC int
fd_mapinit(FLASHDEV *fd)
{
	int	pageno, blockno, status;
	AREAB	areab;
	DIRENT	*dp;

	for (pageno = SBPAGENO; pageno < fd->pages; pageno += fd->ppb) {
		status = fd_readb(fd, pageno, &areab);
		if (status == 0)	/* stop at first good page */
			break;
	}
	if (pageno >= fd->pages || flash_force || 
	    areab.inuse != PAGE_INUSE || areab.sectno != SBSECTNO) {
#ifdef	EPROM
		return(-EIO);
#else
		fd_rebuild(fd);
#endif
	}

	memset((char *) fd->mapping, 0xff, fd->pages * sizeof(fd->mapping[0]));
	for (pageno = SBPAGENO; pageno < fd->pages; ++pageno) {
		status = fd_readb(fd, pageno, &areab);
		if (status != 0) {
			fd->badpages++;
			continue;
		}
		if (areab.inuse != PAGE_INUSE)
			continue;
		if (areab.sectno >= fd->pages)
			printf("flash pageno %d, sectno 0x%x\n",
					pageno, areab.sectno);
		else if (fd->mapping[areab.sectno] != 0xFFFF)
			printf("flash page %d, file crossed\n", pageno);
		else {
			fd->inusepages++;
			fd->mapping[areab.sectno] = pageno;
		}
	}
	ASSERT(fd->mapping[SBSECTNO] != 0xFFFF);

	memset((char *) fd->freemap, 0, fd->pages / 8 );
	memset((char *) fd->blkdelcnt, 0, fd->blocks);
	fd->sectindx = SBSECTNO;
	fd->freeindx = SBPAGENO / 8;

	for (pageno = SBPAGENO; pageno < fd->pages; ++pageno) {
		status = fd_readb(fd, pageno, &areab);
		if (status != 0)
			continue;
		switch(areab.inuse) {
		case PAGE_FREE:
			ASSERT(areab.sectno == 0xFFFF);
			ASSERT(areab.nextno == 0xFFFF);
			fd_freepage(fd, pageno);
			break;
		case PAGE_DELETED:
			blockno = pageno / fd->ppb;
			fd->deletepages++;
			fd->blkdelcnt[blockno]++;
			if (fd->blkdelcnt[blockno] >= fd->ppb)
				fd_reclaim(fd, blockno);
			break;
		case PAGE_INUSE:
			break;
		default:
			printf("flash: pageno %d unknown tag 0x%x\n", 
				pageno, areab.inuse);
			break;
		}
	}

	flash_read(fd, SBSECTNO, (char **) &dp, fd->pagesize);
	ASSERT(S_ISDIR(dp->mode));
	
	PRINTF("flash_dev: free=%d, bad=%d, inuse=%d, delete=%d\n",
		fd->freepages, fd->badpages, fd->inusepages, fd->deletepages);
	fd->initok = 1;
	return(0);
}

STATIC int
fd_init(FLASHDEV *fd, int blocks, int pages, int blocksize, int pagesize)
{
	int	i;
	char	*buf;

	memset(fd, 0, sizeof(*fd));
	i = FLASH_NCACHES * FLASH_SECTSIZE;
	i += blocks + pages / 8 + sizeof(fd->mapping[0]) * pages;
	buf = (char *) malloc(i);
	if (buf == 0)
		return(-ENOMEM);

	fd->blkdelcnt = (unsigned char *) buf + FLASH_NCACHES * FLASH_SECTSIZE;
	fd->freemap = fd->blkdelcnt + blocks;
	fd->mapping = (unsigned short *) (fd->freemap + pages / 8);
	
	fd->blocks = blocks;
	fd->pages = pages;
	fd->blocksize = blocksize;
	fd->pagesize = pagesize;
	fd->ppb = blocksize / pagesize;
	fd->sectsize = FLASH_SECTSIZE;
	if (pagesize == FLASH_PAGE_SIZE)
		fd->rdntsize = FLASH_RDNT_SIZE;

	for (i = 0; i < FLASH_NCACHES; ++i, buf += FLASH_SECTSIZE) {
		fd->cache[i].buf = buf;
		fd->cache[i].sectno = 0xFFFF;
		fd->cache[i].dirty = 0;
		fd->cache[i].access = 0;
	}
		
	return(fd_mapinit(fd));
}

int
fd_mapping(FLASHDEV *fd, int sectno)
{
	ASSERT(sectno < fd->pages);
	return(fd->mapping[sectno]);
}

FLASHDEV *
flash_device(int dev)
{
	int	minor = MINOR(dev);

	ASSERT(minor < NFLASHS && flashs[minor].initok);
	if (minor >= NFLASHS || flashs[minor].initok == 0)
		return(0);
	return(flashs + minor);
}

int
flash_device_init(void)
{
	vubyte	iop, id;
	int	count = 0;
	FLASHDEV *fd = &flashs[0];

	OUTB(0x22, 0x31);		/* Start with ROMCS2 */
	id = INB(0x23) & 0xF0;
	OUTB(0x23, id + 2 * 4);		/* CS2 Selected */

	OUTB(0x3e0, 0x46);
	iop = INB(0x3e1);
	OUTB(0x3e1, iop | 4);		/* enable FLASH_ADDR mapping */

	OUTB(0x22, 0x31);		/* Select ROMCS2 into addr space */
	iop = INB(0x23) & 0x0C;
	if (iop != 8) {
		printf("flash_init: can't select ROMCS2\n");
		return(0);
	}
	OUTB(0x23, iop);

	COMMAND = FLASH_CMD_RESET;
	if (fd_wait4ready(fd) == 0) {
		printf("flash_init: not ready\n");
		return(0);
	}

	COMMAND  = FLASH_CMD_ID_READ;	/* Read Flash ID */
	ADDRESS  = 0x00;
	iop = DATA;			/* ID 0 */
	id  = DATA;			/* ID 1 */

	PRINTF("Flash_init: 0x%x  0x%x\n", iop, id);

	switch(id) {
	case 0x6B:		/* TC5832AFT */
		if (!fd_init(fd, FLASH_BLOCKS, FLASH_PAGES, 
				       FLASH_BLOCK_SIZE, FLASH_PAGE_SIZE))
			count++;
		break;
	case 0x64:		/* TC5816AFT */
		if (!fd_init(fd, FLASH_BLOCKS, FLASH_PAGES, 
				       FLASH_BLOCK_SIZE, FLASH_PAGE_SIZE / 2))
			count++;
		break;
	}
	
	return(count);
}

int
flash_alloc(FLASHDEV *fd, int size)
{
	int	sectno = -1, nextno = 0xFFFF, pageno;

	while(size > 0) {
		size -= fd->pagesize;
		sectno = fd_newsect(fd);
		if (sectno == 0xFFFF)
			return(-1);
		pageno = fd_newpage(fd, sectno, nextno);
		if (pageno == 0xFFFF)
			return(-1);
		nextno = sectno;
	}
	return(sectno);
}

void
flash_delete(FLASHDEV *fd, int sectno)
{
	int	pageno;

	ASSERT(sectno < fd->pages);
	for ( ;; ) {
		flash_invalidate(fd, sectno);
		pageno = fd->mapping[sectno];
		ASSERT(pageno < fd->pages);
		fd->mapping[sectno] = 0xFFFF;
		sectno = fd_delpage(fd, pageno);
		if (sectno == 0xFFFF)
			break;
	}
}

void
flash_del_index(FLASHDEV *fd, int sectno, int level)
{
	ushort	*index, indexbuf[FLASH_PAGE_SIZE];
	int	i;

	flash_read(fd, sectno, (char **) &index, fd->pagesize);
	memcpy(indexbuf, index, fd->pagesize);
	for (i = 0; i < fd->pagesize/sizeof(ushort); ++i) {
		if (indexbuf[i] != 0xFFFF) {
			if (level > 0)
				flash_del_index(fd, indexbuf[i], level - 1);
			else
				flash_delete(fd, indexbuf[i]);
		}
	}
	flash_delete(fd, sectno);
}

int
flash_remap(FLASHDEV *fd, int sectno)
{
	int	pageno, nextno;

	ASSERT(sectno < fd->pages);
	pageno = fd->mapping[sectno];
	ASSERT(pageno < fd->pages);

	nextno = fd_delpage(fd, pageno);
	pageno = fd_newpage(fd, sectno, 0xFFFF);
	if (pageno == 0xFFFF)
		return(-1);
	ASSERT(pageno < fd->pages);
	fd->mapping[sectno] = pageno;
	if (nextno != 0xFFFF)
		flash_delete(fd, nextno);
	return(pageno);
}

static	char	fd_tmpbuf[FLASH_SECTSIZE];

int
flash_xread(FLASHDEV *fd, int sectno, char *buf, int sectsize)
{
	int	pageno, status, size, i;
	AREAB	areab;

	ASSERT(sectno < fd->pages);
	pageno = fd->mapping[sectno];
	ASSERT(pageno < fd->pages);

	if (sectsize <= fd->pagesize)
		return(fd_read(fd, pageno, buf, 0));

	status = fd_read(fd, pageno, fd_tmpbuf, &areab); /* read link & size */
	if (status)
		return(status);
	ASSERT(fd->mapping[areab.sectno] == pageno);

	size = areab.size;
	if (size == 0xFFFF) {  				/* never written */
		memset(buf, 0xFF, sectsize);
		return(0);
	}
	
	ASSERT(size > 0 && size <= sectsize);

	for (i = fd->pagesize; i < size; i += fd->pagesize) {
		ASSERT(areab.nextno < fd->pages);
		pageno = fd->mapping[areab.nextno];
		ASSERT(pageno < fd->pages);
		if (pageno == 0xFFFF)
			return(-1);
		status = fd_read(fd, pageno, fd_tmpbuf + i, &areab);
		if (status)
			return(status);
		ASSERT(fd->mapping[areab.sectno] == pageno);
	}

	ASSERT(areab.nextno == 0xFFFF);
	i = sectsize;
	status = uncompress(buf, &i, fd_tmpbuf, size);
	ASSERT(status == 0);
	ASSERT(i == sectsize);
	if (i != sectsize)
		return(-1);
	
	return(0);
}

int
flash_xwrite(FLASHDEV *fd, int sectno, char *buf, int sectsize)
{
	int	pageno, status, size, i;
	AREAB	areab, *b;

	ASSERT(sectno < fd->pages);
	pageno = fd->mapping[sectno];
	ASSERT(pageno < fd->pages);

	if (sectsize <= fd->pagesize)
		return(fd_write(fd, pageno, buf, 0));

	status = fd_readb(fd, pageno, &areab);	/* read link & size */
	if (status)
		return(status);
	ASSERT(fd->mapping[areab.sectno] == pageno);

	if (areab.size != 0xFFFF) {
		pageno = flash_remap(fd, sectno);
		if (pageno < 0)
			return(-1);
		ASSERT(fd->mapping[sectno] == pageno);
		status = fd_readb(fd, pageno, &areab);
		if (status)
			return(status);
	}
	ASSERT(areab.nextno == 0xFFFF);

	size = sectsize;
	status = compress(fd_tmpbuf, &size, buf, sectsize);
	ASSERT(status == 0);
	ASSERT(size > 0 && size <= sectsize);
	if (status)
		return(-1);

	if ((i = size - fd->pagesize) > 0) {
		areab.nextno = flash_alloc(fd, i);
		if (areab.nextno == 0xFFFF)
			return(-1);
	}
	areab.size = size; /* first sector has size info + the nextno */
	
	for (i = 0, b = &areab; ; b = 0) {
		status = fd_write(fd, pageno, fd_tmpbuf + i, b);
		if (status)
			return(status);
		i += fd->pagesize;
		if (i >= size) {
			ASSERT(areab.nextno == 0xFFFF);
			return(0);
		}
		sectno = areab.nextno;
		ASSERT(sectno < fd->pages);
		pageno = fd->mapping[sectno];
		ASSERT(pageno < fd->pages);
		status = fd_readb(fd, pageno, &areab);	/* follow link */
		if (status)
			return(status);
	}
}

int
flash_read(FLASHDEV *fd, int sectno, char **buf, int sectsize)
{
	FDCACHE	*cp, *lp;
	int	i, status;

	ASSERT(sectno < fd->pages);
	cp = lp = fd->cache;
	for (i = 0; i < FLASH_NCACHES; ++i, ++cp) {
		if (cp->sectno == sectno) {
			ASSERT(sectsize == cp->sectsize);
			*buf = cp->buf;
			cp->access = ++flash_timestamp;
			return(0);
		}
		if (cp->access < lp->access)
			lp = cp;
	}
	if (lp->dirty) {
		lp->dirty = 0;
		flash_xwrite(fd, lp->sectno, lp->buf, lp->sectsize);
	}
	*buf = lp->buf;
	lp->sectno = sectno;
	lp->access = ++flash_timestamp;
	lp->sectsize = sectsize;
	status = flash_xread(fd, lp->sectno, lp->buf, sectsize);
	if (status != 0) {
		lp->sectno = 0xFFFF;
		lp->access = -1;
	}
	return(status);
}

int
flash_write(FLASHDEV *fd, int sectno, char *buf, int sectsize)
{
	FDCACHE	*cp;
	int	i;

	ASSERT(sectno < fd->pages);
	cp = fd->cache;
	for (i = 0; i < FLASH_NCACHES; ++i, ++cp) {
		if (cp->sectno == sectno) {
			ASSERT(cp->sectsize == sectsize);
			if (buf != cp->buf)
				memcpy(cp->buf, buf, sectsize);
			cp->dirty = 1;
			return(0);
		}
	}
	return(flash_xwrite(fd, sectno, buf, sectsize));
}

void
flash_sync(FLASHDEV *fd)
{
	FDCACHE	*cp;
	int	i;

	cp = fd->cache;
	for (i = 0; i < FLASH_NCACHES; ++i, ++cp) {
		if (cp->dirty == 0)
			continue;
		cp->dirty = 0;
		flash_xwrite(fd, cp->sectno, cp->buf, cp->sectsize);
	}
}

void
flash_invalidate(FLASHDEV *fd, int sectno)
{
	FDCACHE	*cp;
	int	i;

	cp = fd->cache;
	for (i = 0; i < FLASH_NCACHES; ++i, ++cp) {
		if (cp->sectno == sectno) {
			cp->dirty = 0;
			cp->access = 0;
			cp->sectno = 0xFFFF;
			break;
		}
	}
}

#ifdef	DEBUG
void
flash_dump(FLASHDEV *fd)
{
	PRINTF("flash_dev: free=%d, bad=%d, inuse=%d, delete=%d\n",
		fd->freepages, fd->badpages, fd->inusepages, fd->deletepages);
}
#endif
