#include <linux/module.h>

#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/stat.h>
#include <linux/string.h>
#include <linux/locks.h>

#include <linux/flash_fs.h>

#include <asm/segment.h>

STATIC struct ffs_dirent *
ffs_empty_dirent(struct inode *dir, const char *name, mode_t mode, struct ffs_dirent *copy, ino_t *ino)
{
	int	status, entsize, offset, off, sectno;
	char	*sectbuf;
	struct 	ffs_dirent *de;
	struct	flash_device *fd = flash_device(dir->i_dev);

	if (name) {
		entsize = strlen(name) + sizeof(*de);
		if (entsize & 1)
			entsize++;
		ASSERT(entsize < 0xFF);
	} else
		entsize = copy->entsize;

	for (offset = 0; ; offset += dir->i_blksize) {
		sectno = ffs_bmap(dir, offset, 1);
		if (sectno < 0)		/* ENOSPC on dir */
			break;
		status = flash_read(fd, sectno, &sectbuf, fd->pagesize);
		if (status != 0)
			break;

		for (off = 0; off < fd->pagesize; off += de->entsize) {
			de = (struct ffs_dirent *) (sectbuf + off);
			if (de->entsize != 0xFF)
				continue;
			if (fd->pagesize - off < entsize)
				break;
			*ino = mk_ino(sectno, off);
			ASSERT(de->size == -1);
			ASSERT(de->mode == 0xFFFF);
			ASSERT(de->name[0] == -1);
			if (copy) {
				PRINTF("ffs_empty_dirent: move to 0x%x\n",
					(int) *ino);
				memcpy(de, copy, entsize);
			} else {
				PRINTF("ffs_empty_dirent: new ino 0x%x\n",
					(int) *ino);
				de->entsize = entsize;
				de->noff = de->name - (char *) de;
				de->mode = mode;
				de->size = 0;
				de->nlink = 1;
				strcpy(de->name, name);
			}
			flash_write(fd, sectno, sectbuf, fd->pagesize);
			flash_sync(fd);
			return(de);
		}
	}

	printk("flashfs: unable to allocate a new dirent\n");
	return(0);
}

void
ffs_reloc_dirent(struct inode *inode)
{
	ino_t	ino;
	struct	ffs_dirent *old_de, *new_de;
	struct 	flash_device *fd;

	inode->i_dirt = 0;
	old_de = ffs_get_dirent(inode);
	if (old_de == 0)
		return;
	old_de->size = inode->i_size;
	old_de->nlink = inode->i_nlink;
	old_de->mode = inode->i_mode;
	PRINTF("ffs_reloc_dirent: old ino 0x%x\n", (int) inode->i_ino);
	new_de = ffs_empty_dirent(inode->i_bound_to, 0, 
				  inode->i_mode, old_de, &ino);
	if (new_de) {
		old_de->mode = 0;	/* zap the old entry */
		fd = flash_device(inode->i_dev);
		flash_write(fd, ino_sectno(inode->i_ino),
				de2buf(old_de, ino_offset(inode->i_ino)),
				fd->pagesize);
		remove_inode_hash(inode);
		inode->i_ino = ino;
		insert_inode_hash(inode);
	}
}

void 
ffs_do_truncate(struct inode *inode)
{
	int	i;
	struct	ffs_dirent  *de;
	struct 	flash_device *fd = flash_device(inode->i_dev);

	de = ffs_get_dirent(inode);
	ASSERT(de);
	if (de == 0)
		return;
	PRINTF("ffs_truncate: inode 0x%x, name: %s.\n", (int) inode, de->name);

	for (i = 0; i < DIRMAP0; ++i) {
		if (de->map[i] != 0xFFFF) 
			flash_delete(fd, de->map[i]);
		de->map[i] = 0xFFFF;
	}

	for (i = 0; i < 3; ++i) {
		if (de->map[DIRMAP0 + i] != 0xFFFF)
			flash_del_index(fd, de->map[DIRMAP0 + i], i);
		de->map[DIRMAP0 + i] = 0xFFFF;
	}
}

void 
ffs_truncate(struct inode *inode)
{
	ffs_do_truncate(inode);
	ffs_reloc_dirent(inode);
	flash_sync(flash_device(inode->i_dev));
}

void 
ffs_put_inode(struct inode *inode)
{
	if (inode->i_nlink == 0) {
		inode->i_size = 0;
		ffs_do_truncate(inode);
		if (ffs_free_inode(inode) < 0)
			ffs_reloc_dirent(inode);
		flash_sync(flash_device(inode->i_dev));
	}
}

int 
ffs_free_inode(struct inode * inode)
{
	struct super_block * sb;
	struct	ffs_dirent *de;
	struct 	flash_device *fd = flash_device(inode->i_dev);

	if (!inode)
		return(-1);
	if (!inode->i_dev) {
		printk("ffs_free_inode: inode has no device\n");
		return(-1);
	}
	if (inode->i_count != 1) {
		printk("ffs_free_inode: inode has count=%d\n", inode->i_count);
		return(-1);
	}
	if (inode->i_nlink) {
		printk("ffs_free_inode: inode has nlink=%d\n", inode->i_nlink);
		return(-1);
	}
	if (!(sb = inode->i_sb)) {
		printk("ffs_free_inode: inode on nonexistent device\n");
		return(-1);
	}

	de = ffs_get_dirent(inode);
	ASSERT(de);
	if (de == 0)
		return(-1);
	de->mode = 0;
	flash_write(fd, ino_sectno(inode->i_ino), 	
			de2buf(de, ino_offset(inode->i_ino)), fd->pagesize);
	clear_inode(inode);
	return(0);
}

struct inode * 
ffs_new_inode(struct inode *dir, const char *name, mode_t mode)
{
	ino_t	ino;
	struct 	inode * inode;
	struct 	ffs_dirent *de;

	PRINTF("ffs_new_inode: name = %s, mode = 0x%x\n", name, mode);
	ASSERT(dir);

	if ((de = ffs_empty_dirent(dir, name, mode, 0, &ino)) == 0)
		return NULL;

	if ((inode = get_empty_inode()) == 0)
		return NULL;

	inode->i_sb = dir->i_sb;
	inode->i_ino = ino;
	inode->i_flags = inode->i_sb->s_flags;
	inode->i_mode = mode;
	inode->i_count = 1;
	inode->i_nlink = 1;
	inode->i_dev = inode->i_sb->s_dev;
	inode->i_uid = current->fsuid;
	inode->i_gid = (dir->i_mode & S_ISGID) ? dir->i_gid : current->fsgid;
	inode->i_dirt = 0;
	inode->i_mtime = inode->i_atime = inode->i_ctime = CURRENT_TIME;
	inode->i_op = NULL;
	inode->i_blksize = S_ISREG(mode) ? FLASH_SECTSIZE : dir->i_blksize;
	inode->i_blocks = inode->i_blksize;
	inode->i_size = 0;		/* ditto */
	inode->i_bound_to = dir;
	insert_inode_hash(inode);
	return inode;
}

void 
ffs_read_inode(struct inode *inode)
{
	struct ffs_dirent *de;
	struct flash_device *fd = flash_device(inode->i_dev);

	/* be ready to bail out */

	inode->i_op = 0;
	inode->i_mode = 0;

	if (inode->i_ino == 0 || ino_sectno(inode->i_ino) > fd->pages) {
		printk("ffs_read_inode: bad ino 0x%x\n", (int) inode->i_ino);
		return;
	}

	de = ffs_get_dirent(inode); 
	if (de == 0)
		return;
	/*
	 * canned stuff
	 */

	inode->i_uid = 0;
	inode->i_gid = 0;

	inode->i_mode = de->mode;

	ffs_set_iops(inode);
	/*
	 * these fields must be filled in from the dir entry, which we don't
	 * have but lookup does.  It will fill them in before letting the
	 * inode out of its grasp.
	 */

	inode->i_atime = 0;
	inode->i_mtime = 0;
	inode->i_ctime = 0;
	inode->i_size = de->size;
	inode->i_rdev = to_kdev_t(de->rdev);

	/*
	 * fill in the rest
	 */
	inode->i_nlink = 1;
	inode->i_blksize = S_ISREG(de->mode) ? FLASH_SECTSIZE : fd->pagesize;
}

void 
ffs_write_inode(struct inode * inode)
{
	struct 	flash_device *fd = flash_device(inode->i_dev);

	if (inode->i_dirt >= 2)
		ffs_reloc_dirent(inode);
	inode->i_dirt = 0;
	flash_sync(fd);
}
