#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/flash_fs.h>
#include <linux/string.h>
#include <linux/stat.h>
#include <linux/errno.h>

/*
 *	ffs_find_entry()
 *
 * finds an entry in the specified directory with the wanted name. It
 * returns the cache buffer in which the entry was found, and the entry
 * itself (as a parameter - res_dir). It does NOT read the inode of the
 * entry - you'll have to do that yourself if you want to.
 */
STATIC struct ffs_dirent *
ffs_find_entry(struct inode *dir, const char * name, int len, ino_t *ino)
{
	int	sectno, offset, off;
	char	*sectbuf;
	struct 	ffs_dirent *de;
	struct 	flash_device *fd = flash_device(dir->i_dev);
	
	PRINTF("ffs_find_entry(%s) ...\n", name);
	if (name[0] == '.') {
		if (len == 1) {
			*ino = dir->i_ino;
			return ffs_get_dirent(dir);
		}
		if (name[1] == '.' && len == 2) {
			ASSERT(dir->i_bound_to);
			*ino = dir->i_bound_to->i_ino;
			return ffs_get_dirent(dir->i_bound_to);
		}
	}

	for (offset = 0; ; offset += dir->i_blksize) {
		sectno = ffs_bmap(dir, offset, 0);
		if (sectno < 0)		/* EOF on dir */
			return(0);
		if (flash_read(fd, sectno, &sectbuf, fd->pagesize) < 0)
			return(0);
		for (off = 0; off < fd->pagesize; off += de->entsize) {
			de = (struct ffs_dirent *) (sectbuf + off);
			if (de->entsize == 0xFF) 
				break;
			ASSERT(de->entsize != 0);
			ASSERT(de->mode != 0xFFFF);
			if (de->mode == 0x0000)
				continue;
			if (de->name[len])
				continue;
			if (memcmp(de->name, name, len) != 0)
				continue;
			*ino = mk_ino(sectno, off);
			return(de);
		}
	}
	return(0);
}

int 
ffs_lookup(struct inode *dir, const char *name, int len, struct inode **result)
{
	ino_t 	ino;
	struct 	ffs_dirent * de;

	*result = NULL;
	if (!dir)
		return -ENOENT;
	if (!S_ISDIR(dir->i_mode)) {
		iput(dir);
		return -ENOENT;
	}
	if ((de = ffs_find_entry(dir, name, len, &ino)) == 0) {
		iput(dir);
		return -ENOENT;
	}
	
	if (!(*result = iget(dir->i_sb, ino))) {
		iput(dir);
		return -EACCES;
	}
	iput(dir);
	return 0;
}

int 
ffs_create(struct inode *dir, const char * name, int len, int mode,
	struct inode ** result)
{
	struct inode * inode;

	*result = NULL;
	if (!dir)
		return -ENOENT;
	inode = ffs_new_inode(dir, name, mode);
	if (!inode) {
		iput(dir);
		return -ENOSPC;
	}
	PRINTF("ffs_create(%s): mode 0x%x ok\n", name, mode);
	inode->i_op = &ffs_file_iops;
	iput(dir);
	*result = inode;
	return 0;
}

int 
ffs_mknod(struct inode *dir, const char *name, int len, int mode, int rdev)
{
	ino_t	ino;
	struct 	inode * inode;
	struct 	ffs_dirent * de;
	struct	flash_device *fd = flash_device(dir->i_dev);

	if (!dir)
		return -ENOENT;
	if ((de = ffs_find_entry(dir, name, len, &ino)) != 0) {
		iput(dir);
		return -EEXIST;
	}
	inode = ffs_new_inode(dir, name, mode);
	if (!inode) {
		iput(dir);
		return -ENOSPC;
	}
	PRINTF("ffs_mknod(%s): mode 0x%x rdev 0x%x ok\n", name, mode, rdev);
	inode->i_ino = ino;
	inode->i_uid = current->fsuid;
	inode->i_mode = mode;
	inode->i_op = NULL;
	if (S_ISBLK(inode->i_mode) || S_ISCHR(inode->i_mode)) {
		inode->i_rdev = to_kdev_t(rdev);
		de->rdev = inode->i_rdev;
		flash_write(fd, ino_sectno(ino), 
			    de2buf(de, ino_offset(ino)), fd->pagesize);
		flash_sync(fd);
	}
		
	if (S_ISDIR(inode->i_mode)) {
		if (dir->i_mode & S_ISGID)
			inode->i_mode |= S_ISGID;
	}
	ffs_set_iops(inode);
	iput(dir);
	iput(inode);
	return 0;
}

int ffs_mkdir(struct inode *dir, const char *name, int len, int mode)
{
	ino_t	ino;
	struct 	inode *inode;
	struct 	ffs_dirent *de;

	if (!dir)
		return -EINVAL;
	if (name[0] == '.') {
		if (len == 1)
			return 0;
		if (name[1] == '.' && len == 2)
			return 0;
	}
	if ((de = ffs_find_entry(dir, name, len, &ino)) != 0) {
		iput(dir);
		return -EEXIST;
	}
	if (dir->i_nlink >= 0xFFFE) {
		iput(dir);
		return -EMLINK;
	}
	mode = S_IFDIR | (mode & 0777 & ~current->fs->umask);
	inode = ffs_new_inode(dir, name, mode);
	if (!inode) {
		iput(dir);
		return -ENOSPC;
	}
	PRINTF("ffs_mkdir(%s): mode 0x%x ok\n", name, mode);
	inode->i_op = &ffs_dir_iops;
	inode->i_size = 100;
	inode->i_ino = ino;
	if (dir->i_mode & S_ISGID)
		inode->i_mode |= S_ISGID;
	iput(dir);
	iput(inode);
	return 0;
}

int ffs_rmdir(struct inode *dir, const char * name, int len)
{
	int 	retval;
	ino_t	ino;
	struct 	inode *inode = 0;
	struct 	ffs_dirent *de;
	struct 	flash_device *fd = flash_device(dir->i_dev);

	inode = NULL;
	retval = -ENOENT;
	if ((de = ffs_find_entry(dir, name, len, &ino)) == 0)
		goto end_rmdir;
	retval = -EPERM;
	if (!(inode = iget(dir->i_sb, ino)))
		goto end_rmdir;
        if ((dir->i_mode & S_ISVTX) && !fsuser() &&
            current->fsuid != inode->i_uid &&
            current->fsuid != dir->i_uid)
		goto end_rmdir;
	if (inode->i_dev != dir->i_dev)
		goto end_rmdir;
	if (inode == dir)	/* we may not delete ".", but "../dir" is ok */
		goto end_rmdir;
	if (!S_ISDIR(inode->i_mode)) {
		retval = -ENOTDIR;
		goto end_rmdir;
	}
	if (inode->i_nlink) {
		retval = -ENOTEMPTY;
		goto end_rmdir;
	}
	if (ino != inode->i_ino) {
		retval = -ENOENT;
		goto end_rmdir;
	}
	if (inode->i_count > 1) {
		retval = -EBUSY;
		goto end_rmdir;
	}
	inode->i_nlink = 0;
	dir->i_nlink--;
	inode->i_ctime = dir->i_ctime = dir->i_mtime = CURRENT_TIME;
	retval = 0;
end_rmdir:
	iput(dir);
	iput(inode);
	PRINTF("ffs_rmdir(%s): retval = %d\n", name, retval);
	return retval;
}

int ffs_unlink(struct inode *dir, const char * name, int len)
{
	int 	retval;
	ino_t	ino;
	struct 	inode *inode;
	struct 	ffs_dirent *de;

repeat:
	retval = -ENOENT;
	inode = NULL;
	if ((de = ffs_find_entry(dir, name, len, &ino)) == 0)
		goto end_unlink;
	if (!(inode = iget(dir->i_sb, ino)))
		goto end_unlink;
	retval = -EPERM;
	if (S_ISDIR(inode->i_mode))
		goto end_unlink;
	if (ino != inode->i_ino) {
		iput(inode);
		current->counter = 0;
		schedule();
		goto repeat;
	}
	if ((dir->i_mode & S_ISVTX) && !fsuser() &&
	    current->fsuid != inode->i_uid &&
	    current->fsuid != dir->i_uid)
		goto end_unlink;
	if (ino != inode->i_ino) {
		retval = -ENOENT;
		goto end_unlink;
	}
	if (!inode->i_nlink) {
		printk("Deleting nonexistent file (%s:%lu), %d\n",
		       kdevname(inode->i_dev),
		       inode->i_ino, inode->i_nlink);
		inode->i_nlink=1;
	}
	dir->i_ctime = dir->i_mtime = CURRENT_TIME;
	inode->i_nlink--;
	inode->i_ctime = dir->i_ctime;
	retval = 0;
end_unlink:
	iput(inode);
	iput(dir);
	PRINTF("ffs_rmdir(%s): retval = %d\n", name, retval);
	return retval;
}

int ffs_symlink(struct inode *dir, const char * name, int len, const char * symname)
{
#ifdef	LATER
	struct ffs_dir_entry * de;
	struct inode * inode;
	struct buffer_head * name_block;
	char * name_block_data;
	struct super_block * sb;
	int i;
	char c;
	struct buffer_head * bh;

	if (!(inode = ffs_new_inode(dir))) {
		iput(dir);
		return -ENOSPC;
	}
	inode->i_mode = S_IFLNK | 0777;
	inode->i_op = &ffs_symlink_iops;
	name_block = ffs_file_bread(inode,0,1);
	if (!name_block) {
		iput(dir);
		inode->i_nlink--;
		inode->i_dirt = 1;
		iput(inode);
		return -ENOSPC;
	}
	sb = inode->i_sb;
	name_block_data = name_block->b_data;
	i = 0;
	while (i < sb->sv_block_size_1 && (c = *(symname++)))
		name_block_data[i++] = c;
	name_block_data[i] = 0;
	mark_buffer_dirty(name_block, 1);
	brelse(name_block);
	inode->i_size = i;
	inode->i_dirt = 1;
	bh = ffs_find_entry(dir,name,len,&de);
	if (bh) {
		inode->i_nlink--;
		inode->i_dirt = 1;
		iput(inode);
		brelse(bh);
		iput(dir);
		return -EEXIST;
	}
	i = ffs_add_entry(dir, name, len, &bh, &de);
	if (i) {
		inode->i_nlink--;
		inode->i_dirt = 1;
		iput(inode);
		iput(dir);
		return i;
	}
	ino = inode->i_ino;
	mark_buffer_dirty(bh, 1);
	brelse(bh);
	iput(dir);
	iput(inode);
	return 0;
#else
	PRINTF("ffs_symlink(%s) -> %s, next time\n", name, symname);
	return -EAGAIN;
#endif
}

int ffs_link(struct inode * oldinode, struct inode *dir, const char * name, int len)
{
#ifdef	LATER
	int error;
	struct ffs_diren * de;

	if (S_ISDIR(oldinode->i_mode)) {
		iput(oldinode);
		iput(dir);
		return -EPERM;
	}
	if (oldinode->i_nlink >= 127) {
		iput(oldinode);
		iput(dir);
		return -EMLINK;
	}
	de = ffs_find_entry(dir, name, len);
	if (de) {
		iput(dir);
		iput(oldinode);
		return -EEXIST;
	}
	error = ffs_add_entry(dir, name, len, &bh, &de);
	if (error) {
		iput(dir);
		iput(oldinode);
		return error;
	}
	de->inode = oldinode->i_ino;
	iput(dir);
	oldinode->i_nlink++;
	oldinode->i_ctime = CURRENT_TIME;
	oldinode->i_dirt = 1;
	iput(oldinode);
	return 0;
#else
	PRINTF("ffs_link(%s): next time\n", name);
	return -EAGAIN;
#endif
}

#ifdef	LATER
/* return 1 if `new' is a subdir of `old' on the same device */
static int subdir(struct inode * new_inode, struct inode * old_inode)
{
	int ino;
	int result;

	new_inode->i_count++;
	result = 0;
	for (;;) {
		if (new_inode == old_inode) {
			result = 1;
			break;
		}
		if (new_inode->i_dev != old_inode->i_dev)
			break;
		ino = new_inode->i_ino;
		if (ffs_lookup(new_inode,"..",2,&new_inode))
			break;
		if (new_inode->i_ino == ino) /* root dir reached ? */
			break;
	}
	iput(new_inode);
	return result;
}

#define PARENT_INO(buffer) \
(((struct ffs_dir_entry *) ((buffer) + 1*SYSV_DIRSIZE))->inode)

/*
 * rename uses retrying to avoid race-conditions: at least they should be minimal.
 * it tries to allocate all the blocks, then sanity-checks, and if the sanity-
 * checks fail, it tries to restart itself again. Very practical - no changes
 * are done until we know everything works ok.. and then all the changes can be
 * done in one fell swoop when we have claimed all the buffers needed.
 *
 * Anybody can rename anything with this: the permission checks are left to the
 * higher-level routines.
 */
static int do_ffs_rename(struct inode * old_dir, const char * old_name, int old_len,
	struct inode * new_dir, const char * new_name, int new_len, int must_be_dir)
{
	struct inode * old_inode, * new_inode;
	struct buffer_head * old_bh, * new_bh, * dir_bh;
	struct ffs_dir_entry * old_de, * new_de;
	int retval;

	goto start_up;
try_again:
	brelse(old_bh);
	brelse(new_bh);
	brelse(dir_bh);
	iput(old_inode);
	iput(new_inode);
	current->counter = 0;
	schedule();
start_up:
	old_inode = new_inode = NULL;
	old_bh = new_bh = dir_bh = NULL;
	old_de = ffs_find_entry(old_dir, old_name, old_len);
	retval = -ENOENT;
	if (!old_de)
		goto end_rename;
	old_inode = __iget(old_dir->i_sb, old_de->inode, 0); /* don't cross mnt-points */
	if (!old_inode)
		goto end_rename;
	if (must_be_dir && !S_ISDIR(old_inode->i_mode))
		goto end_rename;
	retval = -EPERM;
	if ((old_dir->i_mode & S_ISVTX) && 
	    current->fsuid != old_inode->i_uid &&
	    current->fsuid != old_dir->i_uid && !fsuser())
		goto end_rename;
	new_de = ffs_find_entry(new_dir, new_name, new_len);
	if (new_de) {
		new_inode = __iget(new_dir->i_sb, new_de->inode, 0);
		if (!new_inode)
			new_de = NULL;
	}
	if (new_inode == old_inode) {
		retval = 0;
		goto end_rename;
	}
	if (new_inode && S_ISDIR(new_inode->i_mode)) {
		retval = -EISDIR;
		if (!S_ISDIR(old_inode->i_mode))
			goto end_rename;
#ifdef	LATER
		retval = -EINVAL;
		if (subdir(new_dir, old_inode))
			goto end_rename;
		retval = -ENOTEMPTY;
		if (!empty_dir(new_inode))
			goto end_rename;
#endif
		retval = -EBUSY;
		if (new_inode->i_count > 1)
			goto end_rename;
	}
	retval = -EPERM;
	if (new_inode && (new_dir->i_mode & S_ISVTX) && 
	    current->fsuid != new_inode->i_uid &&
	    current->fsuid != new_dir->i_uid && !fsuser())
		goto end_rename;
	if (S_ISDIR(old_inode->i_mode)) {
		retval = -ENOTDIR;
		if (new_inode && !S_ISDIR(new_inode->i_mode))
			goto end_rename;
#ifdef	LATER
		retval = -EINVAL;
		if (subdir(new_dir, old_inode))
			goto end_rename;
		retval = -EIO;
		dir_bh = ffs_file_bread(old_inode,0,0);
		if (!dir_bh)
			goto end_rename;
		if (PARENT_INO(dir_bh->b_data) != old_dir->i_ino)
			goto end_rename;
		retval = -EMLINK;
		if (!new_inode && new_dir->i_nlink >= new_dir->i_sb->sv_link_max)
			goto end_rename;
#endif
	}
	if (!new_de) {
		retval = ffs_add_entry(new_dir,new_name,new_len,&new_bh,&new_de);
		if (retval)
			goto end_rename;
	}
/* sanity checking before doing the rename - avoid races */
	if (new_inode && (new_de->inode != new_inode->i_ino))
		goto try_again;
	if (new_de->inode && !new_inode)
		goto try_again;
	if (old_de->inode != old_inode->i_ino)
		goto try_again;
/* ok, that's it */
	old_de->inode = 0;
	new_de->inode = old_inode->i_ino;
	old_dir->i_ctime = old_dir->i_mtime = CURRENT_TIME;
	old_dir->i_dirt = 1;
	new_dir->i_ctime = new_dir->i_mtime = CURRENT_TIME;
	new_dir->i_dirt = 1;
	if (new_inode) {
		new_inode->i_nlink--;
		new_inode->i_ctime = CURRENT_TIME;
		new_inode->i_dirt = 1;
	}
	mark_buffer_dirty(old_bh, 1);
	mark_buffer_dirty(new_bh, 1);
	if (dir_bh) {
		PARENT_INO(dir_bh->b_data) = new_dir->i_ino;
		mark_buffer_dirty(dir_bh, 1);
		old_dir->i_nlink--;
		old_dir->i_dirt = 1;
		if (new_inode) {
			new_inode->i_nlink--;
			new_inode->i_dirt = 1;
		} else {
			new_dir->i_nlink++;
			new_dir->i_dirt = 1;
		}
	}
	retval = 0;
end_rename:
	brelse(dir_bh);
	brelse(old_bh);
	brelse(new_bh);
	iput(old_inode);
	iput(new_inode);
	iput(old_dir);
	iput(new_dir);
	return retval;
}
#endif

/*
 * Ok, rename also locks out other renames, as they can change the parent of
 * a directory, and we don't want any races. Other races are checked for by
 * "do_rename()", which restarts if there are inconsistencies.
 *
 * Note that there is no race between different filesystems: it's only within
 * the same device that races occur: many renames can happen at once, as long
 * as they are on different partitions.
 */
int ffs_rename(struct inode * old_dir, const char * old_name, int old_len,
	struct inode * new_dir, const char * new_name, int new_len,
	int must_be_dir)
{
#ifdef	LATER
	static struct wait_queue * wait = NULL;
	static int lock = 0;
	int result;

	while (lock)
		sleep_on(&wait);
	lock = 1;
	result = do_ffs_rename(old_dir, old_name, old_len,
		new_dir, new_name, new_len, must_be_dir);
	lock = 0;
	wake_up(&wait);
	return result;
#else
	PRINTF("ffs_rename(%s -> %s): again\n", old_name, new_name);
	return -EAGAIN;
#endif
}
