#include <linux/module.h>
#include <linux/errno.h>
#include <linux/malloc.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/locks.h>
#include <linux/stat.h>
#include <linux/string.h>
#include <asm/bitops.h>
#include <asm/segment.h>

#include <linux/debug.h>
#include <linux/flash_fs.h>
#include <linux/flash.h>

/* super block ops */

void ffs_put_super(struct super_block *);
void ffs_statfs(struct super_block *, struct statfs *, int);
STATIC int ffs_remount_fs(struct super_block *, int *, char *);
static int notify_change(struct inode *, struct iattr *);

static const struct super_operations ffs_sops =
{
	ffs_read_inode,			/* read_inode */
	NULL,				/* notify_change */
	ffs_write_inode,		/* write_inode */
	ffs_put_inode,			/* put_inode */
	ffs_put_super,			/* put_super */
	NULL,				/* write_super */
	ffs_statfs,			/* statfs */
	ffs_remount_fs,			/* remount_fs */
};

/*
 * local time (FlashFS) to GMT (Unix)
 */

static inline time_t local_to_gmt(time_t t)
{
	extern struct timezone sys_tz;
	return t + sys_tz.tz_minuteswest * 60;
}

/* super block ops */

/*
 * mount.  This gets one thing, the root directory inode.  It does a
 * bunch of guessed-at consistency checks.
 */

struct super_block *
ffs_read_super(struct super_block *s, void *options, int silent)
{
	int	error;

	MOD_INC_USE_COUNT;

	/*
	 * Fill in the super block struct
	 */
	if ((error = flash_mount(s->s_dev)) < 0) {
		printk("ffs_read_super(0x%x) = %d\n", s->s_dev, error);
		goto bail;
	}

	/*
	 * fill in standard stuff
	 */

	// s->s_flags |= MS_RDONLY;
	s->s_magic = 0xfdfd;
	s->s_blocksize = 4096;
	s->s_blocksize_bits = 12;
	s->s_op = (struct super_operations *) &ffs_sops;

	/*
	 * all set.  try it out.
	 */
	s->s_mounted = iget(s, FLASH_ROOTINO);
	unlock_super(s);

	if (!s->s_mounted) {
		printk("FlashFS: ffs_read_super: inode get failed\n");
		s->s_dev = 0;
		MOD_DEC_USE_COUNT;
		return 0;
	}

	s->s_mounted->i_atime = local_to_gmt(0);
	s->s_mounted->i_mtime = local_to_gmt(0);
	s->s_mounted->i_ctime = local_to_gmt(0);

	return s;

 bail:
	s->s_dev = 0;
	unlock_super(s);
	MOD_DEC_USE_COUNT;
	return 0;
}

void
ffs_set_iops(struct inode *inode)
{
	if (S_ISREG(inode->i_mode))
		inode->i_op = &ffs_file_iops;
	else if (S_ISDIR(inode->i_mode))
		inode->i_op = &ffs_dir_iops;
	else if (S_ISLNK(inode->i_mode))
		inode->i_op = &ffs_symlink_iops;
	else if (S_ISFIFO(inode->i_mode))
		init_fifo(inode);
	else if (S_ISCHR(inode->i_mode))
		inode->i_op = &chrdev_inode_operations;
	else if (S_ISBLK(inode->i_mode))
		inode->i_op = &blkdev_inode_operations;
}

/*
 * unmount.
 */

void ffs_put_super(struct super_block *s)
{
	ffs_bsync(0);
	s->s_dev = 0;
	MOD_DEC_USE_COUNT;
}

/*
 * statfs.  For free inode counts we report the count of dnodes in the
 * directory band -- not exactly right but pretty analogous.
 */

void ffs_statfs(struct super_block *s, struct statfs *buf, int bufsiz)
{
	struct statfs tmp;
	struct flash_device *fd;

	fd = flash_device(s->s_dev);
	if (fd == 0)
		return;
	/*
	 * fill in the user statfs struct
	 */
	tmp.f_type = s->s_magic;
	tmp.f_bsize = fd->pagesize;
	tmp.f_blocks = fd->pages;
	tmp.f_bfree = fd->freepages;
	tmp.f_bavail = fd->freepages + fd->deletepages;
	tmp.f_files = fd->allinodes;
	tmp.f_ffree = fd->freeinodes;
	tmp.f_namelen = 254;
	memcpy_tofs(buf, &tmp, bufsiz);
}

/*
 * remount.  Don't let read only be turned off.
 */

STATIC int ffs_remount_fs(struct super_block *s, int *flags, char *data)
{
	return 0;
}

/* directory ops */

static struct file_system_type ffs_fs_type = {
        ffs_read_super, "flashfs", 1, NULL
};

int
init_flash_fs(void)
{
	return(register_filesystem(&ffs_fs_type));
}

#ifdef MODULE
int init_module(void)
{
	int status;

	if ((status = init_ffs_fs()) == 0)
		register_symtab(0);
	return status;
}

void cleanup_module(void)
{
        unregister_filesystem(&ffs_fs_type);
}

#endif

