#ifdef	__KERNEL__

#include <linux/module.h>

#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/stat.h>
#include <linux/string.h>
#include <linux/locks.h>
#include <linux/malloc.h>

#include <linux/flash_fs.h>
#include <linux/flash.h>

#include <asm/segment.h>
#else
#include "system.h"
#include "errno.h"
#include "flash_fs.h"
#include "flash.h"
#endif

STATIC struct ffs_dirent *
ffs_ialloc(struct inode *dir, const char *name, mode_t mode, ino_t *ino, kdev_t rdev)
{
	int	entsize;
	struct 	ffs_dirent *de;

	*ino = flash_ialloc(dir->i_ffs_fd, S_ISDIR(mode));
	if (*ino == 0)
		return(0);	

	entsize = strlen(name) + sizeof(*de);
	de = ffs_empty_dirent(dir, entsize);
	if (de) {
		de->noff = de->name - (char *) de;
		de->mode = mode;
		de->ino  = *ino;
		de->rdev = rdev;
		strcpy(de->name, name);
		ffs_bsync(dir);
		return de;
	}

	printf("ffs_ialloc: unable to allocate a new dirent\n");
	flash_irelse(dir->i_ffs_fd, *ino);	/* no longer needed */
	return NULL;
}

void 
ffs_truncate(struct inode *inode)
{
	int	i;
	struct	flash_imap	*ip;

	ip = inode->i_ffs_ip;
	ASSERT(ip);

	for (i = 0; i < ip->mapsize; ++i) {
		if (ip->map[i] != 0) {
			ffs_bdelete(inode, i, ip->map[i]);
			ip->map[i] = 0; 
		}
	}

	inode->i_size = 0;
	ffs_update_dirent(inode);
	ffs_bsync(inode);
}

void 
ffs_put_inode(struct inode *inode)
{
	ASSERT(inode);
	ASSERT(inode->i_dev);
	if (inode->i_nlink == 0) {
		ffs_truncate(inode);
		flash_irelse(inode->i_ffs_fd, inode->i_ino);
		clear_inode(inode);
	} else {
		ffs_update_dirent(inode);
	}
}

#ifdef	__KERNEL__
struct inode * 
ffs_new_inode(struct inode *dir, const char *name, mode_t mode, kdev_t rdev)
{
	ino_t	ino;
	struct 	inode * inode;

	ASSERT(dir);

	if (ffs_ialloc(dir, name, mode, &ino, rdev) == 0)
		return NULL;

	if ((inode = get_empty_inode()) == 0)
		return NULL;

	inode->i_sb = dir->i_sb;
	inode->i_ino = ino;
	inode->i_flags = inode->i_sb->s_flags;
	inode->i_mode = mode;
	inode->i_count = 1;
	inode->i_dev = inode->i_sb->s_dev;
	inode->i_uid = current->fsuid;
	inode->i_gid = (dir->i_mode & S_ISGID) ? dir->i_gid : current->fsgid;
	inode->i_dirt = 0;
	inode->i_mtime = inode->i_atime = inode->i_ctime = 0; // CURRENT_TIME;
	inode->i_op = NULL;
	inode->i_blksize = 0;
	inode->i_blocks = 0;
	inode->i_size = 0;		/* ditto */
	inode->i_rdev = rdev;
	ffs_set_iops(inode);
	inode->i_ffs_fd = dir->i_ffs_fd;
	inode->i_ffs_ip = flash_getimap(inode->i_ffs_fd, inode->i_ino);
	inode->i_ffs_ip->parent = dir->i_ino;
	inode->i_nlink = inode->i_ffs_ip->nlink;
	ASSERT(dir->i_ffs_ip);
	ASSERT(dir->i_nlink == dir->i_ffs_ip->nlink);
	dir->i_nlink++;
	dir->i_ffs_ip->nlink++;
	insert_inode_hash(inode);
	return inode;
}
#endif

int
ffs_new_dirent(struct inode *dir, const char *name, ino_t ino, int mode, int size, kdev_t rdev)
{
	int	entsize;
	struct 	ffs_dirent *de;

	entsize = strlen(name) + sizeof(*de);
	de = ffs_empty_dirent(dir, entsize);
	if (de) {
		de->noff = de->name - (char *) de;
		de->mode = mode;
		de->ino = ino;
		de->size = size;
		de->rdev = rdev;
		de->uid = current->fsuid;
		de->gid = (dir->i_mode & S_ISGID) ? dir->i_gid : current->fsgid;
		strcpy(de->name, name);
		ffs_bsync(dir);
	}
	return((int) de);
}

int
ffs_dup_inode(struct inode *dir, const char *name, struct inode *inode)
{
	ASSERT(dir);

	if (ffs_new_dirent(dir, name, inode->i_ino, inode->i_mode, 
				      inode->i_size, inode->i_rdev) == 0)
		return -ENOSPC;
	ASSERT(dir->i_ffs_ip);
	ASSERT(dir->i_nlink == dir->i_ffs_ip->nlink);
	dir->i_nlink++;
	dir->i_ffs_ip->nlink++;
	return 0;
}

void
ffs_read_inode(struct inode *inode)
{
	struct ffs_dirent *de;

	/* be ready to bail out */

	ASSERT(inode->i_ino);
	inode->i_ffs_fd = flash_device(inode->i_dev);
	if (inode->i_ffs_fd == 0)
		return;
	inode->i_ffs_ip = flash_getimap(inode->i_ffs_fd, inode->i_ino);
	inode->i_nlink = inode->i_ffs_ip->nlink;
	inode->i_op = 0;
	inode->i_mode = 0;
	inode->i_uid = 0;
	inode->i_gid = 0;

	if (inode->i_ino == FLASH_ROOTINO) {
		inode->i_mode = S_IFDIR | 0777;
	} else {
		de = ffs_get_dirent(inode, 0); 
		if (de == 0) {
			printf("ffs_read_inode warning: inode %d, de == 0\n", 
				(int) inode->i_ino);
			return;
		}
		inode->i_mode = de->mode;
		inode->i_size = de->size;
		inode->i_rdev = to_kdev_t(de->rdev);
		if (DE_VERSION(de) > 0) {
			inode->i_uid = de->uid;
			inode->i_gid = de->gid;
		}
	}
	if (S_ISDIR(inode->i_mode))
		inode->i_size = 1024;

	ffs_set_iops(inode);
	/*
	 * these fields must be filled in from the dir entry, which we don't
	 * have but lookup does.  It will fill them in before letting the
	 * inode out of its grasp.
	 */

	inode->i_atime = 0;
	inode->i_mtime = 0;
	inode->i_ctime = 0;
}

void 
ffs_write_inode(struct inode * inode)
{
#ifndef	EPROM
	if (inode->i_dirt)
		ffs_update_dirent(inode);
	inode->i_dirt = 0;
	ffs_bsync(inode);
#endif
}
