#ifdef	__KERNEL__
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/flash_fs.h>
#include <linux/string.h>
#include <linux/stat.h>
#include <linux/errno.h>
#include <linux/flash.h>
#else
#include "system.h"
#include "errno.h"
#include "flash_fs.h"
#include "flash.h"
#endif
/*
 *	ffs_dirlookup()
 *
 * finds an entry in the specified directory with the wanted name. It
 * returns the cache buffer in which the entry was found, and the entry
 * itself (as a parameter - res_dir). It does NOT read the inode of the
 * entry - you'll have to do that yourself if you want to.
 */
STATIC struct ffs_dirent *
ffs_dirlookup(struct inode *dir, const char * name, int len, ino_t *ino)
{
	struct 	ffs_dirent *de;
	
	ASSERT(S_ISDIR(dir->i_mode));
	if (name[0] == '.') {
		if (len == 1) {
			*ino = dir->i_ino;
			return (struct ffs_dirent *) 1;
		}
		if (name[1] == '.' && len == 2) {
			ASSERT(dir->i_ffs_ip && dir->i_ffs_ip->parent);
			*ino = dir->i_ffs_ip->parent;
			return (struct ffs_dirent *) 1;
		}
	}

	de = ffs_find_dirent(dir, name, len, 0);
	if (de)
		*ino = de->ino;
	return(de);
}

#ifndef	__KERNEL__

#define	FFS_ISSLASH(c)	(c == '/')

struct inode *
ffs_lookup(struct inode *root_inode, char *path, struct inode **parent, char **name)
{
	struct	inode	*inode;
	ino_t	ino;
	char	*n, tmp[128];

	if (*path == 0)
		goto error;

	while (FFS_ISSLASH(*path)) 
		path++;
	inode = root_inode;
	if (*path == 0) {
		*parent = 0;
		*name = 0;
		return(inode);
	}

	for ( ;; ) {
		*name = path;
		for (n = tmp; *path && !FFS_ISSLASH(*path); ) {
			if (n >= tmp + sizeof(tmp) - 1)
				goto error;
			*n++ = *path++;
		}
		*n = 0;
		while(FFS_ISSLASH(*path))
			++path;

		if (!S_ISDIR(inode->i_mode)) {
			PRINTF("ffs_lookup(%s): stopped at %s(0x%x).\n",
				path, tmp, inode->i_mode);
			goto error;
		}
		*parent = inode;
		if (ffs_dirlookup(inode, tmp, strlen(tmp), &ino) == 0)
			return(0);
		inode = (struct inode *) iget(inode, ino);
		if (inode == 0 || *path == 0)
			return(inode);
	}
error:
	*parent = 0;
	*name = 0;
	return(0);
}

#else

int 
ffs_lookup(struct inode *dir, const char *name, int len, struct inode **result)
{
	ino_t 	ino;

	*result = NULL;
	if (!dir)
		return -ENOENT;
	if (!S_ISDIR(dir->i_mode)) {
		iput(dir);
		return -ENOENT;
	}
	if (ffs_dirlookup(dir, name, len, &ino) == 0) {
		iput(dir);
		return -ENOENT;
	}
	
	if (!(*result = iget(dir->i_sb, ino))) {
		iput(dir);
		return -EACCES;
	}
	iput(dir);
	return 0;
}

int 
ffs_create(struct inode *dir, const char * name, int len, int mode,
							struct inode ** result)
{
	struct inode * inode;

	*result = NULL;
	if (!dir)
		return -ENOENT;
	ASSERT(strlen(name) == len);
	inode = ffs_new_inode(dir, name, mode, 0);
	if (!inode) {
		iput(dir);
		return -ENOSPC;
	}
	iput(dir);
	*result = inode;
	return 0;
}

int 
ffs_mknod(struct inode *dir, const char *name, int len, int mode, int rdev)
{
	ino_t	ino;
	struct 	inode * inode;

	if (!dir)
		return -ENOENT;
	if (ffs_dirlookup(dir, name, len, &ino) != 0) {
		iput(dir);
		return -EEXIST;
	}
	ASSERT(strlen(name) == len);
	inode = ffs_new_inode(dir, name, mode, to_kdev_t(rdev));
	if (!inode) {
		iput(dir);
		return -ENOSPC;
	}
	inode->i_uid = current->fsuid;
		
	if (S_ISDIR(inode->i_mode)) {
		if (dir->i_mode & S_ISGID)
			inode->i_mode |= S_ISGID;
	}
	iput(dir);
	iput(inode);
	return 0;
}

int 
ffs_mkdir(struct inode *dir, const char *name, int len, int mode)
{
	ino_t	ino;
	int	retval = 0;
	struct 	inode *inode = 0;

	ASSERT(dir);
	if (name[0] == '.') {
		if (len == 1)
			goto ret;
		if (name[1] == '.' && len == 2)
			goto ret;
	}
	if (ffs_dirlookup(dir, name, len, &ino) != 0) {
		retval = -EEXIST;
		goto ret;
	}
	if (dir->i_nlink >= 0xFFFE) {
		retval = -EMLINK;
		goto ret;
	}
	mode = S_IFDIR | (mode & 0777 & ~current->fs->umask);
	if (dir->i_mode & S_ISGID)
		mode |= S_ISGID;
	ASSERT(strlen(name) == len);
	inode = ffs_new_inode(dir, name, mode, 0);
	if (!inode) {
		retval = -ENOSPC;
		goto ret;
	}

ret:
	iput(dir);
	iput(inode);
	return retval;
}

int 
ffs_rmdir(struct inode *dir, const char * name, int len)
{
	int 	retval;
	ino_t	ino;
	struct 	inode *inode = 0;
	struct 	ffs_dirent *de;

	inode = NULL;
	retval = -ENOENT;
	if (ffs_dirlookup(dir, name, len, &ino) == 0)
		goto end_rmdir;
	retval = -EPERM;
	if (!(inode = iget(dir->i_sb, ino)))
		goto end_rmdir;
        if ((dir->i_mode & S_ISVTX) && !fsuser() &&
            current->fsuid != inode->i_uid &&
            current->fsuid != dir->i_uid)
		goto end_rmdir;
	if (inode->i_dev != dir->i_dev)
		goto end_rmdir;
	if (inode == dir)	/* we may not delete ".", but "../dir" is ok */
		goto end_rmdir;
	if (!S_ISDIR(inode->i_mode)) {
		retval = -ENOTDIR;
		goto end_rmdir;
	}
	if (inode->i_nlink != 2) {
		retval = -ENOTEMPTY;
		goto end_rmdir;
	}
	if (inode->i_count > 1) {
		retval = -EBUSY;
		goto end_rmdir;
	}
	ASSERT(inode->i_ffs_ip->nlink == inode->i_nlink);
	inode->i_nlink = 0;
	inode->i_ffs_ip->nlink = 0;
	ASSERT(dir->i_nlink > 0);
	ASSERT(dir->i_ffs_ip->nlink == dir->i_nlink);
	dir->i_nlink--;
	dir->i_ffs_ip->nlink--;
	inode->i_ctime = dir->i_ctime = dir->i_mtime = 0; // CURRENT_TIME;
	de = ffs_get_dirent(inode, 2);	/* set dirty flag for re-write */
	ASSERT(de);
	if (de) {
		de->ino = 0;
		ffs_bsync(dir);
	}
	retval = 0;
end_rmdir:
	iput(dir);
	iput(inode);
	return retval;
}

int 
ffs_unlink(struct inode *dir, const char * name, int len)
{
	int 	retval;
	ino_t	ino;
	struct 	inode *inode;
	struct	ffs_dirent *de;

	retval = -ENOENT;
	inode = NULL;
	if (ffs_dirlookup(dir, name, len, &ino) == 0)
		goto end_unlink;
	if (!(inode = iget(dir->i_sb, ino)))
		goto end_unlink;
	ASSERT(ino == inode->i_ino);
	retval = -EPERM;
	if (S_ISDIR(inode->i_mode))
		goto end_unlink;
	if ((dir->i_mode & S_ISVTX) && !fsuser() &&
	    current->fsuid != inode->i_uid &&
	    current->fsuid != dir->i_uid)
		goto end_unlink;
	if (ino != inode->i_ino) {
		retval = -ENOENT;
		goto end_unlink;
	}
	if (!inode->i_nlink) {
		printf("Deleting nonexistent file (%s:%lu), %d\n",
		       kdevname(inode->i_dev),
		       inode->i_ino, inode->i_nlink);
		inode->i_nlink = 1;
		inode->i_ffs_ip->nlink = 1;
	}
	dir->i_ctime = dir->i_mtime = 0; // CURRENT_TIME;
	ASSERT(inode->i_nlink > 0);
	ASSERT(inode->i_ffs_ip->nlink == inode->i_nlink);
	inode->i_nlink--;
	inode->i_ffs_ip->nlink--;
	ASSERT(dir->i_nlink > 0);
	ASSERT(dir->i_ffs_ip->nlink == dir->i_nlink);
	dir->i_nlink--;
	dir->i_ffs_ip->nlink--;
	inode->i_ctime = dir->i_ctime;
	de = ffs_find_dirent(dir, name, len, 2);
	ASSERT(de != 0);
	if (de) {
		de->ino = 0;
		ffs_bsync(dir);
	}
		
	retval = 0;
end_unlink:
	ASSERT(inode == 0 || inode->i_count > 0);
	ASSERT(dir->i_count > 0);
	iput(inode);
	iput(dir);
	return retval;
}

int 
ffs_symlink(struct inode *dir, const char * name, int len, const char * symname)
{
	struct inode * inode;
	struct ffs_buf *bp;
	ino_t	ino;
	int	i;

	if (ffs_dirlookup(dir, name, len, &ino) != 0) {
		iput(dir);
		return -EEXIST;
	}

	ASSERT(strlen(name) == len);
	inode = ffs_new_inode(dir, name, S_IFLNK | 0777, 0);
	if (inode == 0) {
		iput(dir);
		return -ENOSPC;
	}

	bp = ffs_getblk(inode, 0, 1);
	if (bp == 0) {
		iput(dir);
		inode->i_nlink = 0;
		iput(inode);
		return -ENOSPC;
	}

	strncpy(bp->buf, symname, FFS_BLKSIZE(inode) - 1);
	i = 0;
	while (i < FFS_BLKSIZE(inode) - 1 && (bp->buf[i++] = *symname++));
	inode->i_size = i;
	inode->i_dirt = 1;

	iput(dir);
	iput(inode);
	return 0;
}

int 
ffs_link(struct inode * oldinode, struct inode *dir, const char * name, int len)
{
	ino_t ino;
	int	error;
	extern	int ffs_dup_inode(struct inode *, const char *, struct inode *);

	if (S_ISDIR(oldinode->i_mode)) {
		iput(oldinode);
		iput(dir);
		return -EPERM;
	}
	if (oldinode->i_nlink >= 127) {
		iput(oldinode);
		iput(dir);
		return -EMLINK;
	}
	if (ffs_dirlookup(dir, name, len, &ino) != 0) {
		iput(dir);
		iput(oldinode);
		return -EEXIST;
	}
	ASSERT(strlen(name) == len);
	if ((error = ffs_dup_inode(dir, name, oldinode))) {
		iput(oldinode);
		iput(dir);
		return error;
	}
	iput(dir);
	oldinode->i_nlink++;
	oldinode->i_ffs_ip->nlink++;
	oldinode->i_ctime = CURRENT_TIME;
	iput(oldinode);
	return 0;
}

/* return 1 if `new' is a subdir of `old' on the same device */
STATIC int 
subdir(struct inode * new_inode, struct inode * old_inode)
{
	int 	ino;
	int 	result;
	struct	flash_imap	*ip;

	result = 0;
	if (new_inode->i_dev == old_inode->i_dev) {
		ino = new_inode->i_ino; 
		for (;;) {
			if (ino == FLASH_ROOTINO)
				break;
			ip = flash_getimap(new_inode->i_ffs_fd, ino);
			ino = ip->parent; 
			if (old_inode->i_ino == ino)
				break;;
		}
	}
	return result;
}

 /*
 * Anybody can rename anything with this: the permission checks are left to the
 * higher-level routines.
 */
int 
ffs_rename(struct inode * old_dir, const char * old_name, int old_len,
	struct inode * new_dir, const char * new_name, int new_len, 
	int must_be_dir)
{
	struct inode * old_inode, * new_inode;
	ino_t	old_ino, new_ino;
	int retval;
	struct	ffs_dirent *de = 0;

	old_inode = new_inode = NULL;
	retval = -ENOENT;
	if (ffs_dirlookup(old_dir, old_name, old_len, &old_ino) == 0)
		goto end_rename;
	old_inode = __iget(old_dir->i_sb, old_ino, 0); /* don't cross mnt */
	if (!old_inode)
		goto end_rename;
	if (must_be_dir && !S_ISDIR(old_inode->i_mode))
		goto end_rename;
	retval = -EPERM;
	if ((old_dir->i_mode & S_ISVTX) && 
	    current->fsuid != old_inode->i_uid &&
	    current->fsuid != old_dir->i_uid && !fsuser())
		goto end_rename;
	if (ffs_dirlookup(new_dir, new_name, new_len, &new_ino) != 0)
		new_inode = __iget(new_dir->i_sb, new_ino, 0);
	
	if (new_inode == old_inode) {
		retval = 0;
		goto end_rename;
	}
	if (new_inode && S_ISDIR(new_inode->i_mode)) {
		retval = -EISDIR;
		if (!S_ISDIR(old_inode->i_mode))
			goto end_rename;
		retval = -EINVAL;
		if (subdir(new_dir, old_inode))
			goto end_rename;
		retval = -ENOTEMPTY;
		if (new_inode->i_nlink != 2)	/* non-empty */
			goto end_rename;
		retval = -EBUSY;
		if (new_inode->i_count > 1)
			goto end_rename;
	}
	retval = -EPERM;
	if (new_inode && (new_dir->i_mode & S_ISVTX) && 
	    current->fsuid != new_inode->i_uid &&
	    current->fsuid != new_dir->i_uid && !fsuser())
		goto end_rename;
	if (S_ISDIR(old_inode->i_mode)) {
		retval = -ENOTDIR;
		if (new_inode && !S_ISDIR(new_inode->i_mode))
			goto end_rename;
		retval = -EINVAL;
		if (subdir(new_dir, old_inode))
			goto end_rename;
	}

	if (new_inode) {
		de = ffs_find_dirent(new_dir, new_name, new_len, 2);
		ASSERT(de != 0);
		if (de)
			de->ino = 0;

		ASSERT(new_inode->i_nlink > 0);
		ASSERT(new_inode->i_ffs_ip->nlink == new_inode->i_nlink);
		new_inode->i_nlink--;
		new_inode->i_ffs_ip->nlink--;
		ASSERT(new_dir->i_nlink > 0);
		ASSERT(new_dir->i_ffs_ip->nlink == new_dir->i_nlink);
		new_dir->i_nlink--;
		new_dir->i_ffs_ip->nlink--;
	} 

	if (ffs_rename_dirent(old_dir, old_ino, new_dir, new_name, new_len)) {
		ASSERT(old_dir->i_nlink > 0);
		old_dir->i_ffs_ip->nlink--;
		old_dir->i_nlink--;
		new_dir->i_nlink++;
		new_dir->i_ffs_ip->nlink++;
		retval = 0;
		ffs_bsync(new_dir);
		ffs_bsync(old_dir);
	} else {
		retval = -EIO;
		if (new_inode) {	/* put back deleted inode */
			if (de)
				de->ino = new_ino;
			new_inode->i_nlink++;
			new_inode->i_ffs_ip->nlink++;
			new_dir->i_nlink++;
			new_dir->i_ffs_ip->nlink++;
		}
	}

end_rename:
	ASSERT(old_inode == 0 || old_inode->i_count > 0);
	ASSERT(new_inode == 0 || new_inode->i_count > 0);
	ASSERT(old_dir == 0 || old_dir->i_count > 0);
	ASSERT(new_dir == 0 || new_dir->i_count > 0);

	iput(old_inode);
	iput(new_inode);
	iput(old_dir);
	iput(new_dir);
	return retval;
}
#endif
